/******************************************************************************
 * $Header: /CommonBe/agmsmith/Programming/AGMSRAMFileSystem/RCS/AGMSRAMFSMounter.cpp,v 1.66 2003/05/24 21:04:16 agmsmith Exp $
 *
 * This program is used for mounting the AGMSRAMFileSystem and monitoring
 * its activity.  It also contains a test suite for verifying the functionality
 * of the file system, and a pretty GUI for selecting tests.
 *
 * $Log: AGMSRAMFSMounter.cpp,v $
 * Revision 1.66  2003/05/24 21:04:16  agmsmith
 * New command line arguments to mount() - more generic keyword based
 * rather than raw fixed size binary data.
 *
 * Revision 1.65  2003/05/04 19:23:19  agmsmith
 * Added more Query tests.
 *
 * Revision 1.64  2003/04/25 17:15:02  agmsmith
 * *** empty log message ***
 *
 * Revision 1.63  2003/04/25 01:30:45  agmsmith
 * Oops.
 *
 * Revision 1.62  2003/04/24 23:09:11  agmsmith
 * Move directory tests to a subdirectory, so .indices doesn't interfere.
 *
 * Revision 1.61  2003/04/21 00:23:22  agmsmith
 * More query tests.
 *
 * Revision 1.60  2002/08/16 19:02:46  agmsmith
 * Added an initial Query test.
 *
 * Revision 1.59  2002/07/04 19:36:47  agmsmith
 * Added indices test.
 *
 * Revision 1.58  2002/06/30 17:58:14  agmsmith
 * File system name change.
 *
 * Revision 1.57  2002/06/12 19:37:01  agmsmith
 * Added number of times to run the big buffer test, to reduce
 * swapping with really huge buffers.
 *
 * Revision 1.56  2002/06/08 20:30:49  agmsmith
 * Apparently you can open directories as files with write access,
 * but not actually write to them.
 *
 * Revision 1.55  2002/05/24 16:37:20  agmsmith
 * Added stat_attr tests.
 *
 * Revision 1.54  2002/05/23 22:04:02  agmsmith
 * More attribute tests.
 *
 * Revision 1.53  2002/05/23 19:01:31  agmsmith
 * Volume and attribute tests improved.
 *
 * Revision 1.52  2002/05/14 16:51:11  agmsmith
 * More attribute tests.
 *
 * Revision 1.51  2002/05/12 18:46:19  agmsmith
 * Mostly attribute tests.
 *
 * Revision 1.50  2002/04/29 21:14:08  agmsmith
 * Filesystem name change.
 *
 * Revision 1.49  2002/04/28 21:57:34  agmsmith
 * Added disk volume tests.
 *
 * Revision 1.48  2002/04/28 17:19:21  agmsmith
 * *** empty log message ***
 *
 * Revision 1.47  2002/04/28 17:10:54  agmsmith
 * Don't print output to stdout when flushing vnodes using ls -R.
 *
 * Revision 1.46  2002/04/26 16:12:31  agmsmith
 * Added saving a log of the tests to standard output.
 *
 * Revision 1.45  2002/04/21 21:04:52  agmsmith
 * Also check for attribute changed notification.
 *
 * Revision 1.44  2002/04/21 20:13:22  agmsmith
 * Yet more node monitoring tests.
 *
 * Revision 1.43  2002/04/15 20:44:18  agmsmith
 * Initial change notification test code in and working.
 *
 * Revision 1.42  2002/03/31 20:09:44  agmsmith
 * Added system for limiting the total file size, to avoid using
 * up too much memory and causing a kernel crash.
 *
 * Revision 1.41  2002/03/29 21:00:17  agmsmith
 * Added total file size option to mount test.
 *
 * Revision 1.40  2002/03/17 20:05:56  agmsmith
 * More AGMS link testing, including llink.
 *
 * Revision 1.39  2002/03/16 19:39:52  agmsmith
 * AGMS link testing now working.
 *
 * Revision 1.38  2002/03/13 22:17:32  agmsmith
 * Rename tests done, added mouse wheel support too.
 *
 * Revision 1.37  2002/03/10 16:18:56  agmsmith
 * Added more readlink tests, for NULL buffers etc.
 *
 * Revision 1.36  2002/03/01 19:48:13  agmsmith
 * Include file system exports.
 *
 * Revision 1.35  2002/02/23 20:29:17  agmsmith
 * Stop recursive delete when it encounters an error.
 *
 * Revision 1.34  2002/02/21 21:34:08  agmsmith
 * Improved flashing, and working on AGMS link tests.
 *
 * Revision 1.33  2002/02/21 21:09:53  agmsmith
 * Fix bugs in recursive delete and verify directory tree, add
 * tests for symbolic links.
 *
 * Revision 1.32  2002/02/19 00:24:49  agmsmith
 * VerifyDirectoryTree is mostly working now.
 *
 * Revision 1.31  2002/02/18 00:00:47  agmsmith
 * Under construction, adding AGMS link tests and sidetracked
 * to add a generic directory comparison function.
 *
 * Revision 1.30  2001/12/16 03:56:13  agmsmith
 * Added mount directory extra parameter for mount().
 *
 * Revision 1.29  2001/10/09 20:49:31  agmsmith
 * Added renaming tests.
 *
 * Revision 1.28  2001/10/06 15:56:12  agmsmith
 * Tuning for handling big files.
 *
 * Revision 1.27  2001/09/22 21:31:46  agmsmith
 * Added option to not do the gigabyte SetSize test.
 *
 * Revision 1.26  2001/09/16 01:24:40  agmsmith
 * Super big and sparse file tests in wstat test.
 *
 * Revision 1.25  2001/09/15 21:23:51  agmsmith
 * Added wstat tests.
 *
 * Revision 1.24  2001/09/15 18:00:28  agmsmith
 * Added tests for append mode and truncation when opening files.
 *
 * Revision 1.23  2001/09/15 16:24:48  agmsmith
 * Removed file closing test (now part of unmount), added
 * ability to stop the file read/write test in the middle.
 *
 * Revision 1.22  2001/09/14 21:00:20  agmsmith
 * Added reading and writing speed tests.
 *
 * Revision 1.21  2001/09/13 20:55:55  agmsmith
 * Let the user pick the fake mount directory.
 *
 * Revision 1.20  2001/09/12 22:37:43  agmsmith
 * Permissions tests ready, added umask test.
 *
 * Revision 1.19  2001/09/11 20:43:54  agmsmith
 * Added read-only flag mount option.
 *
 * Revision 1.18  2001/09/10 21:27:15  agmsmith
 * *** empty log message ***
 *
 * Revision 1.17  2001/09/10 18:44:18  agmsmith
 * Not much.
 *
 * Revision 1.16  2001/09/07 00:09:15  agmsmith
 * Added fake mounting for comparison with existing file systems,
 * also cleaned up directory tests.  File tests half done, permission
 * tests just started.
 *
 * Revision 1.15  2001/08/21 01:55:08  agmsmith
 * Adding tests for permissions and creating files.
 *
 * Revision 1.14  2001/07/28 19:10:16  agmsmith
 * Trailing slash in directory names causes problems in rmdir.
 *
 * Revision 1.13  2001/07/28 19:07:48  agmsmith
 * Added rmdir command and tests.
 *
 * Revision 1.12  2001/07/28 00:43:33  agmsmith
 * Added a bunch of directory making tests.
 *
 * Revision 1.11  2001/07/23 01:10:15  agmsmith
 * Changed name of file system to something which fits in the
 * 8 character limit.
 *
 * Revision 1.10  2001/07/12 17:44:31  agmsmith
 * Don't specify device for mounting on, there is none.
 *
 * Revision 1.9  2001/07/06 00:19:52  agmsmith
 * Added an unmount test.
 *
 * Revision 1.8  2001/07/02 21:59:57  agmsmith
 * Initial test sequence for just mounting.
 *
 * Revision 1.7  2001/07/02 17:12:09  agmsmith
 * Added stop at first failure option.
 *
 * Revision 1.6  2001/07/01 21:59:50  agmsmith
 * More testing code.
 *
 * Revision 1.5  2001/06/30 21:58:23  agmsmith
 * Busy adding scrolling subview and array subsubview.
 *
 * Revision 1.4  2001/06/30 15:40:00  agmsmith
 * Under construction.
 *
 * Revision 1.3  2001/06/05 01:02:12  agmsmith
 * Framework is starting to fall into place, class of Tests with
 * subclasses for specific tests, and it's a view too so you see
 * the test and parameters as a row in the display.
 *
 * Revision 1.2  2001/05/31 20:44:20  agmsmith
 * Decided to add a full test suite to the mounter, along with a
 * nice graphical interface to the different tests.  Application
 * coding started.
 *
 * Revision 1.1  2001/05/14 14:16:34  agmsmith
 * Initial revision
 */

/* BeOS headers. */

#include <Application.h>
#include <Alert.h>
#include <Button.h>
#include <CheckBox.h>
#include <Directory.h>
#include <Entry.h>
#include <File.h>
#include <fs_attr.h>
#include <fs_info.h>
#include <fs_index.h>
#include <Looper.h>
#include <MenuBar.h>
#include <MenuItem.h>
#include <NodeMonitor.h>
#include <Path.h>
#include <PopUpMenu.h>
#include <Query.h>
#include <RadioButton.h>
#include <ScrollView.h>
#include <StringView.h>
#include <TextControl.h>
#include <View.h>
#include <Volume.h>
#include <Window.h>

/* Posix headers. */

#include <errno.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

/* File system specific stuff. */

#define AGMSRAMFS_LLINK_INCLUDE_CODE 1
#define AGMSRAMFS_MAXTOTALFILESIZE_INCLUDE_CODE 1
#include "AGMSRAMFSExports.h"



/******************************************************************************
 * Global variables.
 */

static float g_MarginBetweenControls; /* Space of a letter "M" between them. */
static float g_LineOfTextHeight;      /* Height of text the current font. */
static float g_StringViewHeight;      /* Height of a string view text box. */
static float g_ButtonHeight;          /* How many pixels tall buttons are. */
static float g_CheckBoxHeight;        /* Same for check boxes. */
static float g_RadioButtonHeight;     /* Also for radio buttons. */
static float g_PopUpMenuHeight;       /* Again for pop-up menus. */
static float g_TextBoxHeight;         /* Ditto for editable text controls. */

static bool volatile g_StopTestRequested; /* Set when user wants to stop. */
static bool g_FakeMount = false;      /* False for real file system test. */
static char g_FakeDirString [PATH_MAX] = "/boot/var/tmp/Test";

typedef enum MessageCodesEnum
{
  MSG_BIG_BUF_SIZE = 'BiBS',
  MSG_BIG_BUF_TIMES = 'BiBX',
  MSG_FAKE_DIRECTORY = 'FkDr',
  MSG_READ_ONLY_MOUNT = 'RdOy',
  MSG_MAX_TOTAL_FILE_SIZE_MOUNT = 'MxTS',
  MSG_GIGA_SIZE_TEST = 'Gigs',
  MSG_RUN_TESTS = 'RunT',
  MSG_RUN_BUTTON = 'RunB',
  MSG_STOP_AT_FIRST_FAILURE = 'Stp1',
  MSG_TEST_ENABLE = 'Enbl',
} MessageCodes;



#if 0 /* Unused code. */
/******************************************************************************
 * Global utility function to convert a 32 bit number into a 4 character
 * string so that you can see printable message codes etc.
 */

static const char *U32toString (uint32 Number)
{
  union ConverterUnion
  {
    uint32 number;
    char string [4];
  } Converter;

  static char OutputBuffer [5];

  Converter.number = Number;

  OutputBuffer[0] = Converter.string[3];
  OutputBuffer[1] = Converter.string[2];
  OutputBuffer[2] = Converter.string[1];
  OutputBuffer[3] = Converter.string[0];
  OutputBuffer[4] = 0;

  return OutputBuffer;
}
#endif /* Unused code. */



/******************************************************************************
 * Global utility function to display an error message and return.  The message
 * part describes the error, and if ErrorNumber is non-zero, gets the string
 * ", error code $X (standard description)." appended to it.  If the message
 * is NULL then it gets defaulted to "Something went wrong".  The title part
 * doesn't get displayed (no title bar in the dialog box, but you can see it
 * in the debugger as the window thread name), and defaults to "Error Message"
 * if you didn't specify one.
 */

static void DisplayErrorMessage (
  const char *MessageString = NULL,
  int ErrorNumber = 0,
  const char *TitleString = NULL)
{
  BAlert *AlertPntr;
  char ErrorBuffer [B_PATH_NAME_LENGTH + 80 /* error message */ + 80];

  if (TitleString == NULL)
    TitleString = "Error Message";

  if (MessageString == NULL)
  {
    if (ErrorNumber == 0)
      MessageString = "No error, no message, why bother?";
    else
      MessageString = "Something went wrong";
  }

  if (ErrorNumber != 0)
  {
    sprintf (ErrorBuffer, "%s, error code $%X/%d (%s) has occured.",
      MessageString, ErrorNumber, ErrorNumber, strerror (ErrorNumber));
    MessageString = ErrorBuffer;
  }

  AlertPntr = new BAlert (TitleString, MessageString,
    "Acknowledge", NULL, NULL, B_WIDTH_AS_USUAL, B_STOP_ALERT);
  if (AlertPntr != NULL)
    AlertPntr->Go ();
}



/******************************************************************************
 * Delete the given file or directory.  If it is a directory, it will
 * recursively delete the contents.  Avoids BEntry etc, to avoid loads of file
 * opens and locks, which clutter up the log.  Returns 0 if successful, else
 * returns a negative error code and stops deleting things.
 */

static status_t RecursivelyDelete (const char *DeleteName)
{
  dirent_t   *DirEntPntr;
  DIR        *DirHandle;
  status_t    ErrorCode;
  const char *RecursivelyDeleteError = "RecursivelyDelete Error";
  struct stat StatStruct;
  const char *StringPntr;
  char        TempString [B_PATH_NAME_LENGTH + 80];
  BPath       ThingPath;

  ThingPath.SetTo (DeleteName);
  if (ThingPath.InitCheck() != 0)
  {
    sprintf (TempString, "Problems getting full path for thing "
      "named \"%s\"", DeleteName);
    DisplayErrorMessage (TempString,
      ThingPath.InitCheck(), RecursivelyDeleteError);
    return ThingPath.InitCheck();
  }

  /* See what the thing is.  Use lstat, which does not traverse symbolic links,
  so that we can delete the link itself. */

  if (lstat (ThingPath.Path(), &StatStruct))
  {
    ErrorCode = errno;
    sprintf (TempString, "Unable to get status of thing named \"%s\"",
      ThingPath.Path());
    DisplayErrorMessage (TempString, ErrorCode, RecursivelyDeleteError);
    return ErrorCode;
  }

  if (S_ISREG (StatStruct.st_mode))
  {
    // A file.
    if (unlink (ThingPath.Path()))
    {
      ErrorCode = errno;
      sprintf (TempString, "Unable to delete file \"%s\"",
        ThingPath.Path());
      DisplayErrorMessage (TempString, ErrorCode, RecursivelyDeleteError);
      return ErrorCode;
    }
  }
  else if (S_ISDIR (StatStruct.st_mode))
  {
    // A directory.  Go through the contents and delete them.

    DirHandle = opendir (ThingPath.Path());
    if (DirHandle == NULL)
    {
      ErrorCode = errno;
      sprintf (TempString, "Problems opening directory "
        "named \"%s\".", ThingPath.Path());
      DisplayErrorMessage (TempString, ErrorCode, RecursivelyDeleteError);
      return ErrorCode;
    }

    ErrorCode = 0;
    while (ErrorCode == 0 && (DirEntPntr = readdir (DirHandle)) != NULL)
    {
      for (StringPntr = DirEntPntr->d_name; *StringPntr != 0; StringPntr++)
        if (*StringPntr != '.')
          break;
      if (*StringPntr == 0)
        continue; /* Skip all period names, like . or .. or ... etc. */

      strcpy (TempString, ThingPath.Path());
      strcat (TempString, "/");
      strcat (TempString, DirEntPntr->d_name);
      ErrorCode = RecursivelyDelete (TempString);
    }
    closedir (DirHandle);
    if (ErrorCode != 0)
      return ErrorCode;

    // Delete the now empty directory.
    if (rmdir (ThingPath.Path()))
    {
      ErrorCode = errno;
      sprintf (TempString, "Unable to delete directory \"%s\"",
        ThingPath.Path());
      DisplayErrorMessage (TempString, ErrorCode, RecursivelyDeleteError);
      return ErrorCode;
    }
  }
  else if (S_ISLNK (StatStruct.st_mode))
  {
    // Symbolic link.
    if (unlink (ThingPath.Path()))
    {
      ErrorCode = errno;
      sprintf (TempString, "Unable to delete symbolic link \"%s\"",
        ThingPath.Path());
      DisplayErrorMessage (TempString, ErrorCode, RecursivelyDeleteError);
      return ErrorCode;
    }
  }
  else
  {
    // Unknown new type of file system object (attribute index, block device).
    sprintf (TempString, "Unknown type of file system object for %s, "
      "st_mode flags are 0%o.", ThingPath.Path(),  StatStruct.st_mode);
    DisplayErrorMessage (TempString, 0, RecursivelyDeleteError);
    return -1;
  }

  return 0;
}



/******************************************************************************
 * This is a subclass used only by TestAbstract::VerifyDirectoryTree.  It
 * stores the information about each match test done when checking a directory
 * tree.
 */

struct MatchInfo
{
  char     code;
  bool     doVerification;
  bool     fileFound;
  off_t    length;
  char    *name;
  char    *symbolicName;

  MatchInfo ();
  ~MatchInfo ();
  const char *SetFromMatchLine (const char *MatchLine, const char *PathPrefix);

private: /* Disable default copy constructor. */
  MatchInfo (const MatchInfo&);
  MatchInfo& operator= (const MatchInfo&);
};

MatchInfo::MatchInfo ()
 : code ('?'),
   doVerification (false),
   fileFound (false),
   length (0),
   name (NULL),
   symbolicName (NULL)
{
}

MatchInfo::~MatchInfo ()
{
  delete [] name;
  delete [] symbolicName;
}

/* Parse the "My File,f,42\n" line into the match components.  Returns NULL if
it fails, otherwise returns a pointer to just after the matched text (so you
can continue matching). */

const char * MatchInfo::SetFromMatchLine (
  const char *MatchLine,
  const char *PathPrefix)
{
  int         PrefixLength;
  const char *StartPntr = MatchLine;
  int         StringLength;
  const char *StringPntr;

  /* Parse the first argument, the file name. */

  StringPntr = StartPntr;
  while (*StringPntr != ',' && *StringPntr != 0 && *StringPntr != '\n')
    StringPntr++;
  if (*StringPntr != ',')
  {
    DisplayErrorMessage ("Match line doesn't have any commas.");
    return NULL;
  }
  StringLength = StringPntr - StartPntr;
  PrefixLength = strlen (PathPrefix);
  if (StringLength == 0)
  {
    name = new char [PrefixLength + 1];
    if (name == NULL) return NULL;
    strcpy (name, PathPrefix);
  }
  else /* Appending an actual name to the root path. */
  {
    name = new char [StringLength + PrefixLength + 2];
    if (name == NULL) return NULL;
    strcpy (name, PathPrefix);
    name[PrefixLength] = '/';
    strncpy (name + PrefixLength + 1, StartPntr, StringLength);
    name [PrefixLength + StringLength + 1] = 0;
  }
  StartPntr = StringPntr + 1;

  /* Parse the second argument, the single character code. */

  StringPntr = StartPntr;
  while (*StringPntr != ',' && *StringPntr != 0 && *StringPntr != '\n')
    StringPntr++;
  if (*StringPntr != ',')
  {
    DisplayErrorMessage ("Match line doesn't have enough commas.");
    return NULL;
  }
  StringLength = StringPntr - StartPntr;
  if (StringLength != 1)
  {
    DisplayErrorMessage ("Match line code should just be a single letter.");
    return NULL;
  }
  code = *StartPntr;
  StartPntr = StringPntr + 1;

  /* Parse the third argument, the thing to validate against, or "*" for no
  validation needed. */

  StringPntr = StartPntr;
  while (*StringPntr != 0 && *StringPntr != '\n')
    StringPntr++;
  if (*StringPntr != '\n')
  {
    DisplayErrorMessage ("Match line doesn't end with a new line.");
    return NULL;
  }
  StringLength = StringPntr - StartPntr;

  if (code == 'd' /* Directory */ || (StringLength == 1 && *StartPntr == '*'))
    doVerification = false;
  else
  {
    doVerification = true;

    if (code == 'f')
      length = atoll (StartPntr);
    else /* Some kind of symbolic-like link. */
    {
      symbolicName = new char [StringLength + 1];
      if (symbolicName == NULL) return NULL;
      strncpy (symbolicName, StartPntr, StringLength);
      symbolicName [StringLength] = 0;
    }
  }
  StartPntr = StringPntr + 1;

  return StartPntr;
}



/******************************************************************************
 * This NotificationLooper is used for monitoring notification messages.  It
 * saves the info from each notification into a preset array, one element for
 * each notification message opcode (doesn't handle multiple instances of an
 * opcode).  The main application thread does its file operations and checks
 * this BLooper to see what messages it has received.  After allocating a new
 * instance dynamically (doesn't work as a static allocation), use the Run()
 * method to start it up.  When you are done, send it the B_QUIT_REQUESTED (use
 * the PostMessage method).
 */

class NotificationLooper : public BLooper
{
public:
  NotificationLooper (const char *LooperNameString = "NotificationTestLooper");
  virtual ~NotificationLooper ();

  virtual void MessageReceived (BMessage *MessagePntr);

  void       Clear ();
  BMessage * GetMessageForOpCode (int OpCode);
  bool       HasOverflowed ();
  void       PrintAllMessages ();
  int        WaitForOpCodes (int NumberOfOpCodes);

protected:
  static const int MAX_OPCODE = B_DEVICE_UNMOUNTED + 1;
  BMessage *m_MessageArray[MAX_OPCODE];
  bool m_HasOverflowed;
};


NotificationLooper::NotificationLooper (const char *LooperNameString)
: BLooper (LooperNameString),
  m_HasOverflowed (false)
{
  int i;
  for (i = 0; i < MAX_OPCODE; i++)
    m_MessageArray [i] = NULL;
}


NotificationLooper::~NotificationLooper ()
{
  Clear ();
}


void NotificationLooper::MessageReceived (BMessage *MessagePntr)
{
  int32 OpCode;

  switch (MessagePntr->what)
  {
    case B_NODE_MONITOR: /* Save a copy of the notification in our array. */
      /* MessagePntr->PrintToStream (); */
      if (MessagePntr->FindInt32 ("opcode", &OpCode) == B_OK)
      {
        if (OpCode >= 0 && OpCode < MAX_OPCODE)
        {
          if (m_MessageArray [OpCode] != NULL)
            m_HasOverflowed = true;
          else
            m_MessageArray [OpCode] = new BMessage (*MessagePntr);
        }
        else
          DisplayErrorMessage ("NotificationLooper: "
            "Got an unknown change notification opcode!");
      }
      break;

    default:
      BLooper::MessageReceived (MessagePntr);
  }
}


void NotificationLooper::Clear ()
{
  int i;

  for (i = 0; i < MAX_OPCODE; i++)
  {
    if (m_MessageArray [i] != NULL)
    {
      delete m_MessageArray [i];
      m_MessageArray [i] = NULL;
    }
  }
  m_HasOverflowed = false;
}


BMessage * NotificationLooper::GetMessageForOpCode (int OpCode)
{
  if (OpCode < 0 || OpCode >= MAX_OPCODE)
    return NULL;
  return m_MessageArray [OpCode];
}


bool NotificationLooper::HasOverflowed ()
{
  return m_HasOverflowed;
}


void NotificationLooper::PrintAllMessages ()
{
  int i;
  for (i = 0; i < MAX_OPCODE; i++)
  {
    if (m_MessageArray [i] != NULL)
    {
      printf ("Message for opcode %d:\n", i);
      m_MessageArray[i]->PrintToStream ();
    }
  }
}


/* Wait for NumberOfOpCodes different kinds of opcodes (duplicate opcodes just
count as 1) to have been received, or 5 seconds to pass (since the debug output
generated by the file system code can take a while to display at 19200 baud).
Returns the actual number of distinct opcodes received, which will be less than
NumberOfOpCodes if it ran out of time.  Returns -1 if the user aborts the test
run.  Presumably called from a thread different than the BLooper of this
object. */

int NotificationLooper::WaitForOpCodes (int NumberOfOpCodes)
{
  int       Count;
  int       i;
  int       j;
  const int TimeoutSeconds = 5;

  for (i = 0; i < 20 * TimeoutSeconds; i++)
  {
    snooze (50000 /* 1/20 of a second */);
    for (Count = 0, j = 0; j < MAX_OPCODE; j++)
      if (m_MessageArray [j] != NULL)
        Count++;
    if (Count >= NumberOfOpCodes)
      return Count;
    if (g_StopTestRequested)
      return -1;
  }
  return Count;
}



/******************************************************************************
 * This QueryLooper is used for monitoring query messages.  It's similar to
 * NotificationLooper.
 */

class QueryLooper : public NotificationLooper
{
public:
  QueryLooper ();

  virtual void MessageReceived (BMessage *MessagePntr);
};


QueryLooper::QueryLooper ()
: NotificationLooper ("QueryTestLooper")
{
}


void QueryLooper::MessageReceived (BMessage *MessagePntr)
{
  int32 OpCode;

  switch (MessagePntr->what)
  {
    case B_QUERY_UPDATE: /* Save a copy of the query result in our array. */
      if (MessagePntr->FindInt32 ("opcode", &OpCode) == B_OK)
      {
        if (OpCode >= 0 && OpCode < MAX_OPCODE)
        {
          if (m_MessageArray [OpCode] != NULL)
            m_HasOverflowed = true;
          else
            m_MessageArray [OpCode] = new BMessage (*MessagePntr);
        }
        else
          DisplayErrorMessage ("QueryLooper: "
            "Got an unknown query change opcode!");
      }
      break;

    default:
      BLooper::MessageReceived (MessagePntr);
  }
}



/******************************************************************************
 * Each possible test is represented by a subclass of this abstract class.  The
 * subclasses follow this one in the source code, in alphabetical order.
 *
 * The test suite has an array of instances of assorted subclasses (actually
 * stored as the children of the display view), as selected up by the user.
 * They all have in common a call to do the test, a description, and internal
 * parameters.  The user interface portion (they inherit from BView and
 * implement the user interface for a row in the test table) will let the user
 * select which tests to do (by enabling or disabling them in the complete
 * list of all tests), and specify parameters.  Then when the "GO" button is
 * pressed, the program will do the selected tests in the predefined order (so
 * initialization related tests come before shutdown tests).
 */

class TestAbstract : public BView
{
public:
  TestAbstract (BRect NewFrame, const char *NewName);
  virtual ~TestAbstract ();
    /* Constructor and destructor for this virtual base class.  Subclass
    constructors should update m_GUIHeight as needed. */


/* Other Public Methods, in alphabetical order. */

  virtual const char *GetDescription () = 0;
    /* Return a short one line description of the test, it will be displayed
    by the base class to identify the test. */

  virtual void GetPreferredSize (float *WidthPntr, float *HeightPntr);
  bool GetTestWasSuccessful ();
  virtual void MessageReceived (BMessage *MessagePntr);
  virtual void PreTestPreparations ();
  virtual void Pulse ();
  void SetDefaultViewColour ();
  virtual bool SetTestEnabled (bool Enable);
  bool TestAndFrills ();


protected:
  virtual float AddCustomControls (float X);
    /* Subclasses should implement this to add custom controls. */

  virtual bool DoTest () = 0;
    /* This function is called by the application thread to do the test.  It
    should return TRUE if further tests can be done (successful, or failed in
    a way that would allow further tests to work), return FALSE to skip the
    remaining tests.  If there are any errors, it should use SetTestResult to
    show them to the user.  Note that it is called by a non-window thread, so
    you need to use LockLooper when doing any window operations. */

  void SetTestResult (const char *ResultText, bool Successful = true);
  void SetTestResult (const char *ResultText, int ErrorCode);
  bool VerifyDirectoryTree (const char *Root, const char *MatchStrings);

  float m_GUIHeight;
    /* Height of the user interface in pixels.  The constructors set this to
    the maximum of the height of all the contained controls.  Then when the
    controls are created, they are vertically centered in this space. */

  bool m_TestEnabled;
    /* Set to TRUE if this test is enabled (will be done).  Set to FALSE if it
    is disabled (skip the test).  The controls will be drawn in light gray
    when it is disabled. */


private:
  TestAbstract (const TestAbstract&);
  TestAbstract& operator= (const TestAbstract&);
    /* Disable the default copy and assignment operators by privatizing them
    and not providing an implementation. */

  virtual void AttachedToWindow ();

  bool VerifyRecursively (const char *TestName, int NumberOfMatchers,
    MatchInfo *MatchInfoArray);

  BStringView *m_DescriptionViewPntr;
    /* Points to the string control (well, view) used for displaying the
    description.  NULL if not allocated yet. */

  BCheckBox *m_EnableCheckBoxPntr;
    /* Points to the check box control used for displaying the enabled /
    disabled marker.  NULL if not allocated yet. */

  bool m_FlashStateIsOn;
    /* TRUE if the display is currently showing highlighted error text, FALSE
    if it is showing readable error text.  This is useful for avoiding flicker
    during redrawing when the error message is already in the right colours. */

  bigtime_t m_LastFlashTime;
    /* This is used by the flashing code to keep track of the brightness
    of the flashing text over time. */

  BStringView *m_ResultStringViewPntr;
    /* Points to the subview which displays the result string, or NULL. */

  bool m_TestWasSuccessful;
    /* TRUE if the last test was succesful, FALSE if it had failed (and thus
    the test result should be displayed in red and flashing). */
};


/* Construct the new test object, but don't do much here since initialisation
depends on the virtual subclasses being operational, which they aren't until
after this base constructor has run.  The real construction work is in
AttachedToWindow and the subclasses's implementation of AddCustomControls.
Subclass constructors should update m_GUIHeight to fit their needs. */

TestAbstract::TestAbstract (BRect NewFrame, const char *NewName)
: BView (NewFrame, NewName, B_FOLLOW_LEFT | B_FOLLOW_TOP, B_PULSE_NEEDED),
  m_GUIHeight (g_StringViewHeight),
  m_TestEnabled (false),
  m_DescriptionViewPntr (NULL),
  m_EnableCheckBoxPntr (NULL),
  m_FlashStateIsOn (true),
  m_LastFlashTime (0),
  m_ResultStringViewPntr (NULL),
  m_TestWasSuccessful (true)
{
  /* Set the initial maximum height to the space needed for the default
  checkbox and name display.  Subclass constructors run after this and will
  update it as needed. */

  if (m_GUIHeight < g_CheckBoxHeight)
    m_GUIHeight = g_CheckBoxHeight;
}


TestAbstract::~TestAbstract ()
{
  /* Child views (controls) are automatically destroyed and deallocated by
  the parent destructor.  So there isn't much to do here. */

  m_DescriptionViewPntr = NULL;
  m_EnableCheckBoxPntr = NULL;
  m_ResultStringViewPntr = NULL;
}


/* Add the user interface buttons and text boxes needed just for this test.
X specifies the left/right offset in the view to start adding controls.
The default implementation adds nothing (useful for tests with no user
settable parameters).  Return -1.0 on failure, else returns new X value. */

float TestAbstract::AddCustomControls (float X)
{
  return X;
}


/* The view has been added to a window.  If this is the first time, create
all the subviews for the user interface controls.  Can't do that in the
constructor since the subclass constructors haven't been called yet and thus
the virtual functions aren't working yet.  My protocol is to create subviews
after the window has been attached to the GUI server, so that you can size
things according to the current font settings).  So, Test subclasses should
create their subviews in AddCustomControls (which gets called by this) instead
of here. */

void TestAbstract::AttachedToWindow ()
{
  float Height = 0;
  float Margin;
  BRect TempRect;
  float Width = 0;
  float X;

  if (m_EnableCheckBoxPntr != NULL)
    return; /* Already initialised. */

  /* Start with the enable/disable checkbox, width around 60 pixels. */

  Margin = ceilf ((m_GUIHeight - g_CheckBoxHeight) / 2);
  TempRect = Bounds ();
  X = TempRect.left;
  TempRect.top += Margin;
  TempRect.bottom = TempRect.top + g_CheckBoxHeight;

  m_EnableCheckBoxPntr = new BCheckBox (TempRect,
    "EnableBox",
    "Enable",
    new BMessage (MSG_TEST_ENABLE));
  if (m_EnableCheckBoxPntr == NULL) goto ErrorExit;
  AddChild (m_EnableCheckBoxPntr);
  m_EnableCheckBoxPntr->SetTarget (this);
  m_EnableCheckBoxPntr->SetValue (m_TestEnabled);
  m_EnableCheckBoxPntr->ResizeToPreferred ();
  m_EnableCheckBoxPntr->GetPreferredSize (&Width, &Height);
  X += ceilf (Width + g_MarginBetweenControls);

  /* Add the description box, width 300 pixels. */

  Margin = ceilf ((m_GUIHeight - g_StringViewHeight) / 2);
  TempRect = Bounds ();
  TempRect.left = X;
  TempRect.right = TempRect.left + 300;
  TempRect.top += Margin;
  TempRect.bottom = TempRect.top + g_StringViewHeight;

  m_DescriptionViewPntr = new BStringView (TempRect,
    "Description", GetDescription ());
  if (m_DescriptionViewPntr == NULL) goto ErrorExit;
  AddChild (m_DescriptionViewPntr);
  X += ceilf (TempRect.right - TempRect.left + g_MarginBetweenControls);

  /* Add subclass controls, use at least 200 pixels, or more if necessary. */

  TempRect.left = X;
  X = AddCustomControls (X);
  if (X == -1.0) goto ErrorExit;
  if (X < TempRect.left + 200)
    X = TempRect.left + 200;
  X += g_MarginBetweenControls;

  /* Add the test result string subview, wide enough to bring the total width
  to 1500 (see GetPreferredSize for total row width allocated in the view), or
  at least 200 pixels if there isn't enough space. */

  Margin = ceilf ((m_GUIHeight - g_StringViewHeight) / 2);
  TempRect = Bounds ();
  TempRect.left = X;
  TempRect.right = 1500;
  if (TempRect.right < TempRect.left + 200)
    TempRect.right = TempRect.left + 200; /* Smallest size is 200 pixels. */
  TempRect.top += Margin;
  TempRect.bottom = TempRect.top + g_StringViewHeight;

  m_ResultStringViewPntr = new BStringView (TempRect,
    "ResultText", " ");
  if (m_ResultStringViewPntr == NULL) goto ErrorExit;
  AddChild (m_ResultStringViewPntr);
  X += ceilf (TempRect.right - TempRect.left + g_MarginBetweenControls);

  /* Shrink the test view to fit the height used, and use standard width. */

  ResizeToPreferred ();
  SetDefaultViewColour ();
  return;

ErrorExit:
  DisplayErrorMessage ("Unable to initialise a test's UI controls.");
}


/* Returns the size the test view needs, basically enough to hold all the
controls in height, and the standard common row width.  Used by parent views
to adjust scrolling area size. */

void TestAbstract::GetPreferredSize (float *WidthPntr, float *HeightPntr)
{
  if (WidthPntr != NULL)
    *WidthPntr = 1500;
  if (HeightPntr != NULL)
    *HeightPntr = m_GUIHeight;
}


bool TestAbstract::GetTestWasSuccessful ()
{
  return m_TestWasSuccessful;
}


/* A message has been received, usually from a subcontrol like a button or
checkbox.  If you don't understand it, pass it on to the parent class's
MessageReceived method. */

void TestAbstract::MessageReceived (BMessage *MessagePntr)
{
  switch (MessagePntr->what)
  {
    case MSG_TEST_ENABLE:
      SetTestEnabled (!m_TestEnabled);
      break;

    default:
      BView::MessageReceived (MessagePntr);
      break;
  }
}


/* Do various pre-test things, this is called (with the window locked) for all
tests before the actual testing starts.  Subclasses can also do other pre-test
things here, such as reading the current text from text boxes, validating it,
and writing it back (otherwise the user's changes to the box with the current
focus isn't used).  This common abstract method clears out the display prior to
running a test - so that it shows that no test has been done, in case the
testing stops before it reaches this one. */

void TestAbstract::PreTestPreparations ()
{
  SetTestResult ("");
}


/* Periodically called by the OS (maybe twice a second), used for the timing
of flashing error messages in the result string display. */

void TestAbstract::Pulse ()
{
  bigtime_t CurrentTime;
  bigtime_t ElapsedTime;

  if (m_TestWasSuccessful || m_ResultStringViewPntr == NULL)
    return; /* No flashing for successful tests or not ready yet. */

  CurrentTime = system_time ();
  ElapsedTime = CurrentTime - m_LastFlashTime;

  /* Flash brightness varies over approximately 2 seconds. */

  if (ElapsedTime < 1000000)
  {
    if (m_FlashStateIsOn)
    {
      m_ResultStringViewPntr->SetHighColor (255, 128, 128);
      m_ResultStringViewPntr->SetLowColor (128, 0, 0);
      m_ResultStringViewPntr->SetViewColor (128, 0, 0);
      m_FlashStateIsOn = false;
      m_ResultStringViewPntr->Invalidate ();
    }
  }
  else
  {
    if (!m_FlashStateIsOn)
    {
      m_ResultStringViewPntr->SetHighColor (0, 0, 0);
      m_ResultStringViewPntr->SetLowColor (255, 220, 220);
      m_ResultStringViewPntr->SetViewColor (255, 220, 220);
      m_FlashStateIsOn = true;
      m_ResultStringViewPntr->Invalidate ();
    }

    if (ElapsedTime > 2100000)
      m_LastFlashTime = CurrentTime;
  }
}


void TestAbstract::SetDefaultViewColour ()
{
  SetViewColor (255, 255, 255);
/*
  SetViewColor (
    (unsigned char) this,
    (unsigned char) (((int) this) >> 2),
    (unsigned char) (((int) this) >> 4));
*/
}


/* Enables (TRUE) or disables (FALSE) the test.  Returns the new enable state
and updates displayed controls to show the new status.  Subclasses can use
this to hide their controls when disabled.  The final thing that subclasses
should do is to call the parent class's version of this function (which will
update the enable checkbox and set m_TestEnabled to the new value) and should
return the value returned by the parent function.  The default implementation
just updates the enabled checkbox and m_TestEnabled. */

bool TestAbstract::SetTestEnabled (bool Enable)
{
  bool NewEnableState = (Enable != 0);

  /* If the controls exist, update the enable checkbox. */

  if (m_EnableCheckBoxPntr != NULL)
    m_EnableCheckBoxPntr->SetValue (NewEnableState);

  m_TestEnabled = NewEnableState;
  return m_TestEnabled;
}


/* Changes the result text to be the given string (gets copied).  Also prints
it to standard output in case you want to log things.  If Successful is TRUE
then it is drawn in green, FALSE draws it in red.  Safe to call from non-window
threads (it locks the window when it changes the message). */

void TestAbstract::SetTestResult (const char *ResultText, bool Successful)
{
  m_TestWasSuccessful = (Successful != 0);

  if (m_ResultStringViewPntr == NULL || ResultText == NULL)
    return;

  if (LockLooper ()) /* Lock the window so other threads can call this. */
  {
    if (Successful)
    {
      m_ResultStringViewPntr->SetHighColor (0, 128, 0);
      m_ResultStringViewPntr->SetLowColor (255, 255, 255);
      m_ResultStringViewPntr->SetViewColor (255, 255, 255);
    }

    m_ResultStringViewPntr->SetText (ResultText);

    /* Also log the message to standard output (stdout) if it isn't one of the
    meaningless messages used to make things pretty on screen. */

    if (ResultText[0] && strcmp (ResultText, "-") &&
    strcmp (ResultText, "Testing..."))
      printf (" %s: %s\n", Successful ? "OK" : "--", ResultText);

    UnlockLooper ();
  }
  if (!Successful)
    snooze (2000000); /* Let transient errors show for a second or two. */
}


/* This one is used for error results, by specifying an integer error code
as a second argument rather than a boolean yes/no flag.  This result will
always fail the test.  Also, it appends ", error description." to the
given string. */

void TestAbstract::SetTestResult (const char *ResultText, int ErrorCode)
{
  char TempString [1024];

  sprintf (TempString, "%s, %d:%s.",
    ResultText, ErrorCode, strerror (ErrorCode));
  SetTestResult (TempString, false);
}


/* Does the test with graphical user interface frills.  Clears the test result
and highlights the test display to show that the test is in progress, calls
DoTest, then resets the highlighting.  Note that it is called by a non-window
thread (the application thread) so the window needs to be locked when anything
graphical is changed.  Also prints the test name to standard output for logging
purposes.  Returns TRUE if further tests can be run, FALSE to stop the testing.
Note that failure/success of tests is separate from this. */

bool TestAbstract::TestAndFrills ()
{
  bool ReturnCode;

  if (!m_TestEnabled)
  {
    SetTestResult ("-");
    return true;
  }

  /* Show that the test is in progress by changing the background colour of
  the test row to blue. */

  if (!LockLooper ())
    return false; /* Window no longer exists, closed by user etc. */
  SetViewColor (128, 128, 255);
  Invalidate ();
  printf ("%s\n", GetDescription ()); /* Print test name to stdout for log. */
  UnlockLooper ();

  SetTestResult ("Testing...");

  ReturnCode = DoTest ();

  if (LockLooper ())
  {
    SetDefaultViewColour ();
    Invalidate ();
    UnlockLooper ();
  }

  return ReturnCode;
}



/* The TestAbstract::VerifyDirectoryTree function verifies that a directory
tree matches a provided list of files.  The search is done in directory Root
(an absolute path, no trailing slash, "/Test" for example) and all its
subdirectories.  Each file there (except . and ..) must also have a matching
entry in MatchStrings and each entry must have a corresponding file.
MatchStrings is a newline terminated list of strings (a newline after each
entry, including after the last one).  Each string consists of a file name
relative to the root directory, followed by a comma and the file type letter
(d=directory, f=file, s=symbolic link, p=auxiliary parent directory symbolic
link (like "...."), a=AGMS link), followed by a verification value which can be
an asterix (*) for don't care, or size for files (a decimal number), path
string for links.  An example match line would be "SomeDir/MyFile,f,42\n".  For
actual use, the strings are parsed and converted into MatchInfo instances.
Returns TRUE if it's OK, uses SetTestResult with an appropriate error message
and returns FALSE if the directory tree doesn't match the specification. */

bool TestAbstract::VerifyDirectoryTree (
  const char *Root,
  const char *MatchStrings)
{
  int         i;
  int         NumberOfMatchers;
  bool        ReturnCode = false;
  const char *StringPntr;
  char        TempString [B_PATH_NAME_LENGTH + 80];

  /* Count the number of individual match strings inside the big string. */

  NumberOfMatchers = 0;
  StringPntr = MatchStrings;
  while (*StringPntr)
    if (*StringPntr++ == '\n')
      NumberOfMatchers++;
  if (NumberOfMatchers <= 0)
    return false;

  /* Make the MatchInfo array to store the info for each match test, and parse
  the caller's match line into individual tests. */

  MatchInfo * MatchInfoArray = new MatchInfo [NumberOfMatchers];
  if (MatchInfoArray == NULL) return false;

  for (i = 0, StringPntr = MatchStrings;
  i < NumberOfMatchers && StringPntr != NULL; i++)
    StringPntr = MatchInfoArray[i].SetFromMatchLine (StringPntr, Root);
  if (StringPntr == NULL || *StringPntr != 0 /* If excess stuff ignored */)
  {
    SetTestResult ("VerifyDirectoryTree: problems during preparation.",
      false);
    goto ErrorExit;
  }

  /* Recursively iterate through the root directory, checking all files etc. */

  if (!VerifyRecursively (Root, NumberOfMatchers, MatchInfoArray))
    goto ErrorExit;

  /* Check that all the matches have been marked off. */

  for (i = 0; i < NumberOfMatchers; i++)
  {
    if (!MatchInfoArray[i].fileFound)
    {
      sprintf (TempString, "File \"%s\" is missing, it should be there!",
        MatchInfoArray[i].name);
      SetTestResult (TempString, false);
      goto ErrorExit;
    }
  }

  ReturnCode = true; /* Success! */

ErrorExit:
  delete [] MatchInfoArray;
  return ReturnCode;
}


/* Recursive function to test one item and any subitems it has.  Returns TRUE
if successful, FALSE if it couldn't find a match.  Also checks off the items it
has matched from the MatchInfoArray. */

bool TestAbstract::VerifyRecursively (
  const char *TestName,
  int NumberOfMatchers,
  MatchInfo *MatchInfoArray)
{
  dirent_t   *DirEntPntr;
  DIR        *DirHandle;
  int         ErrorCode;
  int         i;
  MatchInfo  *MatchPntr;
  struct stat StatStruct;
  char        SymbolicLinkString [B_PATH_NAME_LENGTH + 1];
  char        TempString [B_PATH_NAME_LENGTH + 80];

  /* Look for the name in the list of things to match. */

  for (i = NumberOfMatchers, MatchPntr = MatchInfoArray;
  i > 0;
  i--, MatchPntr++)
  {
    if (strcmp (MatchPntr->name, TestName) == 0)
      break; /* Found the name in the match list. */
  }
  if (i <= 0)
  {
    sprintf (TempString, "Didn't find \"%s\" in the match list.",
      TestName);
    SetTestResult (TempString, false);
    return false;
  }
  MatchPntr->fileFound = true;

  if (lstat (TestName, &StatStruct))
  {
    ErrorCode = errno;
    sprintf (TempString, "Unable to get status of thing named \"%s\"",
      TestName);
    SetTestResult (TempString, ErrorCode);
    return false;
  }

  if (S_ISREG (StatStruct.st_mode))
  { // A file.
    if (MatchPntr->code != 'f')
    {
      sprintf (TempString, "Code %c: did not expect \"%s\" to be a file.",
        MatchPntr->code, TestName);
      SetTestResult (TempString, false);
      return false;
    }

    if (MatchPntr->doVerification && StatStruct.st_size != MatchPntr->length)
    {
      sprintf (TempString, "Length of file \"%s\" is %qd, was expecting %qd.",
        TestName, StatStruct.st_size, MatchPntr->length);
      SetTestResult (TempString, false);
      return false;
    }
    return true;
  }

  if (S_ISDIR (StatStruct.st_mode))
  { // A directory.
    if (MatchPntr->code != 'd')
    {
      sprintf (TempString, "Code %c: did not expect \"%s\" to be a directory.",
        MatchPntr->code, TestName);
      SetTestResult (TempString, false);
      return false;
    }

    /* Go through the contents and check them. */

    DirHandle = opendir (TestName);
    if (DirHandle == NULL)
    {
      ErrorCode = errno;
      sprintf (TempString, "Unable to open directory \"%s\"",
        TestName);
      SetTestResult (TempString, ErrorCode);
      return false;
    }

    while ((DirEntPntr = readdir (DirHandle)) != NULL)
    {
      if (strcmp (DirEntPntr->d_name, ".") == 0 ||
      strcmp (DirEntPntr->d_name, "..") == 0)
        continue; /* Skip the standard parent names "." and "..". */

      strcpy (TempString, TestName);
      strcat (TempString, "/");
      strcat (TempString, DirEntPntr->d_name);

      if (!VerifyRecursively (TempString, NumberOfMatchers, MatchInfoArray))
      {
        closedir (DirHandle);
        return false;
      }
    }
    closedir (DirHandle);
    return true;
  }


  if (S_ISLNK (StatStruct.st_mode))
  { // Symbolic link.  Also other links currently appear as symbolic.
    // But as a kludge, size is 0 = real symbolic link, 1 = AGMS, 2 = Parent.
    if (StatStruct.st_size == 0)
    {
      if (MatchPntr->code != 's')
      {
        sprintf (TempString, "Code %c: did not expect \"%s\" to be a "
          "regular symbolic link.", MatchPntr->code, TestName);
        SetTestResult (TempString, false);
        return false;
      }
    }
    else if (StatStruct.st_size == 1)
    {
      if (MatchPntr->code != 'a')
      {
        sprintf (TempString, "Code %c: did not expect \"%s\" to be an "
          "AGMS link.", MatchPntr->code, TestName);
        SetTestResult (TempString, false);
        return false;
      }
    }
    else if (StatStruct.st_size == 2)
    {
      if (MatchPntr->code != 'p')
      {
        sprintf (TempString, "Code %c: did not expect \"%s\" to be a "
          "fake symbolic auxiliary parent link.", MatchPntr->code, TestName);
        SetTestResult (TempString, false);
        return false;
      }
    }
    else
    {
      sprintf (TempString, "Size of symbolic link \"%s\" is %qd, "
        "not the expected 0, 1 or 2 secret codes.",
        TestName, StatStruct.st_size);
      SetTestResult (TempString, false);
      return false;
    }

    /* Verify that the symbolic/agms/parent link value is correct. */

    if (MatchPntr->doVerification)
    {
      memset (SymbolicLinkString, 0, sizeof (SymbolicLinkString));
      if (readlink (TestName, SymbolicLinkString,
      sizeof (SymbolicLinkString) - 1) < 0)
      {
        ErrorCode = errno;
        sprintf (TempString, "Unable to read link value of \"%s\"",
          TestName);
        SetTestResult (TempString, ErrorCode);
        return false;
      }

      if (strcmp (SymbolicLinkString, MatchPntr->symbolicName) != 0)
      {
        sprintf (TempString, "Symbolic part of \"%s\" is \"%s\", was "
          "expecting \"%s\".", TestName, SymbolicLinkString,
          MatchPntr->symbolicName);
        SetTestResult (TempString, false);
        return false;
      }
    }
    return true;
  }

  // Unknown new type of file system object (attribute index, block device).
  sprintf (TempString, "Unknown type of file system object for %s, "
    "st_mode flags are 0%o.", TestName,  StatStruct.st_mode);
  SetTestResult (TempString, false);

  return false;
}



/******************************************************************************
 * Test AGMS link operations.
 */

class TestAGMSLink : public TestAbstract
{
public:
  TestAGMSLink (BRect NewFrame)
    : TestAbstract (NewFrame, "TestAGMSLink")
    {};

  virtual bool DoTest ();
  virtual const char *GetDescription ()
    { return "Test AGMS Link (hard link & alias) operations."; };
};


bool TestAGMSLink::DoTest ()
{
  int         ErrorCode;

  /* First try the examples from the AGMSRAMFileSystem code documentation.
  Make the directory structure /Test/a/b/c. */

  if ((ErrorCode = create_directory ("/Test/a/b/c", 0777)) != 0)
  {
    SetTestResult ("Failed to make /Test/a/b/c directories", ErrorCode);
    goto ErrorExit;
  }

  /* Make AGMS link d, pointing back to b. */

  if (link ("/Test/a/b", "/Test/a/b/c/d"))
  {
    SetTestResult ("Unable to make link /Test/a/b/c/d to directory b", errno);
    goto ErrorExit;
  }

  if (!VerifyDirectoryTree ("/Test/a",
  ",d,*\n"
  "b,d,*\n"
  "b/...,p,/Test/a/b/c\n"
  "b/c,d,*\n"
  "b/c/d,a,/Test/a/b\n"))
    goto ErrorExit;

  /* Try deleting the a/b link, should fail because it would isolate a loop. */

  ErrorCode = 0;
  if (!rmdir ("/Test/a/b") || (ErrorCode = errno) != ELOOP)
  {
    SetTestResult ("Expected ELOOP while deleting cyclic directory /Test/a/b",
      ErrorCode);
    goto ErrorExit;
  }

  /* Make AGMS link Test/a/x, pointing down to c. */

  if (link ("/Test/a/b/c", "/Test/a/x"))
  {
    SetTestResult ("Unable to make link /Test/a/x to directory c", errno);
    goto ErrorExit;
  }

  if (!VerifyDirectoryTree ("/Test/a",
  ",d,*\n"
  "x,a,/Test/a/b/c\n"
  "b,d,*\n"
  "b/...,p,/Test/a/b/c\n"
  "b/c,d,*\n"
  "b/c/...,p,/Test/a\n"
  "b/c/d,a,/Test/a/b\n"))
    goto ErrorExit;

  /* Try deleting the a/b link, should rearrange the tree a lot. */

  if (rmdir ("/Test/a/b"))
  {
    SetTestResult ("Unable to delete /Test/a/b which should work", errno);
    goto ErrorExit;
  }

  /* Check that the directory structure is correct. */

  if (!VerifyDirectoryTree ("/Test/a",
  ",d,*\n"
  "x,d,*\n"
  "x/...,p,/Test/a/x/d\n"
  "x/d,d,*\n"
  "x/d/c,a,/Test/a/x\n"))
    goto ErrorExit;

  /* Wipe the slate clean.  Deleting AGMS links may not work, since it may
  rearrange the directory structure, confusing the recursive delete. */

  RecursivelyDelete ("/Test/a/x");
  if (!VerifyDirectoryTree ("/Test/a",
  ",d,*\n"))
    goto ErrorExit;

  /* Try testing multiple links to the same object, from the same directory.
  Note that the names are Link1, link2, Link3 so that some of them come before
  their target directory "c" and some after, which should make the recursive
  delete more interesting. */

  if ((ErrorCode = create_directory ("/Test/a/b/c", 0777)) != 0)
  {
    SetTestResult ("Failed to make /Test/a/b/c directories again", ErrorCode);
    goto ErrorExit;
  }

  if (link ("/Test/a/b/c", "/Test/a/b/Link1"))
  {
    SetTestResult ("Unable to make link /Test/a/b/Link1", errno);
    goto ErrorExit;
  }

  if (link ("/Test/a/b/c", "/Test/a/b/link2"))
  {
    SetTestResult ("Unable to make link /Test/a/b/link2", errno);
    goto ErrorExit;
  }

  if (link ("/Test/a/b/c", "/Test/a/b/Link3"))
  {
    SetTestResult ("Unable to make link /Test/a/b/Link3", errno);
    goto ErrorExit;
  }

  if (!VerifyDirectoryTree ("/Test/a",
  ",d,*\n"
  "b,d,*\n"
  "b/Link1,a,/Test/a/b/c\n"
  "b/link2,a,/Test/a/b/c\n"
  "b/Link3,a,/Test/a/b/c\n"
  "b/c,d,*\n"
  "b/c/...,p,/Test/a/b\n"
  "b/c/....,p,/Test/a/b\n"
  "b/c/.....,p,/Test/a/b\n"))
    goto ErrorExit;

  RecursivelyDelete ("/Test/a/b");
  if (!VerifyDirectoryTree ("/Test/a",
  ",d,*\n"))
    goto ErrorExit;

  /* Test making an AGMS link to its own parent directory. */

  if (link ("/Test/a", "/Test/a/SelfLink"))
  {
    SetTestResult ("Unable to make link /Test/a/SelfLink", errno);
    goto ErrorExit;
  }

  if (!VerifyDirectoryTree ("/Test/a",
  ",d,*\n"
  "...,p,/Test/a\n"
  "SelfLink,a,/Test/a\n"))
    goto ErrorExit;

  RecursivelyDelete ("/Test/a/SelfLink");

  /* Try making a link to a link.  Should end up linking to the target. */

  if ((ErrorCode = create_directory ("/Test/a/b/c", 0777)) != 0)
  {
    SetTestResult ("Failed to make /Test/a/b/c directories yet again",
      ErrorCode);
    goto ErrorExit;
  }

  if (link ("/Test/a/b/c", "/Test/a/b/SittingLink"))
  {
    SetTestResult ("Unable to make link /Test/a/b/SittingLink", errno);
    goto ErrorExit;
  }

  if (link ("/Test/a/b/SittingLink", "/Test/a/b/LinkToLink"))
  {
    SetTestResult ("Unable to make link /Test/a/b/LinkToLink", errno);
    goto ErrorExit;
  }

  if (!VerifyDirectoryTree ("/Test/a",
  ",d,*\n"
  "b,d,*\n"
  "b/SittingLink,a,/Test/a/b/c\n"
  "b/LinkToLink,a,/Test/a/b/c\n"
  "b/c,d,*\n"
  "b/c/...,p,/Test/a/b\n"
  "b/c/....,p,/Test/a/b\n"))
    goto ErrorExit;

  /* Test making a link to something which doesn't exist. */

  ErrorCode = 0;
  if (!link ("/Test/a/Zot", "/Test/a/Junk") || (ErrorCode = errno) != ENOENT)
  {
    SetTestResult ("Linking to a nonexistent thing should fail with ENOENT",
      ErrorCode);
    goto ErrorExit;
  }

  /* Test making a link in a non-existent directory. */

  ErrorCode = 0;
  if (!link ("/Test/a/b/c", "/Test/a/Zot/Junk") ||
  (ErrorCode = errno) != ENOENT)
  {
    SetTestResult ("Creating a link in a nonexistent directory should fail "
      "with ENOENT", ErrorCode);
    goto ErrorExit;
  }

  /* Test overwriting an existing link.  Should fail. */

  ErrorCode = 0;
  if (!link ("/Test/a", "/Test/a/b/SittingLink") ||
  (ErrorCode = errno) != EEXIST)
  {
    SetTestResult ("Creating a link that replaces an existing link "
      "should fail with EEXIST", ErrorCode);
    goto ErrorExit;
  }

  /* Test the llink() function for making a link to a link. */

  if (llink ("/Test/a/b/SittingLink", "/Test/a/b/RealLinkToLink"))
  {
    SetTestResult ("Unable to make link /Test/a/b/RealLinkToLink", errno);
    goto ErrorExit;
  }

  if (llink ("/Test/a/b/SittingLink", "/Test/a/b/c/AnotherRealLink"))
  {
    SetTestResult ("Unable to make link /Test/a/b/c/AnotherRealLink", errno);
    goto ErrorExit;
  }

  if (llink ("/Test/a/b/SittingLink", "/Test/a/ALinkAgain"))
  {
    SetTestResult ("Unable to make link /Test/a/ALinkAgain", errno);
    goto ErrorExit;
  }

  if (!VerifyDirectoryTree ("/Test/a",
  ",d,*\n"
  "ALinkAgain,a,/Test/a/b/SittingLink\n"
  "b,d,*\n"
  "b/SittingLink,a,/Test/a/b/c\n"
  "b/LinkToLink,a,/Test/a/b/c\n"
  "b/RealLinkToLink,a,/Test/a/b/SittingLink\n"
  "b/c,d,*\n"
  "b/c/...,p,/Test/a/b\n"
  "b/c/....,p,/Test/a/b\n"
  "b/c/AnotherRealLink,a,/Test/a/b/SittingLink\n"))
    goto ErrorExit;

  /* Now delete the middle link, it should merge with the link to the link, and
  the other links that formerly pointed to it should point to the one it merged
  with. */

  if (unlink ("/Test/a/b/SittingLink"))
  {
    SetTestResult ("Unable to delete /Test/a/b/SittingLink", errno);
    goto ErrorExit;
  }

  if (!VerifyDirectoryTree ("/Test/a",
  ",d,*\n"
  "ALinkAgain,a,/Test/a/b/RealLinkToLink\n"
  "b,d,*\n"
  "b/LinkToLink,a,/Test/a/b/c\n"
  "b/RealLinkToLink,a,/Test/a/b/c\n"
  "b/c,d,*\n"
  "b/c/...,p,/Test/a/b\n"
  "b/c/....,p,/Test/a/b\n"
  "b/c/AnotherRealLink,a,/Test/a/b/RealLinkToLink\n"))
    goto ErrorExit;

  SetTestResult ("AGMS Link test OK.");

ErrorExit:
  RecursivelyDelete ("/Test/a");
  return true;
}



/******************************************************************************
 * Test Attribute operations.
 */

class TestAttributes : public TestAbstract
{
public:
  TestAttributes (BRect NewFrame)
    : TestAbstract (NewFrame, "TestAttributes")
    {};

  virtual bool DoTest ();
  virtual const char *GetDescription ()
    { return "Test Attributes (read, write, dir, etc)."; };
};


bool TestAttributes::DoTest ()
{
  attr_info   AttributeInfo;
  int         ErrorCode;
  BFile       File;
  bool        GotAnother;
  bool        GotMyAttr;
  int         i;
  gid_t       OldGroupID = getegid ();
  uid_t       OldUserID = geteuid ();
  char        TempString [1024];
  const char *WriteString = "This is a string for attribute testing.";

  if (mkdir ("/Test/Holder", 0755 /* rwxr-xr-x */))
  {
    SetTestResult ("mkdir failed for /Test/Holder", errno);
    goto ErrorExit;
  }

  /* First make a simple file, with some other user and group to see if write
  access tests are done. */

  ErrorCode =
    File.SetTo ("/Test/Holder/TestFile", B_READ_WRITE | B_CREATE_FILE);
  if (ErrorCode)
  {
    SetTestResult ("Unable to create /Test/Holder/TestFile", ErrorCode);
    goto ErrorExit;
  }

  if (File.Write (WriteString, strlen (WriteString)) !=
  (int) strlen (WriteString))
  {
    SetTestResult ("Error while writing data to /Test/Holder/TestFile", errno);
    goto ErrorExit;
  }

  File.SetOwner (1234);
  File.SetGroup (5678);
  File.SetPermissions (0644);
  File.Unset ();

  setgid (1111);
  setuid (2222);

  ErrorCode = File.SetTo ("/Test/Holder/TestFile", B_READ_ONLY);
  if (ErrorCode)
  {
    SetTestResult ("Unable to reopen /Test/Holder/TestFile", ErrorCode);
    goto ErrorExit;
  }

  ErrorCode = File.WriteAttr ("MyAttr",  B_STRING_TYPE, 0 /* offset */,
    WriteString, strlen (WriteString));
  if (ErrorCode != EACCES)
  {
    SetTestResult ("Writing an attribute to a read only file should fail with "
      "EACCES (EBADF according to the docs).  BFile was opened read only, "
      "permissions read only, not root user", ErrorCode);
    if (!g_FakeMount)
      goto ErrorExit;
  }
  File.Unset ();

  setuid (1234); /* Be the owner of the file. */

  ErrorCode = File.SetTo ("/Test/Holder/TestFile", B_READ_WRITE);
  if (ErrorCode)
  {
    SetTestResult ("Unable to open again /Test/Holder/TestFile", ErrorCode);
    goto ErrorExit;
  }

  ErrorCode = File.WriteAttr ("MyAttr",  B_STRING_TYPE, 0 /* offset */,
    "Blah!", 6);
  if (ErrorCode != 6)
  {
    SetTestResult ("Writing an attribute failed", ErrorCode);
    goto ErrorExit;
  }

  /* Test removing the only attribute. */

  ErrorCode = File.RemoveAttr ("MyAttr");
  if (ErrorCode != 0)
  {
    SetTestResult ("Unable to remove only attribute", ErrorCode);
    goto ErrorExit;
  }

  /* More writing tests. */

  ErrorCode = File.WriteAttr ("MyAttr",  B_STRING_TYPE, 0 /* offset */,
    WriteString, strlen (WriteString) + 1);
  if (ErrorCode != (int) strlen (WriteString) + 1)
  {
    SetTestResult ("Rewriting an attribute failed", ErrorCode);
    goto ErrorExit;
  }

  ErrorCode = File.WriteAttr ("Another",  B_STRING_TYPE, 0 /* offset */,
    WriteString, strlen (WriteString));
  if (ErrorCode == (int) strlen (WriteString))
  {
    SetTestResult ("Was able to write a string attribute without the trailing "
      "NUL byte (not really an error, but the interface layer could check "
      "attribute validity in the future).",
      false);
  }

  ErrorCode = File.WriteAttr ("Another", 12345678, 0 /* offset */,
    WriteString, strlen (WriteString));
  if (ErrorCode != (int) strlen (WriteString))
  {
    SetTestResult ("Was unable to overwrite a string attribute with a "
      "different kind of attribute type", ErrorCode);
    goto ErrorExit;
  }

  /* Check the attribute stat function. */

  ErrorCode = File.GetAttrInfo ("Another", &AttributeInfo);
  if (ErrorCode != 0)
  {
    SetTestResult ("Unable to read attribute info", ErrorCode);
    goto ErrorExit;
  }
  if (AttributeInfo.type != 12345678 ||
  AttributeInfo.size != strlen (WriteString))
  {
    SetTestResult ("Attribute info (size or type) is wrong.", false);
    goto ErrorExit;
  }

  ErrorCode = File.GetAttrInfo ("None", &AttributeInfo);
  if (ErrorCode != ENOENT)
  {
    SetTestResult ("Expected ENOENT when getting info on "
      "nonexistent attribute", ErrorCode);
    goto ErrorExit;
  }

  /* Test the attribute directory listing functions. */

  GotAnother = false;
  GotMyAttr = false;
  ErrorCode = File.RewindAttrs ();
  if (ErrorCode != 0)
  {
    SetTestResult ("Unable to rewind attribute directory", ErrorCode);
    goto ErrorExit;
  }

  ErrorCode = File.GetNextAttrName (TempString);
  if (ErrorCode != 0)
  {
    SetTestResult ("Unable to read first name in attribute directory",
      ErrorCode);
    goto ErrorExit;
  }
  if (strcmp (TempString, "Another") == 0)
    GotAnother = true;
  else if (strcmp (TempString, "MyAttr") == 0)
    GotMyAttr = true;
  else
  {
    SetTestResult ("Got unexpected attribute name when reading list.", false);
    goto ErrorExit;
  }

  ErrorCode = File.GetNextAttrName (TempString);
  if (ErrorCode != 0)
  {
    SetTestResult ("Unable to read second name in attribute directory",
      ErrorCode);
    goto ErrorExit;
  }
  if (strcmp (TempString, "Another") == 0)
    GotAnother = true;
  else if (strcmp (TempString, "MyAttr") == 0)
    GotMyAttr = true;
  else
  {
    SetTestResult ("Got unexpected attribute name when reading list.", false);
    goto ErrorExit;
  }

  ErrorCode = File.GetNextAttrName (TempString);
  if (ErrorCode != ENOENT)
  {
    SetTestResult ("Should have reached end of name list and got ENOENT",
      ErrorCode);
    goto ErrorExit;
  }

  if (!GotAnother || !GotMyAttr)
  {
    SetTestResult ("Didn't get all the names it should have read from the "
      "name list.", false);
    goto ErrorExit;
  }

  /* More writing tests. */

  ErrorCode = File.WriteAttr ("MAIL:subject",  B_STRING_TYPE, 0 /* offset */,
    "12345678901234567890123456789012345678901234567890" /*  50 */
    "12345678901234567890123456789012345678901234567890" /* 100 */
    "12345678901234567890123456789012345678901234567890" /* 150 */
    "12345678901234567890123456789012345678901234567890" /* 200 */
    "12345678901234567890123456789012345678901234567890" /* 250 */
    "1234567890123456789012345678901234567890123456789"  /* 299 */, 300);
  if (ErrorCode != 300)
  {
    SetTestResult ("Unable to write a long (300, more than the 255 maximum "
      "for indexed things) byte string", ErrorCode);
    goto ErrorExit;
  }

  /* Test reading an attribute. */

  memset (TempString, '?', sizeof (TempString));
  ErrorCode = File.ReadAttr ("MyAttr",  B_STRING_TYPE, 0 /* offset */,
    TempString, sizeof (TempString));
  if (ErrorCode != (int) strlen (WriteString) + 1)
  {
    SetTestResult ("Unable to read an ordinary attribute", ErrorCode);
    goto ErrorExit;
  }
  if (strcmp (WriteString, TempString) != 0)
  {
    SetTestResult ("Data read back from an ordinary attribute is wrong.",
      false);
    goto ErrorExit;
  }

  ErrorCode = File.ReadAttr ("NoName",  B_STRING_TYPE, 0 /* offset */,
    TempString, sizeof (TempString));
  if (ErrorCode != ENOENT)
  {
    SetTestResult ("Expected ENOENT when reading an non-existent attribute",
       ErrorCode);
    goto ErrorExit;
  }

  /* Test removing one of many attributes. */

  ErrorCode = File.RemoveAttr ("Another");
  if (ErrorCode != 0)
  {
    SetTestResult ("Unable to remove one of many attributes", ErrorCode);
    goto ErrorExit;
  }

  ErrorCode = File.RemoveAttr ("NoName");
  if (ErrorCode != ENOENT)
  {
    SetTestResult ("Expected ENOENT when "
      "removing a non-existent attribute", ErrorCode);
    goto ErrorExit;
  }

  /* Test writing a big attribute, 1 exabyte (10e18). */

  ErrorCode = File.WriteAttr ("Exabyte", 'EXAB',
    1000000000000000000ULL /* offset */, WriteString, strlen (WriteString));
  if (ErrorCode != EFBIG)
  {
    SetTestResult ("While writing at offset 10e18, expected EFBIG (possibly "
      "offsets aren't implemented)", ErrorCode);
    if (!g_FakeMount)
      goto ErrorExit;
  }

  /* See if the uninitialised data in an attribute is zero, and test if the
  offset operations work. */

  ErrorCode = File.WriteAttr ("Uninit", 'TEST',
    10000 /* offset */, WriteString, strlen (WriteString) + 1);
  if (ErrorCode != (int) strlen (WriteString) + 1)
  {
    SetTestResult ("Error while writing at offset 10000", ErrorCode);
    goto ErrorExit;
  }

  memset (TempString, '?', sizeof (TempString));
  ErrorCode = File.ReadAttr ("Uninit",  'TEST', 0 /* offset */,
    TempString, sizeof (TempString));
  if (ErrorCode != sizeof (TempString))
  {
    SetTestResult ("Unable to read full buffer doing uninit/offset "
      "test (possibly offsets are ignored)", ErrorCode);
    if (!g_FakeMount)
      goto ErrorExit;
  }
  for (i = 0; i < (int) sizeof (TempString); i++)
  {
    if (TempString[i] != 0)
    {
      SetTestResult ("Supposedly uninitialised data isn't zero.", false);
      if (!g_FakeMount)
        goto ErrorExit;
      break;
    }
  }

  memset (TempString, '?', sizeof (TempString));
  ErrorCode = File.ReadAttr ("Uninit", 'TEST', 10000 /* offset */,
    TempString, sizeof (TempString));
  if (ErrorCode != (int) strlen (WriteString) + 1)
  {
    SetTestResult ("Got wrong amount when reading at offset 10000, "
      "partly going past end of attribute", ErrorCode);
    if (!g_FakeMount)
      goto ErrorExit;
  }
  if (strcmp (WriteString, TempString) != 0)
  {
    SetTestResult ("Data read back from attribute at offset 10000 is wrong.",
      false);
    if (!g_FakeMount)
      goto ErrorExit;
  }

  /* See if we can write to a reserved word attribute. */

  ErrorCode = File.ReadAttr ("size", B_UINT64_TYPE, 0 /* offset */,
    TempString, sizeof (TempString));
  if (ErrorCode <= 0)
  {
    SetTestResult ("Unable to read reserved word \"size\" attribute "
      "(normal, but fancier file systems could export stat info as "
      "attributes in the future)", ErrorCode);
  }

  ErrorCode = File.WriteAttr ("size", B_STRING_TYPE, 0 /* offset */,
    WriteString, strlen (WriteString) + 1);
  if (ErrorCode == (int) strlen (WriteString) + 1)
  {
    SetTestResult ("Was able to write to reserved name \"size\" attribute, "
      "and with the wrong data type too (BFS has this bug).", false);
  }

  /* Try renaming attributes.  First a name to a different name. */

  if (File.WriteAttr ("OldName", 'CSTR', 0 /* offset */, "Testing", 8) != 8)
  {
    SetTestResult ("Error while writing OldName.", errno);
    goto ErrorExit;
  }

  ErrorCode = File.RenameAttr ("OldName", "NewName");
  if (ErrorCode != 0)
  {
    SetTestResult ("Error while renaming OldName to NewName", ErrorCode);
    if (!g_FakeMount)
      goto ErrorExit;
  }

  /* Try renaming over an existing attribute. */

  if (!g_FakeMount)
  {
    ErrorCode = File.RenameAttr ("NewName", "Uninit");
    if (ErrorCode != 0)
    {
      SetTestResult ("Error while renaming NewName to Uninit", ErrorCode);
      goto ErrorExit;
    }

    /* Try renaming from a non-existent attribute. */

    ErrorCode = File.RenameAttr ("NoSuchName", "Uninit");
    if (ErrorCode != ENOENT)
    {
      SetTestResult ("Should get ENOENT when renaming a nonexistent attribute",
        ErrorCode);
      goto ErrorExit;
    }

    /* See that the result has the right value. */

    memset (TempString, '?', sizeof (TempString));
    ErrorCode = File.ReadAttr ("Uninit",  'CSTR', 0 /* offset */,
      TempString, sizeof (TempString));
    if (ErrorCode != 8)
    {
      SetTestResult ("Error while reading renamed attribute", ErrorCode);
      goto ErrorExit;
    }
    if (strcmp (TempString, "Testing") != 0)
    {
      SetTestResult ("Contents of renamed attribute are wrong.", false);
      goto ErrorExit;
    }
  }

  SetTestResult ("Attribute test finished successfully.");

ErrorExit:
  File.Unset ();
  setgid (OldGroupID);
  setuid (OldUserID);
  RecursivelyDelete ("/Test/Holder");
  return true;
}



/******************************************************************************
 * Test file creation and opening.
 */

class TestCreateFile : public TestAbstract
{
public:
  TestCreateFile (BRect NewFrame)
    : TestAbstract (NewFrame, "TestCreateFile")
    {};

  virtual bool DoTest ();

  virtual const char *GetDescription ()
    { return "Test creation & deletion of files."; };
};


bool TestCreateFile::DoTest ()
{
  int  ErrorCode;
  int  FileHandle0 = -1;
  int  FileHandle1 = -1;
  char TempString [1024];

  /* Try a plain open of a file which doesn't exist. */

  ErrorCode = 0;
  FileHandle0 = open ("/Test/File1", O_RDWR, 0777);
  if (FileHandle0 >= 0 || (ErrorCode = errno) != ENOENT)
  {
    SetTestResult ("Plain open of non-existing file /Test/File1 should fail "
      "with ENOENT", ErrorCode);
    goto ErrorExit;
  }

  /* Should not be able to overwrite an existing directory with the same name,
  even if special open flags used.  Or at least that's my theory. */

  if (mkdir ("/Test/Create", S_IRWXU))
  {
    SetTestResult ("mkdir failed for /Test/Create", errno);
    goto ErrorExit;
  }

  ErrorCode = 0;
  FileHandle0 = creat ("/Test/Create", (unsigned int) -1 /* permissions */);
  if (FileHandle0 >= 0 || (ErrorCode = errno) != EISDIR)
  {
    SetTestResult ("Create /Test/Create should fail with EISDIR", ErrorCode);
    goto ErrorExit;
  }

  /* Should not be able to create files with magic names. */

  FileHandle0 = creat ("/Test/..", 0777 /* access mode flags */);
  if (FileHandle0 >= 0)
  {
    SetTestResult ("Should not be able to create \"/Test/..\".", false);
    goto ErrorExit;
  }

  if (!g_FakeMount) /* Regular BFS allows ... as a name so don't test BFS. */
  {
    FileHandle0 = creat ("/Test/...", 0777 /* access mode flags */);
    if (FileHandle0 >= 0)
    {
      SetTestResult ("Should not be able to create \"/Test/...\".",
        false);
      goto ErrorExit;
    }
  }

  /* Try a plain ordinary file creation.  Note that default open mode for the
  "creat" call seems to be 0x601 == O_CREAT | O_TRUNC | O_WRONLY. */

  FileHandle0 = creat ("/Test/File1", (unsigned int) -1 /* permissions */);
  if (FileHandle0 < 0)
  {
    SetTestResult ("Failed to create Test/File1", errno);
    goto ErrorExit;
  }

  strcpy (TempString, "This should be wiped out by the next create - "
    "which will truncate the old file.");
  if (write (FileHandle0, TempString, strlen (TempString)) !=
  (int) strlen (TempString))
  {
    SetTestResult ("Failed to write full string to Test/File1", errno);
    goto ErrorExit;
  }

  /* See if close works. */

  if (close (FileHandle0))
  {
    SetTestResult ("Close of /Test/File1 failed", errno);
    goto ErrorExit;
  }
  FileHandle0 = -1;

  /* Create it again.  Should work.  Or does it internally use open?
  Apparently it does use internal create on existing files.  And it should
  truncate the previous data in the file. */

  FileHandle0 = creat ("/Test/File1", 0377 /* permission flags */);
  if (FileHandle0 < 0)
  {
    SetTestResult ("Failed second create Test/File1", errno);
    goto ErrorExit;
  }

  /* Write some data.  We'll check reading it later. */

  if (write (FileHandle0, "Test Data", 9) != 9)
  {
    SetTestResult ("Unable to write to /Test/File1", errno);
    goto ErrorExit;
  }

  if (close (FileHandle0))
  {
    SetTestResult ("Close of second /Test/File1 failed", errno);
    goto ErrorExit;
  }
  FileHandle0 = -1;

  /* Try creating another one with the same name, but use open(),
  should fail because it already exists and we used O_EXCL. */

  ErrorCode = 0;
  FileHandle1 = open ("/Test/File1", O_RDWR | O_CREAT | O_EXCL,
    -1 /* access mode flags */);
  if (FileHandle1 >= 0 || (ErrorCode = errno) != EEXIST)
  {
    SetTestResult ("Created existing file /Test/File1, should fail "
      "with EEXIST", ErrorCode);
    goto ErrorExit;
  }

  /* Try opening a file read only and doing a read and a write. */

  FileHandle0 = open ("/Test/File1", O_RDONLY, S_IRWXU);
  if (FileHandle0 < 0)
  {
    SetTestResult ("Should be able to open /Test/File1 as "
      "a read only file", errno);
    goto ErrorExit;
  }

  memset (TempString, 0, sizeof (TempString));
  if (read (FileHandle0, TempString, sizeof (TempString)) != 9)
  {
    SetTestResult ("Failed to read expected amount of data from /Test/File1",
      errno);
    goto ErrorExit;
  }
  if (strcmp (TempString, "Test Data") != 0)
  {
    SetTestResult ("Failed to read back correct data from /Test/File1.",
      false);
    goto ErrorExit;
  }

  ErrorCode = 0;
  if (write (FileHandle0, "Huh?", 5) >= 0 || (ErrorCode = errno) != EBADF)
  {
    SetTestResult ("Should get error EBADF when writing data to "
      "/Test/File1 opened readonly", ErrorCode);
    goto ErrorExit;
  }
  close (FileHandle0);

  /* Open write-only and try reading. */

  FileHandle0 = open ("/Test/File1", O_WRONLY, S_IRWXU);
  if (FileHandle0 < 0)
  {
    SetTestResult ("Should be able to open /Test/File1 as "
      "a write only file", errno);
    goto ErrorExit;
  }

  ErrorCode = 0;
  if (read (FileHandle0, TempString, sizeof (TempString)) >= 0 ||
  (ErrorCode = errno) != EBADF)
  {
    SetTestResult ("Should get EBADF when reading from a write only "
      "/Test/File1", ErrorCode);
    goto ErrorExit;
  }
  close (FileHandle0);

  /* Try append mode. */

  FileHandle0 = open ("/Test/File1", O_RDWR | O_APPEND, 0777);
  if (FileHandle0 < 0)
  {
    SetTestResult ("Open with append of existing file /Test/File1 failed",
      errno);
    goto ErrorExit;
  }

  if (write (FileHandle0, "Appended", 8) != 8)
  {
    SetTestResult ("Unable to append to /Test/File1", errno);
    goto ErrorExit;
  }

  /* Go back to the beginning of the file and write a bit there.  In append
  mode, it will ignore your lseeks for writing, and write it at the end. */

  if (lseek (FileHandle0, 0, SEEK_SET) < 0)
  {
    SetTestResult ("Seek to zero of file /Test/File1 failed", errno);
    goto ErrorExit;
  }

  if (write (FileHandle0, "Good", 4) != 4)
  {
    SetTestResult ("Unable to overwrite /Test/File1", errno);
    goto ErrorExit;
  }

  /* Go back to the beginning of the file and read it. */

  if (lseek (FileHandle0, 0, SEEK_SET) < 0)
  {
    SetTestResult ("Seek to zero of file /Test/File1 failed", errno);
    goto ErrorExit;
  }

  memset (TempString, 0, sizeof (TempString));
  if (read (FileHandle0, TempString, sizeof (TempString)) != 21)
  {
    SetTestResult ("Didn't read right amount of data from append test.",
      false);
    goto ErrorExit;
  }

  if (strcmp (TempString, "Test DataAppendedGood") != 0)
  {
    SetTestResult ("Failed to read back correct append data from /Test/File1.",
      false);
    goto ErrorExit;
  }

  SetTestResult ("File creation tests finished successfully.");

ErrorExit:
  if (FileHandle0 >= 0)
    close (FileHandle0);
  FileHandle0 = -1;
  if (FileHandle1 >= 0)
    close (FileHandle1);
  unlink ("/Test/File1");
  rmdir ("/Test/Create");

  return true;
}



/******************************************************************************
 * Instead of mounting the file system, make a symbolic link to a place like
 * /boot/var/tmp/Test and use that instead, so that we can run our tests
 * against a real file system.
 */

class TestFakeSymbolicMount : public TestAbstract
{
public:
  TestFakeSymbolicMount (BRect NewFrame);
  virtual ~TestFakeSymbolicMount ();
  virtual void MessageReceived (BMessage *MessagePntr);
  virtual float AddCustomControls (float X);
  virtual bool DoTest ();

  virtual const char *GetDescription ()
    { return "Fake mount - symlink /Test to a directory."; };

  BTextControl *m_DirectoryTextBoxPntr;
};


TestFakeSymbolicMount::TestFakeSymbolicMount (BRect NewFrame)
: TestAbstract (NewFrame, "TestFakeSymbolicMount"),
  m_DirectoryTextBoxPntr (NULL)
{
  /* Make sure there is space for our "directory" text box. */

  if (m_GUIHeight < g_TextBoxHeight)
    m_GUIHeight = g_TextBoxHeight;
}


TestFakeSymbolicMount::~TestFakeSymbolicMount ()
{
  /* Child views (controls) are automatically destroyed and deallocated by
  the parent destructor.  So there isn't much to do here. */

  m_DirectoryTextBoxPntr = NULL;
}


void TestFakeSymbolicMount::MessageReceived (BMessage *MessagePntr)
{
  switch (MessagePntr->what)
  {
    case MSG_FAKE_DIRECTORY:
      if (m_DirectoryTextBoxPntr != NULL)
      {
        strncpy (g_FakeDirString, m_DirectoryTextBoxPntr->Text (),
          sizeof (g_FakeDirString) - 1);
        g_FakeDirString [sizeof (g_FakeDirString) - 1] = 0;
      }
      break;

    default:
      TestAbstract::MessageReceived (MessagePntr);
      break;
  }
}


float TestFakeSymbolicMount::AddCustomControls (float X)
{
  float Margin;
  BRect TempRect;

  /* Add the text control used to specify the directory to use. */

  Margin = ceilf ((m_GUIHeight - g_TextBoxHeight) / 2);
  TempRect = Bounds ();
  TempRect.top += Margin;
  TempRect.bottom = TempRect.top + g_TextBoxHeight;
  TempRect.left = X;
  TempRect.right = X + 100;

  m_DirectoryTextBoxPntr = new BTextControl (TempRect,
    "Fake Directory Text Control",
    "Dir:" /* label */,
    g_FakeDirString,
    new BMessage (MSG_FAKE_DIRECTORY),
    B_FOLLOW_LEFT | B_FOLLOW_TOP,
    B_WILL_DRAW | B_FRAME_EVENTS | B_FULL_UPDATE_ON_RESIZE | B_NAVIGABLE);
  if (m_DirectoryTextBoxPntr == NULL) goto ErrorExit;
  m_DirectoryTextBoxPntr->SetDivider (be_plain_font->StringWidth ("Dir: "));
  AddChild (m_DirectoryTextBoxPntr);
  m_DirectoryTextBoxPntr->SetTarget (this);
  X += ceilf (m_DirectoryTextBoxPntr->Bounds().Width());

  return X;

ErrorExit:
  return -1.0;
}


bool TestFakeSymbolicMount::DoTest ()
{
  int  ErrorCode;
  char TempString [PATH_MAX + 80];

  if (mkdir (g_FakeDirString, 0777 /* access mode flags */))
  {
    ErrorCode = errno;
    if (ErrorCode != EEXIST)
    {
      sprintf (TempString, "Can't make \"%s\" directory", g_FakeDirString);
      SetTestResult (TempString, ErrorCode);
      return false;
    }
  }

  if (symlink (g_FakeDirString, "/Test"))
  {
    ErrorCode = errno;
    sprintf (TempString, "Problems setting up /Test link to \"%s\" directory",
      g_FakeDirString);
    SetTestResult (TempString, ErrorCode);
    return false;
  }

  sprintf (TempString, "Symlink /Test now points to directory \"%s\".",
    g_FakeDirString);
  SetTestResult (TempString);
  g_FakeMount = true;
  return true;
}



/******************************************************************************
 * Test Index operations (just creating and listing etc, not actually doing
 * queries).
 */

class TestIndices : public TestAbstract
{
public:
  TestIndices (BRect NewFrame)
    : TestAbstract (NewFrame, "TestIndices")
    {};

  virtual bool DoTest ();
  virtual const char *GetDescription ()
    { return "Test Indices (create, list, remove)."; };
};


bool TestIndices::DoTest ()
{
  int32       AnInt32;
  int64       AnInt64;
  dev_t       DeviceID;
  dirent_t   *DirentPntr;
  DIR        *DirPntr = NULL;
  int         ErrorCode;
  BFile       File;
  char        FirstName [NAME_MAX];
  bool        FoundOurIndex;
  index_info  IndexInfo;
  off_t       IndexSize;
  char        OldWorkingDirectory [PATH_MAX];

  if (getcwd (OldWorkingDirectory, sizeof (OldWorkingDirectory)) !=
  OldWorkingDirectory)
  {
    SetTestResult ("Unable to get current working directory.", false);
    return false;
  }

  if (chdir ("/Test") != 0)
  {
    SetTestResult ("Unable to change working directory to /Test", errno);
    goto ErrorExit;
  }

  DeviceID = dev_for_path (".");
  if (DeviceID < 0)
  {
    SetTestResult ("Unable to get device ID for /Test", errno);
    goto ErrorExit;
  }

  if (fs_create_index (DeviceID, "JunkIndex", B_INT32_TYPE, 0) != 0)
  {
    SetTestResult ("Unable to create JunkIndex with type B_INT32_TYPE", errno);
    goto ErrorExit;
  }

  /* Create it again, should fail with an error. */

  ErrorCode = 0;
  if (fs_create_index (DeviceID, "JunkIndex", B_INT32_TYPE, 0) == 0 ||
  (ErrorCode = errno) != EEXIST)
  {
    SetTestResult ("Duplicate create of JunkIndex should fail with EEXIST",
      ErrorCode);
    if (!g_FakeMount)
      goto ErrorExit;
  }

  /* Get info about the index.  Remember the size for later. */

  if (fs_stat_index (DeviceID, "JunkIndex", &IndexInfo))
  {
    SetTestResult ("fs_stat_index failed on JunkIndex", errno);
    goto ErrorExit;
  }

  if (IndexInfo.type != B_INT32_TYPE)
  {
    SetTestResult ("Index type from fs_stat_index doesn't original type.",
      false);
    goto ErrorExit;
  }

  IndexSize = IndexInfo.size;

  /* Add a file to the system with an attribute that will be in the index. */

  ErrorCode = File.SetTo ("/Test/File", B_READ_WRITE | B_CREATE_FILE);
  if (ErrorCode)
  {
    SetTestResult ("Unable to create /Test/File", ErrorCode);
    goto ErrorExit;
  }

  AnInt32 = 1234567890;
  if (File.WriteAttr ("JunkIndex",  B_INT32_TYPE, 0 /* offset */,
  &AnInt32, sizeof (AnInt32)) != sizeof (AnInt32))
  {
    SetTestResult ("Unable to add int32 attribute to /Test/File", errno);
    goto ErrorExit;
  }
  File.Unset ();

  /* Check if the index size has gotten bigger. */

  fs_stat_index (DeviceID, "JunkIndex", &IndexInfo);
  if (IndexSize == IndexInfo.size)
  {
    SetTestResult ("Index size didn't change after adding a file!", false);
    if (!g_FakeMount)
      goto ErrorExit;
  }

  /* List all the indices, see if our new one is there too. */

  DirPntr = fs_open_index_dir (DeviceID);
  if (DirPntr == NULL)
  {
    SetTestResult ("Unable to open directory of indices", errno);
    goto ErrorExit;
  }

  FirstName[0] = 0;
  FoundOurIndex = false;
  errno = 0;
  while ((DirentPntr = fs_read_index_dir (DirPntr)) != NULL)
  {
    if (FirstName[0] == 0)
      strcpy (FirstName, DirentPntr->d_name);

    if (strcmp (DirentPntr->d_name, "JunkIndex") == 0)
      FoundOurIndex = true;
  }
  if (errno != 0)
  {
    SetTestResult ("Got an odd error while reading indices directory", errno);
    goto ErrorExit;
  }
  if (!FoundOurIndex)
  {
    SetTestResult ("Didn't find our new index in the list of indices.", false);
    goto ErrorExit;
  }

  /* Test the rewind feature.  For some reason it doesn't use errno. */

  fs_rewind_index_dir (DirPntr);

  if ((DirentPntr = fs_read_index_dir (DirPntr)) == NULL)
  {
    SetTestResult ("Failed to read index directory after rewind", errno);
    goto ErrorExit;
  }

  if (strcmp (FirstName, DirentPntr->d_name) != 0)
  {
    SetTestResult ("First entry after rewind doesn't match original one.",
      false);
    goto ErrorExit;
  }

  if (fs_close_index_dir (DirPntr))
  {
    SetTestResult ("fs_close_index_dir failed", errno);
    goto ErrorExit;
  }
  DirPntr = NULL;

  /* Try removing the index. */

  if (fs_remove_index (DeviceID, "JunkIndex"))
  {
    SetTestResult ("Unable to remove JunkIndex", errno);
    goto ErrorExit;
  }

  /* fs_stat_index should fail when the index is gone. */

  ErrorCode = 0;
  if (fs_stat_index (DeviceID, "JunkIndex", &IndexInfo) == 0 ||
  (ErrorCode = errno) != ENOENT)
  {
    SetTestResult ("fs_stat_index should fail with ENOENT after looking "
      "for a deleted index", ErrorCode);
    goto ErrorExit;
  }

  /* Try to make an index with an unknown data type (the file system wouldn't
  know how to sort it so it would be useless). */

  ErrorCode = 0;
  if (fs_create_index (DeviceID, "JunkIndex", 'Zort', 0) == 0 ||
  (ErrorCode = errno) != ERANGE)
  {
    SetTestResult ("Expected ERANGE when creating an index with an "
      "unknown data type", ErrorCode);
    if (!g_FakeMount)
      goto ErrorExit;
  }

  /* Try the MIMS mime string datatype which BeOS converts to a CSTR index, to
  see if you can use the true type as the index type. */

  if (fs_create_index (DeviceID, "JunkIndex", 'MIMS', 0))
  {
    SetTestResult ("Unable to create an index of type MIMS (MIME string)",
      errno);
    goto ErrorExit;
  }

  fs_stat_index (DeviceID, "JunkIndex", &IndexInfo);
  if (IndexInfo.type != 'MIMS')
  {
    SetTestResult ("MIMS index shows up as some other data type.", false);
    if (!g_FakeMount)
      goto ErrorExit;
  }

  fs_remove_index (DeviceID, "JunkIndex");

  /* Try making a B_INT32_TYPE index and add some attributes which aren't quite
  the right type.  CSTR attributes and some double attributes.  Does it check
  the index type vs the attribute type? */

  if (fs_create_index (DeviceID, "JunkIndex", B_INT32_TYPE, 0))
  {
    SetTestResult ("Unable to create an index of type B_INT32_TYPE", errno);
    goto ErrorExit;
  }

  ErrorCode = File.SetTo ("/Test/File", B_READ_WRITE);
  if (ErrorCode)
  {
    SetTestResult ("Unable to reopen /Test/File", ErrorCode);
    goto ErrorExit;
  }

  if (File.WriteAttr ("JunkIndex", B_STRING_TYPE, 0 /* offset */,
  "Hi", 3) != 3)
  {
    SetTestResult ("Unable to add short B_STRING_TYPE attribute to /Test/File",
      errno);
    goto ErrorExit;
  }

  AnInt64 = -1000000000000LL;
  if (File.WriteAttr ("JunkIndex",  B_INT64_TYPE, 0 /* offset */,
  &AnInt64, sizeof (AnInt64)) != sizeof (AnInt64))
  {
    SetTestResult ("Unable to add B_INT64_TYPE attribute to /Test/File",
      errno);
    goto ErrorExit;
  }
  File.Unset ();

  SetTestResult ("Indices test finished successfully.");

ErrorExit:
  if (DirPntr != NULL)
    fs_close_index_dir (DirPntr);
  File.Unset ();
  unlink ("/Test/File");
  fs_remove_index (DeviceID, "JunkIndex");
  chdir (OldWorkingDirectory);
  return true;
}



/******************************************************************************
 * Try making directories, and getting info about them.
 */

class TestMakeDirectory : public TestAbstract
{
public:
  TestMakeDirectory (BRect NewFrame)
    : TestAbstract (NewFrame, "TestMakeDirectory")
    {};

  virtual bool DoTest ();

  virtual const char *GetDescription ()
    { return "Test mkdir, rmdir, readdir & subfunctions."; };
};


bool TestMakeDirectory::DoTest ()
{
  dirent_t       DirEntArray [400];
  dirent_t      *DirEntPntr;
  DIR           *DirHandle = NULL;
  int            ErrorCode;
  int            FileHandle0 = -1;
  int32          NumberRead;
  char           TempString [1024];
  BDirectory     TestDir;

  if (mkdir ("/Test/DirTest", S_IRWXU))
  {
    ErrorCode = errno;
    SetTestResult ("mkdir failed for /Test/DirTest", ErrorCode);
    goto ErrorExit;
  }

  /* Make directories /Test/DirTest/Directory2, 1, 3, 5, 4 in that order. */

  if (mkdir ("/Test/DirTest/Directory2", S_IRWXU))
  {
    ErrorCode = errno;
    SetTestResult ("mkdir failed for /Test/DirTest/Directory2", ErrorCode);
    goto ErrorExit;
  }

  if (mkdir ("/Test/DirTest/Directory1", S_IRWXU))
  {
    ErrorCode = errno;
    SetTestResult ("mkdir failed for /Test/DirTest/Directory1", ErrorCode);
    goto ErrorExit;
  }

  if (mkdir ("/Test/DirTest/Directory3", S_IRWXU))
  {
    ErrorCode = errno;
    SetTestResult ("mkdir failed for /Test/DirTest/Directory3", ErrorCode);
    goto ErrorExit;
  }

  /* Check for duplicate name detection. */

  ErrorCode = 0;
  if (!mkdir ("/Test/DirTest/Directory1", S_IRWXU) || (ErrorCode = errno) != EEXIST)
  {
    SetTestResult ("Was able to make a duplicate /Test/DirTest/Directory1",
      ErrorCode);
    goto ErrorExit;
  }

  /* Check for subdirectory creation allowed.  Well, it's kind of already tested. */

  if (mkdir ("/Test/DirTest/Directory3/SubDirInThree", S_IRWXU))
  {
    ErrorCode = errno;
    SetTestResult ("makedir failed for /Test/DirTest/Directory3/SubDirInThree",
      ErrorCode);
    goto ErrorExit;
  }

  /* Check for opening a directory as a file, it should work, even if you open
  it as writeable.  But you shouldn't be able to actually write to it. */

  FileHandle0 = open ("/Test/DirTest/Directory1", O_RDWR, S_IRWXU);
  if (FileHandle0 < 0)
  {
    SetTestResult ("Should be able to open /Test/DirTest/Directory1 as "
      "a writeable file", errno);
  }
  ErrorCode = 0;
  if (write (FileHandle0, "Huh?", 5) >= 0 || (ErrorCode = errno) != EISDIR)
  {
    SetTestResult ("Should get error EISDIR when writing data to "
      "/Test/DirTest/Directory1 which was opened read/write", ErrorCode);
    goto ErrorExit;
  }
  close (FileHandle0);

  /* You should be able to open a directory as a file in read only mode, but
  not actually read from it. */

  FileHandle0 = open ("/Test/DirTest/Directory1", O_RDONLY, S_IRWXU);
  if (FileHandle0 < 0)
  {
    SetTestResult ("Should be able to open /Test/DirTest/Directory1 as "
      "a read only file", errno);
    goto ErrorExit;
  }

  ErrorCode = 0;
  if (read (FileHandle0, TempString, sizeof (TempString)) >= 0 ||
  (ErrorCode = errno) != EISDIR)
  {
    SetTestResult ("Should get error EISDIR when reading data from "
      "/Test/DirTest/Directory1", ErrorCode);
    goto ErrorExit;
  }

  ErrorCode = 0;
  if (write (FileHandle0, "Huh?", 5) >= 0 || (ErrorCode = errno) != EBADF)
  {
    SetTestResult ("Should get error EBADF when writing data to "
      "/Test/DirTest/Directory1 opened readonly", ErrorCode);
    goto ErrorExit;
  }
  close (FileHandle0);

  /* Another directory creation call, plus subdir at the same time,
  plus see if umask makes a difference. */

  ErrorCode = create_directory ("/Test/DirTest/Directory4/SubDirInFour",
    (unsigned int) -1 /* All mode bits on */);
  if (ErrorCode != 0)
  {
    SetTestResult ("create_directory failed for "
      "/Test/DirTest/Directory4/SubDirInFour", ErrorCode);
    goto ErrorExit;
  }

  /* Yet another way of making directories. */

  if (TestDir.SetTo ("/Test"))
  {
    SetTestResult ("SetTo of a BDirectory failed for /Test.", false);
    goto ErrorExit;
  }

  if (!TestDir.IsRootDirectory ())
  {
    SetTestResult ("BDirectory /Test should be a root directory.", false);
    if (!g_FakeMount)
      goto ErrorExit; /* Only real one is a root of a file system. */
  }

  ErrorCode = TestDir.CreateDirectory ("DirTest/Directory5", NULL);
  if (ErrorCode != 0)
  {
    SetTestResult ("BDirectory::CreateDirectory failed for Directory5",
      ErrorCode);
    goto ErrorExit;
  }

  /* Test for name too long.  255 should be OK, 256 should be too long (the
  limit is 256 bytes including the NUL). */

  if (mkdir ("/Test/DirTest/"
  "Testing with a name that's the maximum length....." /* 50 */
  "12345678901234567890123456789012345678901234567890" /* 50 */
  "12345678901234567890123456789012345678901234567890" /* 50 */
  "12345678901234567890123456789012345678901234567890" /* 50 */
  "12345678901234567890123456789012345678901234567890" /* 50 */
  "12345", S_IRWXU))
  {
    ErrorCode = errno;
    SetTestResult ("mkdir with 255 character name failed", ErrorCode);
    goto ErrorExit;
  }

  ErrorCode = 0;
  if (!mkdir ("/Test/DirTest/"
  "This name should be too long. 12345678901234567890" /* 50 */
  "12345678901234567890123456789012345678901234567890" /* 50 */
  "12345678901234567890123456789012345678901234567890" /* 50 */
  "12345678901234567890123456789012345678901234567890" /* 50 */
  "12345678901234567890123456789012345678901234567890" /* 50 */
  "123456", S_IRWXU) ||
  (ErrorCode = errno) != ENAMETOOLONG)
  {
    SetTestResult ("Was able to make a directory with a name too long",
      ErrorCode);
    goto ErrorExit;
  }

  /* Names that are all periods should be invalid. */

  ErrorCode = 0;
  if (!mkdir ("/Test/DirTest/.", S_IRWXU) || (ErrorCode = errno) != EINVAL)
  {
    SetTestResult ("Was able to make a directory with a name of \".\"",
      ErrorCode);
    goto ErrorExit;
  }

  ErrorCode = 0;
  if (!mkdir ("/Test/DirTest/..", S_IRWXU) || (ErrorCode = errno) != EINVAL)
  {
    SetTestResult ("Was able to make a directory with a name of \"..\"",
      ErrorCode);
    goto ErrorExit;
  }

  ErrorCode = 0;
  if (!mkdir ("/Test/DirTest/...", S_IRWXU) || (ErrorCode = errno) != EINVAL)
  {
    SetTestResult ("Was able to make a directory with a name of \"...\"",
      ErrorCode);
    if (!g_FakeMount)
      goto ErrorExit;
    /* Regular BFS allows ... as a user name, so we will ignore it. */
    rmdir ("/Test/DirTest/...");
  }

  /* Read back the list of directories, should be in alphabetical order. */

  DirHandle = opendir ("/Test/DirTest");
  if (DirHandle == NULL)
  {
    ErrorCode = errno;
    SetTestResult ("Unable to opendir /Test/DirTest", ErrorCode);
    goto ErrorExit;
  }

  DirEntPntr = readdir (DirHandle);
  if (DirEntPntr == NULL || strcmp (DirEntPntr->d_name, "."))
  {
    SetTestResult ("First directory isn't \".\".", false);
    goto ErrorExit;
  }

  DirEntPntr = readdir (DirHandle);
  if (DirEntPntr == NULL || strcmp (DirEntPntr->d_name, ".."))
  {
    SetTestResult ("Second directory isn't \"..\".", false);
    goto ErrorExit;
  }

  DirEntPntr = readdir (DirHandle);
  if (DirEntPntr == NULL || strcmp (DirEntPntr->d_name, "Directory1"))
  {
    SetTestResult ("Third directory isn't \"Directory1\".", false);
    goto ErrorExit;
  }

  DirEntPntr = readdir (DirHandle);
  if (DirEntPntr == NULL || strcmp (DirEntPntr->d_name, "Directory2"))
  {
    SetTestResult ("Fourth directory isn't \"Directory2\".", false);
    goto ErrorExit;
  }

  DirEntPntr = readdir (DirHandle);
  if (DirEntPntr == NULL || strcmp (DirEntPntr->d_name, "Directory3"))
  {
    SetTestResult ("Fifth directory isn't \"Directory3\".", false);
    goto ErrorExit;
  }

  DirEntPntr = readdir (DirHandle);
  if (DirEntPntr == NULL || strcmp (DirEntPntr->d_name, "Directory4"))
  {
    SetTestResult ("Sixth directory isn't \"Directory4\".", false);
    goto ErrorExit;
  }

  DirEntPntr = readdir (DirHandle);
  if (DirEntPntr == NULL || strcmp (DirEntPntr->d_name, "Directory5"))
  {
    SetTestResult ("Seventh directory isn't \"Directory5\".", false);
    goto ErrorExit;
  }

  DirEntPntr = readdir (DirHandle);
  if (DirEntPntr == NULL || strcmp (DirEntPntr->d_name,
  "Testing with a name that's the maximum length....." /* 50 */
  "12345678901234567890123456789012345678901234567890" /* 50 */
  "12345678901234567890123456789012345678901234567890" /* 50 */
  "12345678901234567890123456789012345678901234567890" /* 50 */
  "12345678901234567890123456789012345678901234567890" /* 50 */
  "12345"))
  {
    SetTestResult ("Eighth directory has the wrong name.", false);
    goto ErrorExit;
  }

  DirEntPntr = readdir (DirHandle);
  if (DirEntPntr != NULL)
  {
    SetTestResult ("Shouldn't be a 9th directory.", false);
    goto ErrorExit;
  }

  /* Test rewinddir. */

  rewinddir (DirHandle);
  DirEntPntr = readdir (DirHandle);
  DirEntPntr = readdir (DirHandle);
  DirEntPntr = readdir (DirHandle);
  if (DirEntPntr == NULL || strcmp (DirEntPntr->d_name, "Directory1"))
  {
    SetTestResult ("rewinddir didn't work properly.", false);
    goto ErrorExit;
  }

  if (closedir (DirHandle))
  {
    ErrorCode = errno;
    SetTestResult ("closedir failed", ErrorCode);
    goto ErrorExit;
  }
  DirHandle = NULL;

  /* BFS and most other file systems only return one dirent at a time, so skip
  all these multiple dirent tests for them. */

  TestDir.SetTo ("/Test/DirTest");
  if (!g_FakeMount)
  {
    /* Try the BDirectory multiple dirent reading facility. */

    TestDir.Rewind ();
    NumberRead = TestDir.GetNextDirents (DirEntArray, sizeof (DirEntArray));
    if (NumberRead != 8)
    {
      SetTestResult ("GetNextDirents didn't read enough Dirents.", false);
      goto ErrorExit;
    }

    /* Check for error handling, silly buffer size. */

    TestDir.Rewind ();
    NumberRead = TestDir.GetNextDirents (DirEntArray, 10);
    if (NumberRead != -1)
    {
      SetTestResult ("GetNextDirents should have failed with a bad buffer size.",
        false);
      goto ErrorExit;
    }

    /* Check for filling the buffer, enough space only for ".", "..",
    "Directory1" entry structures.  Make it just a bit too small first. */

    TestDir.Rewind ();
    NumberRead = TestDir.GetNextDirents (DirEntArray,
      sizeof (dirent_t) * 3 + 1 + 2 + strlen ("Directory1") - 1);
    if (NumberRead != 2)
    {
      SetTestResult ("GetNextDirents should only have read 2 entries.", false);
      goto ErrorExit;
    }

    TestDir.Rewind ();
    NumberRead = TestDir.GetNextDirents (DirEntArray,
      sizeof (dirent_t) * 3 + 1 + 2 + strlen ("Directory1") - 0);
    if (NumberRead != 3)
    {
      SetTestResult ("GetNextDirents should only have read 3 entries.", false);
      goto ErrorExit;
    }
  }

  /* Test removal of directories. */

  if (rmdir ("/Test/DirTest/Directory1"))
  {
    ErrorCode = errno;
    SetTestResult ("rmdir failed for /Test/DirTest/Directory1", ErrorCode);
    goto ErrorExit;
  }

  NumberRead = TestDir.CountEntries() + 2 /* For . and .. */;
  if (NumberRead != 7)
  {
    sprintf (TempString, "After removing a directory, the number of entries "
      "in it is %d, should be 7.", (int) NumberRead);
    SetTestResult (TempString, false);
    goto ErrorExit;
  }

  /* Should not be able to remove non-existent directories. */

  ErrorCode = 0;
  if (!rmdir ("/Test/DirTest/Directory1") || (ErrorCode = errno) != ENOENT)
  {
    ErrorCode = errno;
    SetTestResult ("rmdir should not work for non-existant directories",
      ErrorCode);
    goto ErrorExit;
  }

  if (rmdir ("/Test/DirTest/Directory2"))
  {
    ErrorCode = errno;
    SetTestResult ("rmdir failed for /Test/DirTest/Directory2", ErrorCode);
    goto ErrorExit;
  }

  ErrorCode = 0;
  if (!rmdir ("/Test/DirTest/Directory3") || (ErrorCode = errno) != ENOTEMPTY)
  {
    ErrorCode = errno;
    SetTestResult ("rmdir should not work for a non-empty directory",
      ErrorCode);
    goto ErrorExit;
  }

  if (rmdir ("/Test/DirTest/Directory3/SubDirInThree"))
  {
    ErrorCode = errno;
    SetTestResult ("rmdir failed for /Test/DirTest/Directory3/SubDirInThree",
      ErrorCode);
    goto ErrorExit;
  }

  if (rmdir ("/Test/DirTest/Directory3"))
  {
    ErrorCode = errno;
    SetTestResult ("rmdir failed for /Test/DirTest/Directory3",
      ErrorCode);
    goto ErrorExit;
  }

  SetTestResult ("mkdir, rmdir, readdir etc. successful.");

ErrorExit:
  if (DirHandle != NULL)
    closedir (DirHandle);
  if (FileHandle0 >= 0)
    close (FileHandle0);
  rmdir ("/Test/DirTest/Directory1");
  rmdir ("/Test/DirTest/Directory2");
  rmdir ("/Test/DirTest/Directory3/SubDirInThree");
  rmdir ("/Test/DirTest/Directory3");
  rmdir ("/Test/DirTest/Directory4/SubDirInFour");
  rmdir ("/Test/DirTest/Directory4");
  rmdir ("/Test/DirTest/Directory5");
  rmdir ("/Test/DirTest/"
    "Testing with a name that's the maximum length....." /* 50 */
    "12345678901234567890123456789012345678901234567890" /* 50 */
    "12345678901234567890123456789012345678901234567890" /* 50 */
    "12345678901234567890123456789012345678901234567890" /* 50 */
    "12345678901234567890123456789012345678901234567890" /* 50 */
    "12345");
  rmdir ("/Test/DirTest");

  return true;
}



/******************************************************************************
 * Test maximum total file size limit system.
 */

class TestMaxTotalSize : public TestAbstract
{
public:
  TestMaxTotalSize (BRect NewFrame)
    : TestAbstract (NewFrame, "TestMaxTotalSize")
    {};

  virtual bool DoTest ();
  virtual const char *GetDescription ()
    { return "Test total size limiter."; };
};


bool TestMaxTotalSize::DoTest ()
{
  int      ErrorCode;
  int      FileHandle = -1;
  int32    OldLimit = 0;
  BFile    TestFile;

  /* Get the existing maximum size.  This will fail on most file systems. */

  OldLimit = ARFS_GetMaxTotalFileSize ("/Test");
  if (OldLimit < 0)
  {
    SetTestResult ("Can't get current maximum total file size, "
      "perhaps this is not an AGMSRAMFileSystem", errno);
    goto ErrorExit;
  }

  if (mkdir ("/Test/a", 0777 /* access mode flags */))
  {
    SetTestResult ("Failed to make /Test/a directory", errno);
    goto ErrorExit;
  }

  /* Should permit writing only five bytes to a file.  Note that we can't use
  BFile for testing since it doesn't permit partial writes. */

  if (ARFS_SetMaxTotalFileSize ("/Test", 5) < 0)
  {
    SetTestResult ("Failed to set maximum total file size to five", errno);
    goto ErrorExit;
  }

  FileHandle = creat ("/Test/a/LimitedFile", (unsigned int) -1 /* perms */);
  if (FileHandle < 0)
  {
    SetTestResult ("Unable to create /Test/a/LimitedFile", errno);
    goto ErrorExit;
  }

  if (write (FileHandle, "Hi", 2) != 2)
  {
    SetTestResult ("Unable to write two bytes", errno);
    goto ErrorExit;
  }

  if ((ErrorCode = write (FileHandle, "There", 5)) != 3)
  {
    SetTestResult ("Expected to write only three bytes, but didn't.", errno);
    goto ErrorExit;
  }

  ErrorCode = 0;
  if (write (FileHandle, "Folks", 5) >= 0 || (ErrorCode = errno) != ENOSPC)
  {
    SetTestResult ("Expected write to fail with ENOSPC", ErrorCode);
    goto ErrorExit;
  }

  /* Truncate the file to 3 bytes.  Leaves 2 bytes for other files. */

  if (ftruncate (FileHandle, 3))
  {
    SetTestResult ("Unable to reduce the size of the file", errno);
    goto ErrorExit;
  }
  close (FileHandle); FileHandle = -1;

  /* Create another file, and set its size to a large value. */

  FileHandle = creat ("/Test/a/AnotherFile", (unsigned int) -1 /* perms */);
  if (FileHandle < 0)
  {
    SetTestResult ("Unable to create /Test/a/AnotherFile", errno);
    goto ErrorExit;
  }

  if (!ftruncate (FileHandle, 10) || (ErrorCode = errno) != ENOSPC)
  {
    SetTestResult ("Enlarging a file past the limit should have failed "
      "with ENOSPC", ErrorCode);
    goto ErrorExit;
  }

  /* Try writing, only 2 bytes should get in. */

  if ((ErrorCode = write (FileHandle, "Two", 3)) != 2)
  {
    SetTestResult ("Expected to write only two bytes, but didn't.", errno);
    goto ErrorExit;
  }

  /* Delete the first file, should get another 3 bytes in. */

  if (unlink ("/Test/a/LimitedFile"))
  {
    SetTestResult ("Unlink of /Test/a/LimitedFile failed", errno);
    goto ErrorExit;
  }

  if ((ErrorCode = write (FileHandle, "More", 4)) != 3)
  {
    SetTestResult ("Expected to write again three bytes, but didn't.", errno);
    goto ErrorExit;
  }

  SetTestResult ("Maximum size tests successful.");

ErrorExit:
  if (FileHandle >= 0)
    close (FileHandle);
  RecursivelyDelete ("/Test/a");
  if (OldLimit > 0)
    ARFS_SetMaxTotalFileSize ("/Test", OldLimit);

  return true;
}



/******************************************************************************
 * Mount the test file system.  Includes making the mount point directory.
 * There is an option to do a read-only mount and another to specify the total
 * file size limit.
 */

class TestMount : public TestAbstract
{
public:
  TestMount (BRect NewFrame);
  virtual ~TestMount ();
  virtual void MessageReceived (BMessage *MessagePntr);
  virtual float AddCustomControls (float X);
  virtual void PreTestPreparations ();
  virtual bool DoTest ();

  virtual const char *GetDescription ()
    { return "Mount the AGMS RAM File System on /Test."; };

private:
  void GetCheckAndSetTextBoxes ();

  int           m_MaxTotalFileSize;
  BTextControl *m_MaxTotalFileSizeTextBoxPntr;
  bool          m_ReadOnly;
  BCheckBox    *m_ReadOnlyCheckBoxPntr;
};


TestMount::TestMount (BRect NewFrame)
: TestAbstract (NewFrame, "TestMount"),
  m_MaxTotalFileSize (100000000),
  m_MaxTotalFileSizeTextBoxPntr (NULL),
  m_ReadOnly (false),
  m_ReadOnlyCheckBoxPntr (NULL)
{
  /* Make sure there is space for our "read-only" checkbox and maximum size
  text box. */

  if (m_GUIHeight < g_CheckBoxHeight)
    m_GUIHeight = g_CheckBoxHeight;
  if (m_GUIHeight < g_TextBoxHeight)
    m_GUIHeight = g_TextBoxHeight;
}


TestMount::~TestMount ()
{
  /* Child views (controls) are automatically destroyed and deallocated by
  the parent destructor.  So there isn't much to do here. */

  m_MaxTotalFileSizeTextBoxPntr = NULL;
  m_ReadOnlyCheckBoxPntr = NULL;
}


void TestMount::MessageReceived (BMessage *MessagePntr)
{
  switch (MessagePntr->what)
  {
    case MSG_MAX_TOTAL_FILE_SIZE_MOUNT:
      GetCheckAndSetTextBoxes ();
      break;

    case MSG_READ_ONLY_MOUNT:
      m_ReadOnly = m_ReadOnlyCheckBoxPntr->Value ();
      break;

    default:
      TestAbstract::MessageReceived (MessagePntr);
      break;
  }
}


float TestMount::AddCustomControls (float X)
{
  float Height;
  float Margin;
  BRect TempRect;
  char  TempString [80];
  float Width;

  /* Add the checkbox used to specify read-only mode. */

  Margin = ceilf ((m_GUIHeight - g_CheckBoxHeight) / 2);
  TempRect = Bounds ();
  TempRect.top += Margin;
  TempRect.bottom = TempRect.top + g_CheckBoxHeight;
  TempRect.left = X;

  m_ReadOnlyCheckBoxPntr = new BCheckBox (TempRect,
    "ReadOnlyBox",
    "Read-Only",
    new BMessage (MSG_READ_ONLY_MOUNT));
  if (m_ReadOnlyCheckBoxPntr == NULL) goto ErrorExit;
  AddChild (m_ReadOnlyCheckBoxPntr);
  m_ReadOnlyCheckBoxPntr->SetTarget (this);
  m_ReadOnlyCheckBoxPntr->SetValue (m_ReadOnly);
  m_ReadOnlyCheckBoxPntr->ResizeToPreferred ();
  m_ReadOnlyCheckBoxPntr->GetPreferredSize (&Width, &Height);
  X += ceilf (Width);

  /* Add the text control used to specify the maximum RAM disk size. */

  Margin = ceilf ((m_GUIHeight - g_TextBoxHeight) / 2);
  TempRect = Bounds ();
  TempRect.top += Margin;
  TempRect.bottom = TempRect.top + g_TextBoxHeight;
  TempRect.left = X;
  TempRect.right = X + 120;

  sprintf (TempString, "%d", m_MaxTotalFileSize);
  m_MaxTotalFileSizeTextBoxPntr = new BTextControl (TempRect,
    "Max Total File Size Text Control",
    "Max size:" /* label, no trailing space */,
    TempString,
    new BMessage (MSG_MAX_TOTAL_FILE_SIZE_MOUNT),
    B_FOLLOW_LEFT | B_FOLLOW_TOP,
    B_WILL_DRAW | B_FRAME_EVENTS | B_FULL_UPDATE_ON_RESIZE | B_NAVIGABLE);
  if (m_MaxTotalFileSizeTextBoxPntr == NULL) goto ErrorExit;
  m_MaxTotalFileSizeTextBoxPntr->SetDivider (be_plain_font->StringWidth (
    "Max size: " /* includes a trailing space for spacing */));
  AddChild (m_MaxTotalFileSizeTextBoxPntr);
  m_MaxTotalFileSizeTextBoxPntr->SetTarget (this);
  X += ceilf (m_MaxTotalFileSizeTextBoxPntr->Bounds().Width());

  return X;

ErrorExit:
  return -1.0;
}


void TestMount::PreTestPreparations ()
{
  GetCheckAndSetTextBoxes ();
  TestAbstract::PreTestPreparations ();
}


bool TestMount::DoTest ()
{
  int   ErrorCode;
  char *MountDirectory = "/Test";
  char  Parameters [1024];

  ErrorCode = mkdir (MountDirectory, 0777 /* access mode flags */);
  if (ErrorCode < 0 && ErrorCode != EEXIST)
  {
    SetTestResult ("Can't Make /Test directory", ErrorCode);
    return false;
  }

  sprintf (Parameters,
    "MountDir=%s\tMaxVolumeSize=%d\tDumpFilePath=/var/tmp/Junk.dump\t"
    "DumpCompressionMethod=Blort",
    MountDirectory, m_MaxTotalFileSize);

  ErrorCode = mount ("ARFSv0", MountDirectory,
    NULL /* Don't specify a device, needs to be NULL, not an empty string! */,
    m_ReadOnly ? B_MOUNT_READ_ONLY : 0 /* flags */,
    Parameters, strlen (Parameters) + 1);

  if (ErrorCode < 0)
  {
    SetTestResult ("Mount failed", ErrorCode);
    return false;
  }

  SetTestResult ("Mounted on /Test.");
  g_FakeMount = false;
  return true;
}


void TestMount::GetCheckAndSetTextBoxes ()
{
  char TempString [80];

  if (m_MaxTotalFileSizeTextBoxPntr != NULL)
  {
    m_MaxTotalFileSize = atoi (m_MaxTotalFileSizeTextBoxPntr->Text ());
    if (m_MaxTotalFileSize < 0)
      m_MaxTotalFileSize = 0;
    sprintf (TempString, "%d", m_MaxTotalFileSize);
    m_MaxTotalFileSizeTextBoxPntr->SetText (TempString);
  }
}



/******************************************************************************
 * The TestNotification class tests the change notification system.
 */

class TestNotification : public TestAbstract
{
public:
  TestNotification (BRect NewFrame)
    : TestAbstract (NewFrame, "TestNotification")
    {};

  virtual bool DoTest ();
  virtual const char *GetDescription ()
    { return "Test Change Notification."; };
};


bool TestNotification::DoTest ()
{
  int                 ErrorCode;
  BFile               File;
  int32               LocalInt32;
  int64               LocalInt64;
  const char         *LocalStringPntr;
  BMessage           *MsgPntr;
  struct node_ref     NodeRefDirA;
  struct node_ref     NodeRefDirB;
  struct node_ref     NodeRefOverwrittenFile;
  struct node_ref     NodeRefTheFile;
  NotificationLooper *NotificationLooperPntr = NULL;
  struct stat         StatInfo;
  const char         *WriteString = "Notification test data.\n";

  /* Make the directory structure /Test/a/b/c. */

  if ((ErrorCode = create_directory ("/Test/dirs/a", 0777)) != 0)
  {
    SetTestResult ("Failed to make /Test/dirs/a directories", ErrorCode);
    goto ErrorExit;
  }
  if ((ErrorCode = create_directory ("/Test/dirs/b", 0777)) != 0)
  {
    SetTestResult ("Failed to make /Test/dirs/b directories", ErrorCode);
    goto ErrorExit;
  }

  /* Start up the notifier monitoring BLooper. */

  NotificationLooperPntr = new NotificationLooper;
  ErrorCode = 0;
  if (NotificationLooperPntr == NULL ||
  (ErrorCode = NotificationLooperPntr->Run ()) < 0)
  {
    SetTestResult ("Failed to start up the looper for monitoring "
      "notification messages", ErrorCode);
    goto ErrorExit;
  }

  /* Monitor the "a" directory, then make a file /Test/dirs/a/TheFile.  Should
  trigger a creation notification. */

  if (stat ("/Test/dirs/a", &StatInfo)) goto ErrorExit;
  NodeRefDirA.device = StatInfo.st_dev;
  NodeRefDirA.node = StatInfo.st_ino;

  if (stat ("/Test/dirs/b", &StatInfo)) goto ErrorExit;
  NodeRefDirB.device = StatInfo.st_dev;
  NodeRefDirB.node = StatInfo.st_ino;

  ErrorCode = watch_node (&NodeRefDirA, B_WATCH_ALL,
    NotificationLooperPntr, NotificationLooperPntr);
  if (ErrorCode != 0)
  {
    SetTestResult ("watch_node failed for /Test/dirs/a", ErrorCode);
    goto ErrorExit;
  }

  ErrorCode =
    File.SetTo ("/Test/dirs/a/TheFile", B_READ_WRITE | B_CREATE_FILE);
  if (ErrorCode != 0)
  {
    SetTestResult ("Unable to create /Test/dirs/a/TheFile", ErrorCode);
    goto ErrorExit;
  }
  File.GetNodeRef (&NodeRefTheFile);

  if (NotificationLooperPntr->WaitForOpCodes (1) != 1 ||
  NotificationLooperPntr->HasOverflowed () || (MsgPntr =
  NotificationLooperPntr->GetMessageForOpCode (B_ENTRY_CREATED)) == NULL)
  {
    SetTestResult ("Should just get a single B_ENTRY_CREATED notification "
      "message for the creation of a file.", false);
    goto ErrorExit;
  }

  /* Verify the contents of the notification message. */

  if (MsgPntr->FindInt32 ("device", &LocalInt32) != B_OK ||
  LocalInt32 != NodeRefDirA.device)
  {
    SetTestResult ("B_ENTRY_CREATED notification message has an incorrect "
      "device field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt64 ("directory", &LocalInt64) != B_OK ||
  LocalInt64 != NodeRefDirA.node)
  {
    SetTestResult ("B_ENTRY_CREATED notification message has an incorrect "
      "directory field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt64 ("node", &LocalInt64) != B_OK ||
  LocalInt64 != NodeRefTheFile.node)
  {
    SetTestResult ("B_ENTRY_CREATED notification message has an incorrect "
      "node field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindString ("name", &LocalStringPntr) != B_OK ||
  strcmp (LocalStringPntr, "TheFile") != 0)
  {
    SetTestResult ("B_ENTRY_CREATED notification message has an incorrect "
      "name field.", false);
    goto ErrorExit;
  }

  /* Write some data to the file, this will make it longer and possibly change
  the STAT info (file size to be particular).  Some file systems will report
  the change immediately, others when the file is closed, and mine does it once
  per second (so that the Tracker GUI display shows the current size). */

  stop_watching (NotificationLooperPntr, NotificationLooperPntr);
  ErrorCode = watch_node (&NodeRefTheFile, B_WATCH_ALL,
    NotificationLooperPntr, NotificationLooperPntr);
  if (ErrorCode != 0)
  {
    SetTestResult ("watch_node failed for /Test/dirs/a/TheFile", ErrorCode);
    goto ErrorExit;
  }
  NotificationLooperPntr->Clear ();

  if (File.Write (WriteString, strlen (WriteString)) !=
  (int) strlen (WriteString))
  {
    SetTestResult ("Error while writing data to /Test/dirs/a/TheFile", errno);
    goto ErrorExit;
  }

  if (NotificationLooperPntr->GetMessageForOpCode (B_STAT_CHANGED) != NULL)
    SetTestResult ("Excessive!  This file system sends B_STAT_CHANGED "
    "whenever the file size changes.", false);

  snooze (2000000); /* Wait 2 seconds, testing delayed write notifications. */
  NotificationLooperPntr->Clear ();

  if (File.Write (WriteString, strlen (WriteString)) !=
  (int) strlen (WriteString))
  {
    SetTestResult ("Error while writing data to /Test/dirs/a/TheFile", errno);
    goto ErrorExit;
  }

  snooze (1100000);

  if (NotificationLooperPntr->GetMessageForOpCode (B_STAT_CHANGED) != NULL)
    SetTestResult ("Good!  This file system sends B_STAT_CHANGED even "
    "while the file is open and is being written to.", false);
  else
    SetTestResult ("Not very good, cheap in fact.  This file system does not "
      "send B_STAT_CHANGED at all while the file is open and being written "
      "to.", false);

  File.Unset (); /* Close the file.  Lazy file systems should notify now. */

  if (NotificationLooperPntr->WaitForOpCodes (1) != 1 ||
  (MsgPntr = NotificationLooperPntr->GetMessageForOpCode (B_STAT_CHANGED))
  == NULL)
  {
    SetTestResult ("Should just get B_STAT_CHANGED notification "
      "message(s) after writing to a file and then closing it.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt32 ("device", &LocalInt32) != B_OK ||
  LocalInt32 != NodeRefTheFile.device)
  {
    SetTestResult ("B_STAT_CHANGED notification message has an incorrect "
      "device field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt64 ("node", &LocalInt64) != B_OK ||
  LocalInt64 != NodeRefTheFile.node)
  {
    SetTestResult ("B_STAT_CHANGED notification message has an incorrect "
      "node field.", false);
    goto ErrorExit;
  }

  /* Try changing status information.  Should trigger a B_STAT_CHANGED. */

  ErrorCode = File.SetTo ("/Test/dirs/a/TheFile", B_READ_WRITE);
  if (ErrorCode != 0)
  {
    SetTestResult ("Unable to reopen /Test/dirs/a/TheFile", ErrorCode);
    goto ErrorExit;
  }

  NotificationLooperPntr->Clear ();

  ErrorCode = File.SetPermissions (0642);
  if (ErrorCode != 0)
  {
    SetTestResult ("SetPermissions failed", ErrorCode);
    goto ErrorExit;
  }

  if (NotificationLooperPntr->WaitForOpCodes (1) != 1 ||
  NotificationLooperPntr->HasOverflowed () || (MsgPntr =
  NotificationLooperPntr->GetMessageForOpCode (B_STAT_CHANGED)) == NULL)
  {
    SetTestResult ("Should just get a single B_STAT_CHANGED notification "
      "message for changing permissions.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt32 ("device", &LocalInt32) != B_OK ||
  LocalInt32 != NodeRefTheFile.device)
  {
    SetTestResult ("B_STAT_CHANGED notification message has an incorrect "
      "device field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt64 ("node", &LocalInt64) != B_OK ||
  LocalInt64 != NodeRefTheFile.node)
  {
    SetTestResult ("B_STAT_CHANGED notification message has an incorrect "
      "node field.", false);
    goto ErrorExit;
  }

  NotificationLooperPntr->Clear ();

  ErrorCode = File.SetSize (54321);
  if (ErrorCode != 0)
  {
    SetTestResult ("SetSize failed", ErrorCode);
    goto ErrorExit;
  }

  if (NotificationLooperPntr->WaitForOpCodes (1) != 1 ||
  NotificationLooperPntr->HasOverflowed () || (MsgPntr =
  NotificationLooperPntr->GetMessageForOpCode (B_STAT_CHANGED)) == NULL)
  {
    SetTestResult ("Should just get a single B_STAT_CHANGED notification "
      "message for changing file size.", false);
    goto ErrorExit;
  }

  /* B_STAT_CHANGED quirks - try setting the size to the same size as it
  currently has.  Could also test for failed changes (permission denied etc)
  not triggering a notification message. */

  NotificationLooperPntr->Clear ();

  ErrorCode = File.SetSize (54321);
  if (ErrorCode != 0)
  {
    SetTestResult ("SetSize failed later", ErrorCode);
    goto ErrorExit;
  }

  if (NotificationLooperPntr->WaitForOpCodes (1) != 1 ||
  NotificationLooperPntr->HasOverflowed () || (MsgPntr =
  NotificationLooperPntr->GetMessageForOpCode (B_STAT_CHANGED)) == NULL)
  {
    SetTestResult ("Expecting one notification for SetSize when using "
      "the existing size (even if size isn't changed, should notify).",
      false);
    goto ErrorExit;
  }

  /* Check the B_ATTR_CHANGED notification for writing an attribute. */

  NotificationLooperPntr->Clear ();

  if (File.WriteAttr ("AttributeName", B_STRING_TYPE, 0,
  WriteString, strlen (WriteString)) != (int) strlen (WriteString))
  {
    SetTestResult ("WriteAttr failed", errno);
    goto ErrorExit;
  }

  if (NotificationLooperPntr->WaitForOpCodes (1) != 1 ||
  NotificationLooperPntr->HasOverflowed () || (MsgPntr =
  NotificationLooperPntr->GetMessageForOpCode (B_ATTR_CHANGED)) == NULL)
  {
    SetTestResult ("Should just get a single B_ATTR_CHANGED notification "
      "when writing an attribute.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt32 ("device", &LocalInt32) != B_OK ||
  LocalInt32 != NodeRefTheFile.device)
  {
    SetTestResult ("B_ATTR_CHANGED notification message has an incorrect "
      "device field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt64 ("node", &LocalInt64) != B_OK ||
  LocalInt64 != NodeRefTheFile.node)
  {
    SetTestResult ("B_ATTR_CHANGED notification message has an incorrect "
      "node field.", false);
    goto ErrorExit;
  }

  /* Check the B_ATTR_CHANGED notification for deleting an attribute. */

  NotificationLooperPntr->Clear ();

  if (File.RemoveAttr ("AttributeName") != B_OK)
  {
    SetTestResult ("RemoveAttr failed", errno);
    goto ErrorExit;
  }

  if (NotificationLooperPntr->WaitForOpCodes (1) != 1 ||
  NotificationLooperPntr->HasOverflowed () || (MsgPntr =
  NotificationLooperPntr->GetMessageForOpCode (B_ATTR_CHANGED)) == NULL)
  {
    SetTestResult ("Should just get a single B_ATTR_CHANGED notification "
      "when removing an attribute.", false);
    goto ErrorExit;
  }

  /* Check the B_ATTR_CHANGED notification for renaming an attribute. */

  if (File.WriteAttr ("FirstAttribute", B_STRING_TYPE, 0,
  WriteString, strlen (WriteString)) != (int) strlen (WriteString))
  {
    SetTestResult ("WriteAttr failed for FirstAttribute", errno);
    goto ErrorExit;
  }
  if (File.WriteAttr ("SecondAttribute", B_STRING_TYPE, 0,
  WriteString, strlen (WriteString)) != (int) strlen (WriteString))
  {
    SetTestResult ("WriteAttr failed for SecondAttribute", errno);
    goto ErrorExit;
  }

  NotificationLooperPntr->Clear ();

  if (File.RenameAttr ("FirstAttribute", "SecondAttribute") != B_OK)
  {
    SetTestResult ("RenameAttr failed", errno);
    if (!g_FakeMount) /* BFS doesn't have renaming. */
      goto ErrorExit;
  }

  if (NotificationLooperPntr->WaitForOpCodes (1) != 1 ||
  NotificationLooperPntr->HasOverflowed () || (MsgPntr =
  NotificationLooperPntr->GetMessageForOpCode (B_ATTR_CHANGED)) == NULL)
  {
    SetTestResult ("Should just get a single B_ATTR_CHANGED notification "
      "when renaming an attribute.", false);
    if (!g_FakeMount) /* BFS doesn't have renaming. */
      goto ErrorExit;
  }

  /* Check the B_ENTRY_MOVED notification.  First from the source directory. */

  File.Unset ();
  stop_watching (NotificationLooperPntr, NotificationLooperPntr);
  watch_node (&NodeRefDirA, B_WATCH_ALL,
    NotificationLooperPntr, NotificationLooperPntr);
  NotificationLooperPntr->Clear ();

  if (rename ("/Test/dirs/a/TheFile", "/Test/dirs/b/AnotherFile"))
  {
    SetTestResult ("Renaming failed", errno);
    goto ErrorExit;
  }

  if (NotificationLooperPntr->WaitForOpCodes (1) != 1 ||
  NotificationLooperPntr->HasOverflowed () || (MsgPntr =
  NotificationLooperPntr->GetMessageForOpCode (B_ENTRY_MOVED)) == NULL)
  {
    SetTestResult ("Should just get a single B_ENTRY_MOVED notification "
      "message from the source directory.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt32 ("device", &LocalInt32) != B_OK ||
  LocalInt32 != NodeRefDirA.device)
  {
    SetTestResult ("B_ENTRY_MOVED notification message has an incorrect "
      "device field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt64 ("from directory", &LocalInt64) != B_OK ||
  LocalInt64 != NodeRefDirA.node)
  {
    SetTestResult ("B_ENTRY_MOVED notification message has an incorrect "
      "from directory field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt64 ("to directory", &LocalInt64) != B_OK ||
  LocalInt64 != NodeRefDirB.node)
  {
    SetTestResult ("B_ENTRY_MOVED notification message has an incorrect "
      "to directory field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt64 ("node", &LocalInt64) != B_OK ||
  LocalInt64 != NodeRefTheFile.node)
  {
    SetTestResult ("B_ENTRY_MOVED notification message has an incorrect "
      "node field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindString ("name", &LocalStringPntr) != B_OK ||
  strcmp (LocalStringPntr, "AnotherFile") != 0)
  {
    SetTestResult ("B_ENTRY_MOVED notification message has an incorrect "
      "name field.", false);
    goto ErrorExit;
  }

  /* Check the B_ENTRY_MOVED destination directory notification. */

  NotificationLooperPntr->Clear ();

  if (rename ("/Test/dirs/b/AnotherFile", "/Test/dirs/a/FinalFile"))
  {
    SetTestResult ("Renaming failed", errno);
    goto ErrorExit;
  }

  if (NotificationLooperPntr->WaitForOpCodes (1) != 1 ||
  NotificationLooperPntr->HasOverflowed () || (MsgPntr =
  NotificationLooperPntr->GetMessageForOpCode (B_ENTRY_MOVED)) == NULL)
  {
    SetTestResult ("Should just get a single B_ENTRY_MOVED notification "
      "message from the destination directory.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt32 ("device", &LocalInt32) != B_OK ||
  LocalInt32 != NodeRefDirA.device)
  {
    SetTestResult ("B_ENTRY_MOVED notification message has an incorrect "
      "device field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt64 ("from directory", &LocalInt64) != B_OK ||
  LocalInt64 != NodeRefDirB.node)
  {
    SetTestResult ("B_ENTRY_MOVED notification message has an incorrect "
      "from directory field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt64 ("to directory", &LocalInt64) != B_OK ||
  LocalInt64 != NodeRefDirA.node)
  {
    SetTestResult ("B_ENTRY_MOVED notification message has an incorrect "
      "to directory field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt64 ("node", &LocalInt64) != B_OK ||
  LocalInt64 != NodeRefTheFile.node)
  {
    SetTestResult ("B_ENTRY_MOVED notification message has an incorrect "
      "node field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindString ("name", &LocalStringPntr) != B_OK ||
  strcmp (LocalStringPntr, "FinalFile") != 0)
  {
    SetTestResult ("B_ENTRY_MOVED notification message has an incorrect "
      "name field.", false);
    goto ErrorExit;
  }

  /* B_ENTRY_MOVED for the file being renamed, and the file being overwritten
  should also get a B_ENTRY_REMOVED. */

  ErrorCode =
    File.SetTo ("/Test/dirs/b/OverWrittenFile", B_READ_WRITE | B_CREATE_FILE);
  if (ErrorCode != 0)
  {
    SetTestResult ("Unable to create /Test/dirs/b/OverWrittenFile", ErrorCode);
    goto ErrorExit;
  }
  File.GetNodeRef (&NodeRefOverwrittenFile);
  File.Unset ();

  stop_watching (NotificationLooperPntr, NotificationLooperPntr);
  watch_node (&NodeRefTheFile, B_WATCH_ALL,
    NotificationLooperPntr, NotificationLooperPntr);
  watch_node (&NodeRefOverwrittenFile, B_WATCH_ALL,
    NotificationLooperPntr, NotificationLooperPntr);
  NotificationLooperPntr->Clear ();

  if (rename ("/Test/dirs/a/FinalFile", "/Test/dirs/b/OverWrittenFile"))
  {
    SetTestResult ("Renaming failed", errno);
    goto ErrorExit;
  }

  if (NotificationLooperPntr->WaitForOpCodes (2) != 2 ||
  NotificationLooperPntr->HasOverflowed () || (MsgPntr =
  NotificationLooperPntr->GetMessageForOpCode (B_ENTRY_MOVED)) == NULL ||
  NotificationLooperPntr->GetMessageForOpCode (B_ENTRY_REMOVED) == NULL)
  {
    SetTestResult ("Renaming a file over another should get both a "
      "B_ENTRY_MOVED notification and a B_ENTRY_REMOVED one too.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt32 ("device", &LocalInt32) != B_OK ||
  LocalInt32 != NodeRefDirA.device)
  {
    SetTestResult ("B_ENTRY_MOVED notification message has an incorrect "
      "device field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt64 ("from directory", &LocalInt64) != B_OK ||
  LocalInt64 != NodeRefDirA.node)
  {
    SetTestResult ("B_ENTRY_MOVED notification message has an incorrect "
      "from directory field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt64 ("to directory", &LocalInt64) != B_OK ||
  LocalInt64 != NodeRefDirB.node)
  {
    SetTestResult ("B_ENTRY_MOVED notification message has an incorrect "
      "to directory field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt64 ("node", &LocalInt64) != B_OK ||
  LocalInt64 != NodeRefTheFile.node)
  {
    SetTestResult ("B_ENTRY_MOVED notification message has an incorrect "
      "node field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindString ("name", &LocalStringPntr) != B_OK ||
  strcmp (LocalStringPntr, "OverWrittenFile") != 0)
  {
    SetTestResult ("B_ENTRY_MOVED notification message has an incorrect "
      "name field.", false);
    goto ErrorExit;
  }

  MsgPntr = NotificationLooperPntr->GetMessageForOpCode (B_ENTRY_REMOVED);

  if (MsgPntr->FindInt32 ("device", &LocalInt32) != B_OK ||
  LocalInt32 != NodeRefDirB.device)
  {
    SetTestResult ("B_ENTRY_REMOVED notification message has an incorrect "
      "device field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt64 ("directory", &LocalInt64) != B_OK ||
  LocalInt64 != NodeRefDirB.node)
  {
    SetTestResult ("B_ENTRY_REMOVED notification message has an incorrect "
      "directory field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt64 ("node", &LocalInt64) != B_OK ||
  LocalInt64 != NodeRefOverwrittenFile.node)
  {
    SetTestResult ("B_ENTRY_REMOVED notification message has an incorrect "
      "node field.", false);
    goto ErrorExit;
  }

  /* Check the B_ENTRY_REMOVED notification for deleting a file, should get one
  from the directory and another from monitoring the file. */

  stop_watching (NotificationLooperPntr, NotificationLooperPntr);
  watch_node (&NodeRefTheFile, B_WATCH_ALL,
    NotificationLooperPntr, NotificationLooperPntr);
  watch_node (&NodeRefDirB, B_WATCH_ALL,
    NotificationLooperPntr, NotificationLooperPntr);
  NotificationLooperPntr->Clear ();

  if (unlink ("/Test/dirs/b/OverWrittenFile"))
  {
    SetTestResult ("Unable to delete /Test/dirs/b/OverWrittenFile", errno);
    goto ErrorExit;
  }

  if (NotificationLooperPntr->WaitForOpCodes (1) != 1 ||
  !NotificationLooperPntr->HasOverflowed () || (MsgPntr =
  NotificationLooperPntr->GetMessageForOpCode (B_ENTRY_REMOVED)) == NULL)
  {
    SetTestResult ("Didn't get two B_ENTRY_REMOVED notifications when "
      "removing a file from a directory.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt32 ("device", &LocalInt32) != B_OK ||
  LocalInt32 != NodeRefDirB.device)
  {
    SetTestResult ("B_ENTRY_REMOVED notification message has an incorrect "
      "device field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt64 ("directory", &LocalInt64) != B_OK ||
  LocalInt64 != NodeRefDirB.node)
  {
    SetTestResult ("B_ENTRY_REMOVED notification message has an incorrect "
      "directory field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt64 ("node", &LocalInt64) != B_OK ||
  LocalInt64 != NodeRefTheFile.node)
  {
    SetTestResult ("B_ENTRY_REMOVED notification message has an incorrect "
      "node field.", false);
    goto ErrorExit;
  }

  SetTestResult ("Change Notification test OK.");

ErrorExit:
  stop_watching (NotificationLooperPntr, NotificationLooperPntr);
  if (NotificationLooperPntr != NULL)
    NotificationLooperPntr->PostMessage (B_QUIT_REQUESTED);
  File.Unset ();
  RecursivelyDelete ("/Test/dirs");
  return true;
}



/******************************************************************************
 * Test file and directory access permissions.
 */

class TestPermissions : public TestAbstract
{
public:
  TestPermissions (BRect NewFrame)
    : TestAbstract (NewFrame, "TestPermissions")
    {};

  virtual bool DoTest ();
  virtual const char *GetDescription ()
    { return "Test file and directory access permissions."; };
};



bool TestPermissions::DoTest ()
{
  int         ErrorCode = 0;
  int         FileHandle0 = -1;
  int         FileHandle1 = -1;
  gid_t       OldGroupID = getegid ();
  mode_t      OldUMask = umask (0 /* specifies rwxrwxrwx bits to disable */);
  uid_t       OldUserID = geteuid ();
  BDirectory  PermsDir;
  struct stat StatBuf;
  gid_t       TempGroup;
  BFile       TestFile;

  if (OldUserID != 0)
  {
    SetTestResult ("Not running as root user, need to to be able to pretend "
      "to be other users.", false);
    goto ErrorExit;
  }

  /* Make a directory with write access for all, owned by root, otherwise the
  rest of our tests won't work in fake mode.  Also set it to group 1234, so
  that new objects created in it have that group. */

  if (mkdir ("/Test/DirPermTest", 0777))
  {
    SetTestResult ("mkdir failed for /Test/DirPermTest", errno);
    goto ErrorExit;
  }
  if ((ErrorCode = PermsDir.SetTo ("/Test/DirPermTest")) != 0)
  {
    SetTestResult ("SetTo failed for /Test/DirPermTest", ErrorCode);
    goto ErrorExit;
  }
  if ((ErrorCode = PermsDir.SetGroup (1234)) != 0)
  {
    SetTestResult ("SetGroup failed for /Test/DirPermTest", ErrorCode);
    goto ErrorExit;
  }
  PermsDir.Unset ();

  memset (&StatBuf, 0, sizeof (StatBuf));
  if (stat ("/Test/DirPermTest", &StatBuf) != 0)
  {
    SetTestResult ("Stat of /Test/DirPermTest failed", errno);
    goto ErrorExit;
  }
  if (StatBuf.st_gid != 1234)
  {
    SetTestResult ("New directory /Test/DirPermTest has wrong group.", false);
    goto ErrorExit;
  }

  /* Check directory permission flags.  First pretend to be some user other
  than the root user, which is the default user in BeOS.  Otherwise a lot of
  these permission tests won't be very meaningful. */

  setuid (5); /* Fortunately, as root we can change to be anyone else. */
  setgid (5555);

  if (mkdir ("/Test/DirPermTest/FiveOwnerDir", S_IRWXU))
  {
    SetTestResult ("mkdir failed for /Test/DirPermTest/FiveOwnerDir", errno);
    goto ErrorExit;
  }

  memset (&StatBuf, 0, sizeof (StatBuf));
  if (stat ("/Test/DirPermTest/FiveOwnerDir", &StatBuf) != 0)
  {
    SetTestResult ("Stat of /Test/DirPermTest/FiveOwnerDir failed", errno);
    goto ErrorExit;
  }
  if (StatBuf.st_gid != 1234)
  {
    SetTestResult ("New directory /Test/DirPermTest/FiveOwnerDir has "
      "wrong group.", false);
    if (!g_FakeMount) /* BFS gets this wrong. */
      goto ErrorExit;
  }
  if (StatBuf.st_uid != 5)
  {
    SetTestResult ("New directory /Test/DirPermTest/FiveOwnerDir has "
      "wrong owner.", false);
    goto ErrorExit;
  }

  /* Some file permission and owner tests. */

  ErrorCode = TestFile.SetTo ("/Test/DirPermTest/FiveOwnerFile",
    B_READ_WRITE | B_CREATE_FILE);
  if (ErrorCode != 0)
  {
    SetTestResult ("Failed to create /Test/DirPermTest/FiveOwnerFile",
      ErrorCode);
    goto ErrorExit;
  }
  TestFile.Write ("This is a test.", 15);

  TempGroup = 5678;
  if (TestFile.GetGroup (&TempGroup) != 0 || TempGroup != 1234)
  {
    SetTestResult ("Group of /Test/DirPermTest/FiveOwnerFile is wrong.",
      false);
    if (!g_FakeMount) /* BFS gets this wrong. */
      goto ErrorExit;
  }

  ErrorCode = TestFile.SetPermissions (0644);
  if (ErrorCode != 0)
  {
    SetTestResult ("Unable to set permission flags of "
      "/Test/DirPermTest/FiveOwnerFile", ErrorCode);
    goto ErrorExit;
  }

  TestFile.Unset ();

  /* Testing new file group - group should be the same as its parent directory,
  not this process's group ID. */

  memset (&StatBuf, 0, sizeof (StatBuf));
  if (stat ("/Test/DirPermTest/FiveOwnerFile", &StatBuf) != 0)
  {
    SetTestResult ("Stat of /Test/DirPermTest/FiveOwnerFile failed", errno);
    goto ErrorExit;
  }
  if (StatBuf.st_uid != 5)
  {
    SetTestResult ("New file /Test/DirPermTest/FiveOwnerFile has wrong owner.",
      false);
    goto ErrorExit;
  }
  if (StatBuf.st_gid != 1234)
  {
    SetTestResult ("New file /Test/DirPermTest/FiveOwnerFile has wrong group.",
      false);
    if (!g_FakeMount) /* BFS gets this wrong. */
      goto ErrorExit;
  }
  if (StatBuf.st_size != 15)
  {
    SetTestResult ("New file /Test/DirPermTest/FiveOwnerFile has wrong size.",
      false);
    goto ErrorExit;
  }
  if ((StatBuf.st_mode & S_IFMT) != S_IFREG ||
  (StatBuf.st_mode & S_IUMSK) != 0644)
  {
    SetTestResult ("Wrong file mode flags for "
      "/Test/DirPermTest/FiveOwnerFile.", false);
    goto ErrorExit;
  }

  /* Try to create or read as a different user.  The directory we are testing
  in should have rwx only for user 5, and as we are some other user, we should
  be rejected. */

  setuid (6);

  ErrorCode = 0;
  if (mkdir ("/Test/DirPermTest/FiveOwnerDir/ShouldNotExist", S_IRWXU) == 0 ||
  (ErrorCode = errno) != EACCES)
  {
    SetTestResult ("mkdir should fail with EACCES for "
      "/Test/DirPermTest/FiveOwnerDir/ShouldNotExist", ErrorCode);
    rmdir ("/Test/DirPermTest/FiveOwnerDir/ShouldNotExist");
    goto ErrorExit;
  }

  /* Also try writing to a file which only some other owner has write
  permission, we should only have read permission. */

  ErrorCode = 0;
  FileHandle1 = open ("/Test/DirPermTest/FiveOwnerFile", O_RDWR, -1);
  if (FileHandle1 >= 0 || (ErrorCode = errno) != EACCES)
  {
    SetTestResult ("Should get EACCES when opening file "
      "/Test/DirPermTest/FiveOwnerFile", ErrorCode);
    goto ErrorExit;
  }

  FileHandle1 = open ("/Test/DirPermTest/FiveOwnerFile", O_RDONLY, -1);
  if (FileHandle1 < 0)
  {
    SetTestResult ("Failed to do legal read only open", errno);
    goto ErrorExit;
  }
  close (FileHandle1); FileHandle1 = -1;

  ErrorCode = TestFile.SetTo ("/Test/DirPermTest/FiveOwnerFile", B_READ_WRITE);
  if (ErrorCode == 0)
  {
    SetTestResult ("No error when opening a BFile for writing when "
      "write permission is lacking (should be EACCES).  Will try actually "
      "writing.", false);
    ErrorCode = TestFile.Write ("This is a test.", 15);
    if (ErrorCode == 15)
    {
      SetTestResult ("Was able to write to a read only file.)", false);
      goto ErrorExit;
    }
    SetTestResult ("Belatedly OK, didn't actually write to the read only file",
      ErrorCode);
  }
  else if (ErrorCode != EACCES)
  {
    SetTestResult ("Should get EACCES when opening a BFile for writing when "
      "write permission is lacking", ErrorCode);
  }
  TestFile.Unset ();

  ErrorCode = TestFile.SetTo ("/Test/DirPermTest/FiveOwnerFile", B_READ_ONLY);
  if (ErrorCode != 0)
  {
    SetTestResult ("Should be able to open a BFile for reading when "
      "permission allows just read access", ErrorCode);
    goto ErrorExit;
  }
  TestFile.Unset ();

  SetTestResult ("Permissions test OK.");

ErrorExit:
  TestFile.Unset ();
  if (setuid (OldUserID))
    SetTestResult ("Failed to restore old user ID after permissions test",
    errno);
  if (setgid (OldGroupID))
    SetTestResult ("Failed to restore old group ID after permissions test",
    errno);
  umask (OldUMask);
  if (FileHandle0 >= 0)
    close (FileHandle0);
  if (FileHandle1 >= 0)
    close (FileHandle1);
  unlink ("/Test/DirPermTest/FiveOwnerFile");
  rmdir ("/Test/DirPermTest/FiveOwnerDir");
  rmdir ("/Test/DirPermTest");

  return true;
}



/******************************************************************************
 * The TestQuery class tests the query (finding files by using indices and
 * query expressions) system.
 */

class TestQuery : public TestAbstract
{
public:
  TestQuery (BRect NewFrame)
    : TestAbstract (NewFrame, "TestQuery")
    {};

  virtual bool DoTest ();
  virtual const char *GetDescription ()
    { return "Test Queries."; };
};


bool TestQuery::DoTest ()
{
  uint8               DirEntBuffer [1024];
  dirent             *DirEntPntr;
  BEntry              Entry;
  int                 ErrorCode;
  BFile               File;
  int32               LocalInt32;
  int64               LocalInt64;
  const char         *LocalStringPntr;
  BMessage           *MsgPntr;
  struct node_ref     NodeRefDirStuff;
  struct node_ref     NodeRefTheFile;
  BQuery              Query;
  QueryLooper        *QueryLooperPntr = NULL;
  struct stat         StatInfo;
  BVolume             Volume;

  /* Start up the live query monitoring BLooper. */

  QueryLooperPntr = new QueryLooper;
  ErrorCode = 0;
  if (QueryLooperPntr == NULL ||
  (ErrorCode = QueryLooperPntr->Run ()) < 0)
  {
    SetTestResult ("Failed to start up the looper for monitoring "
      "query messages", ErrorCode);
    return false; /* Can't use a goto here due to BMessenger construction. */
  }

  /* Construct the BMessenger on the stack, since it can't be constructed and
  configured separately. */

  BMessenger MessengerToQueryLooper (NULL, QueryLooperPntr);

  /* Make the directory structure /Test/Stuff */

  if ((ErrorCode = create_directory ("/Test/Stuff", 0777)) != 0)
  {
    SetTestResult ("Failed to make /Test/Stuff directory", ErrorCode);
    goto ErrorExit;
  }

  /* Get the inode and volume numbers of the Stuff directory. */

  if (stat ("/Test/Stuff", &StatInfo)) goto ErrorExit;
  NodeRefDirStuff.device = StatInfo.st_dev;
  NodeRefDirStuff.node = StatInfo.st_ino;

  /* Start a simple query which our new file will trigger. */

  Volume.SetTo (NodeRefDirStuff.device);
  Query.SetVolume (&Volume);
  Query.SetPredicate ("name=Blort*");
  Query.SetTarget (MessengerToQueryLooper);
  if ((ErrorCode = Query.Fetch ()) != 0)
  {
    SetTestResult ("Query's Fetch failed", ErrorCode);
    goto ErrorExit;
  }

  /* Make a test file. */

  ErrorCode =
    File.SetTo ("/Test/Stuff/Blortiferous", B_READ_WRITE | B_CREATE_FILE);
  if (ErrorCode != 0)
  {
    SetTestResult ("Unable to create /Test/Stuff/Blortiferous", ErrorCode);
    goto ErrorExit;
  }
  File.GetNodeRef (&NodeRefTheFile);

  if (QueryLooperPntr->WaitForOpCodes (2) != 1 ||
  QueryLooperPntr->HasOverflowed () || (MsgPntr =
  QueryLooperPntr->GetMessageForOpCode (B_ENTRY_CREATED)) == NULL)
  {
    SetTestResult ("Should just get a single B_ENTRY_CREATED query "
      "message for the creation of a file.", false);
    QueryLooperPntr->PrintAllMessages ();
    goto ErrorExit;
  }

  /* Verify the contents of the query message. */

  if (MsgPntr->FindInt32 ("device", &LocalInt32) != B_OK ||
  LocalInt32 != NodeRefDirStuff.device)
  {
    SetTestResult ("B_ENTRY_CREATED query message has an incorrect "
      "device field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt64 ("directory", &LocalInt64) != B_OK ||
  LocalInt64 != NodeRefDirStuff.node)
  {
    SetTestResult ("B_ENTRY_CREATED query message has an incorrect "
      "directory field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt64 ("node", &LocalInt64) != B_OK ||
  LocalInt64 != NodeRefTheFile.node)
  {
    SetTestResult ("B_ENTRY_CREATED query message has an incorrect "
      "node field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindString ("name", &LocalStringPntr) != B_OK ||
  strcmp (LocalStringPntr, "Blortiferous") != 0)
  {
    SetTestResult ("B_ENTRY_CREATED query message has an incorrect "
      "name field.", false);
    goto ErrorExit;
  }

  /* Rename the file, which should remove it from the query set. */

  File.Unset ();
  Entry.SetTo ("/Test/Stuff/Blortiferous");
  QueryLooperPntr->Clear ();
  Entry.Rename ("SomeOtherName");
  Entry.Unset ();

  if (QueryLooperPntr->WaitForOpCodes (2) != 1 ||
  QueryLooperPntr->HasOverflowed () || (MsgPntr =
  QueryLooperPntr->GetMessageForOpCode (B_ENTRY_REMOVED)) == NULL)
  {
    SetTestResult ("Should just get a single B_ENTRY_REMOVED query "
      "message for the renaming of a file.", false);
    QueryLooperPntr->PrintAllMessages ();
    goto ErrorExit;
  }

  /* Verify the contents of the query message. */

  if (MsgPntr->FindInt32 ("device", &LocalInt32) != B_OK ||
  LocalInt32 != NodeRefDirStuff.device)
  {
    SetTestResult ("B_ENTRY_REMOVED query message has an incorrect "
      "device field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt64 ("directory", &LocalInt64) != B_OK ||
  LocalInt64 != NodeRefDirStuff.node)
  {
    SetTestResult ("B_ENTRY_REMOVED query message has an incorrect "
      "directory field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt64 ("node", &LocalInt64) != B_OK ||
  LocalInt64 != NodeRefTheFile.node)
  {
    SetTestResult ("B_ENTRY_REMOVED query message has an incorrect "
      "node field.", false);
    goto ErrorExit;
  }

  if ((ErrorCode = MsgPntr->FindString ("name", &LocalStringPntr)) != B_OK)
  {
    SetTestResult ("B_ENTRY_REMOVED query message doesn't have a name field, "
      "a known defect in BeOS R5", ErrorCode);
  }
  else if (strcmp (LocalStringPntr, "Blortiferous") != 0)
  {
    SetTestResult ("B_ENTRY_REMOVED query message has an incorrect "
      "name field (should be the old name).", false);
    goto ErrorExit;
  }

  /* Test for query composite logical operations and lack of update messages if
  the result is true both before and after.  In other words, use a query
  expression for size < 10 || size > 100, and then change the size from 5 to
  500.  Should result in no query messages. */

  Query.Clear ();
  QueryLooperPntr->Clear ();
  Query.SetVolume (&Volume);
  Query.SetPredicate ("\tsize  <  10  ||  \t \"size\" >\n100  ");
  Query.SetTarget (MessengerToQueryLooper);
  if ((ErrorCode = Query.Fetch ()) != 0)
  {
    SetTestResult ("Unable to set up a query with odd white space", ErrorCode);
    goto ErrorExit;
  }

  ErrorCode =
    File.SetTo ("/Test/Stuff/SizeTest", B_READ_WRITE | B_CREATE_FILE);
  if (ErrorCode != 0)
  {
    SetTestResult ("Unable to create /Test/Stuff/SizeTest", ErrorCode);
    goto ErrorExit;
  }
  File.GetNodeRef (&NodeRefTheFile);

  QueryLooperPntr->Clear ();
  File.SetSize (500);
  File.Unset (); /* Force it to flush the file and update the size. */

  ErrorCode = QueryLooperPntr->WaitForOpCodes (3);
  if (QueryLooperPntr->HasOverflowed ())
  {
    SetTestResult ("Too many notifications in composite logic test.", false);
    QueryLooperPntr->PrintAllMessages ();
    goto ErrorExit;
  }
  else if (ErrorCode == 0)
  {
    SetTestResult ("Good, no notifications for a file size change that leaves "
      "the file in the query set.", false);
  }
  else if (ErrorCode == 2)
  {
    if (QueryLooperPntr->GetMessageForOpCode (B_ENTRY_CREATED) == NULL ||
    QueryLooperPntr->GetMessageForOpCode (B_ENTRY_REMOVED) == NULL)
    {
      SetTestResult ("Didn't get a pair of entry created/removed for change "
        "in file size.", false);
      QueryLooperPntr->PrintAllMessages ();
      goto ErrorExit;
    }
    MsgPntr = QueryLooperPntr->GetMessageForOpCode (B_ENTRY_CREATED);
    if (MsgPntr->FindInt64 ("node", &LocalInt64) != B_OK ||
    LocalInt64 != NodeRefTheFile.node)
    {
      SetTestResult ("B_ENTRY_CREATED query message in composite test "
        "has an incorrect node field.", false);
      QueryLooperPntr->PrintAllMessages ();
      goto ErrorExit;
    }
    MsgPntr = QueryLooperPntr->GetMessageForOpCode (B_ENTRY_REMOVED);
    if (MsgPntr->FindInt64 ("node", &LocalInt64) != B_OK ||
    LocalInt64 != NodeRefTheFile.node)
    {
      SetTestResult ("B_ENTRY_REMOVED query message in composite test "
        "has an incorrect node field.", false);
      QueryLooperPntr->PrintAllMessages ();
      goto ErrorExit;
    }
    SetTestResult ("Average performance.  Got a remove/create pair of "
      "notifications for a file size change that leaves "
      "the file in the query set.", false);
  }
  else
  {
    SetTestResult ("Composite logic test didn't get the right notification "
      "messages.", false);
    QueryLooperPntr->PrintAllMessages ();
    goto ErrorExit;
  }

  /* Check the parsing ability of the file system.  Will it detect a query with
  garbage after the sensible part? */

  Query.Clear ();
  Query.SetVolume (&Volume);
  Query.SetPredicate ("name=\"Blort*\"<42");
  if ((ErrorCode = Query.Fetch ()) == 0)
  {
    SetTestResult ("Query's Fetch succeeded even though 'name=\"Blort*\"<42' "
      "should be a syntax error.", false);
    if (!g_FakeMount) /* BFS doesn't have syntax checking. */
      goto ErrorExit;
  }

  /* Does the query allow you to specify double quote marks by preceding them
  with a backslash?  Also tests reading the static query results rather than
  the live notification messages.  And see if it can return a batch of multiple
  results. */

  ErrorCode =
    File.SetTo ("/Test/Stuff/Quote\"File", B_READ_WRITE | B_CREATE_FILE);
  if (ErrorCode != 0)
  {
    SetTestResult ("Unable to create /Test/Stuff/Quote\"File", ErrorCode);
    goto ErrorExit;
  }
  ErrorCode =
    File.SetTo ("/Test/Stuff/\"Double Quote\"", B_READ_WRITE | B_CREATE_FILE);
  if (ErrorCode != 0)
  {
    SetTestResult ("Unable to create /Test/Stuff/\"Double Quote\"", ErrorCode);
    goto ErrorExit;
  }
  File.GetNodeRef (&NodeRefTheFile); /* This one is first in alpha order. */
  File.Unset ();

  Query.Clear ();
  Query.SetVolume (&Volume);
  Query.SetPredicate ("name=\"*\\\"*\"");
  if ((ErrorCode = Query.Fetch ()) != 0)
  {
    SetTestResult ("Failed to make query containing a quote mark", ErrorCode);
    goto ErrorExit;
  }

  ErrorCode = Query.GetNextDirents ((dirent *) DirEntBuffer,
    sizeof (DirEntBuffer), 123);
  if (ErrorCode <= 0 || ErrorCode > 2)
  {
    SetTestResult ("Query.GetNextDirents failed, should have returned "
      "1 or 2 results", ErrorCode);
    goto ErrorExit;
  }

  if (ErrorCode == 1)
    SetTestResult ("Query.GetNextDirents only returned 1 result, "
    "someone used a simpler implementation.", false);

  DirEntPntr = (dirent *) DirEntBuffer;
  if (DirEntPntr->d_dev != NodeRefTheFile.device ||
  DirEntPntr->d_pdev != NodeRefDirStuff.device ||
  DirEntPntr->d_ino != NodeRefTheFile.node ||
  DirEntPntr->d_pino != NodeRefDirStuff.node)
  {
    SetTestResult ("Query.GetNextDirents returned wrong inode or "
      "device numbers.", false);
    goto ErrorExit;
  }
  if (strcmp (DirEntPntr->d_name, "\"Double Quote\"") != 0)
  {
    SetTestResult ("Query.GetNextDirents returned wrong file name.", false);
    if (!g_FakeMount) /* Real disk volumes have lots of files with quotes. */
      goto ErrorExit;
  }
  if (DirEntPntr->d_reclen != sizeof (dirent) + strlen (DirEntPntr->d_name))
  {
    SetTestResult ("Query.GetNextDirents returned bad d_reclen value, "
      "should be able to add it to the record start to get the next "
      "record.", false);
    if (!g_FakeMount) /* BFS gets this wrong. */
      goto ErrorExit;
  }

  if (ErrorCode == 1)
  {
    ErrorCode = Query.GetNextDirents ((dirent *) DirEntBuffer,
      sizeof (DirEntBuffer), 123);
    if (ErrorCode != 1)
    {
      SetTestResult ("Query.GetNextDirents second call didn't "
        "return anything.", false);
      goto ErrorExit;
    }
  }
  else /* Already in the buffer, advance to next record. */
    DirEntPntr = (dirent *) (((uint8 *) DirEntPntr) + DirEntPntr->d_reclen);

  if (strcmp (DirEntPntr->d_name, "Quote\"File") != 0)
  {
    SetTestResult ("Query.GetNextDirents second returned "
      "wrong file name.", false);
    if (!g_FakeMount) /* Real disk volumes have lots of files with quotes. */
      goto ErrorExit;
  }

  /* Should be no more entries left. */

  ErrorCode = Query.GetNextDirents ((dirent *) DirEntBuffer,
    sizeof (DirEntBuffer), 123);
  if (ErrorCode != 0)
  {
    SetTestResult ("Query.GetNextDirents final call should "
      "have returned zero.", false);
    if (!g_FakeMount) /* Real disk volumes have lots of files with quotes. */
      goto ErrorExit;
  }

  /* See if rewind works. */

  ErrorCode = Query.Rewind ();
  if (ErrorCode != 0)
  {
    SetTestResult ("Query.Rewind not supported, stupid OS", ErrorCode);
  }

  SetTestResult ("Query test OK.");

ErrorExit:
  Query.Clear ();
  if (QueryLooperPntr != NULL)
    QueryLooperPntr->PostMessage (B_QUIT_REQUESTED);
  File.Unset ();
  RecursivelyDelete ("/Test/Stuff");
  return true;
}



/******************************************************************************
 * Test the rename feature.
 */

class TestRename : public TestAbstract
{
public:
  TestRename (BRect NewFrame)
    : TestAbstract (NewFrame, "TestRename")
    {};

  virtual bool DoTest ();
  virtual const char *GetDescription ()
    { return "Test renaming."; };
};


bool TestRename::DoTest ()
{
  int         ErrorCode;
  BFile       File;
  char        TempString [B_PATH_NAME_LENGTH];
  const char *WriteString = "This is a file with a string in it.\n";

  /* Test a few invalid cases before making any test files. */

  ErrorCode = 0;
  if (!rename ("/Test/Blah", "/Test/Blort") ||
  (ErrorCode = errno) != ENOENT)
  {
    SetTestResult ("Renaming non-existant files should fail with ENOENT",
      ErrorCode);
    goto ErrorExit;
  }

  ErrorCode = 0;
  if (!rename (NULL, "Blort") ||
  (ErrorCode = errno) != EFAULT)
  {
    SetTestResult ("Renaming from NULL should fail with EFAULT",
      ErrorCode);
    goto ErrorExit;
  }

  ErrorCode = 0;
  if (!rename ("Blort", NULL) ||
  (ErrorCode = errno) != EFAULT)
  {
    SetTestResult ("Renaming to NULL should fail with EFAULT",
      ErrorCode);
    goto ErrorExit;
  }

  /* Make the directory structure /Test/a/(b,bb/cc,bbb). */

  if ((ErrorCode = create_directory ("/Test/a/b", 0777)) != 0 ||
  (ErrorCode = create_directory ("/Test/a/bb/cc", 0777)) != 0 ||
  (ErrorCode = create_directory ("/Test/a/bbb", 0777)) != 0)
  {
    SetTestResult ("Failed to make /Test/a/(b,bb/cc,bbb) directories",
      ErrorCode);
    goto ErrorExit;
  }

  /* Make a file /Test/a/OldFile. */

  ErrorCode = File.SetTo ("/Test/a/OldFile", B_READ_WRITE | B_CREATE_FILE);
  if (ErrorCode != 0)
  {
    SetTestResult ("Unable to create /Test/a/OldFile", ErrorCode);
    goto ErrorExit;
  }
  ErrorCode = File.Write (WriteString, strlen (WriteString));
  if (ErrorCode != (int) strlen (WriteString))
  {
    SetTestResult ("Unable to write to /Test/a/OldFile", ErrorCode);
    goto ErrorExit;
  }
  File.Unset ();

  if (!VerifyDirectoryTree ("/Test/a",
  ",d,*\n"
  "b,d,*\n"
  "bb,d,*\n"
  "bb/cc,d,*\n"
  "bbb,d,*\n"
  "OldFile,f,36\n"))
    goto ErrorExit;

  /* Test a simple name change. */

  if (rename ("/Test/a/OldFile", "/Test/a/NewFile"))
  {
    SetTestResult ("Rename /Test/a/OldFile to /Test/a/NewFile failed", errno);
    goto ErrorExit;
  }

  if (!VerifyDirectoryTree ("/Test/a",
  ",d,*\n"
  "b,d,*\n"
  "bb,d,*\n"
  "bb/cc,d,*\n"
  "bbb,d,*\n"
  "NewFile,f,36\n"))
    goto ErrorExit;

  /* Try opening the new file name, should end up reading the original data. */

  ErrorCode = File.SetTo ("/Test/a/NewFile", B_READ_WRITE);
  if (ErrorCode != 0)
  {
    SetTestResult ("Unable to open /Test/a/NewFile", ErrorCode);
    goto ErrorExit;
  }
  ErrorCode = File.Read (TempString, sizeof (TempString));
  if (ErrorCode != (int) strlen (WriteString))
  {
    SetTestResult ("Read wrong amount of data from /Test/a/NewFile",
      ErrorCode);
    goto ErrorExit;
  }
  File.Unset ();
  if (memcmp (TempString, WriteString, strlen (WriteString)) != 0)
  {
    SetTestResult ("Incorrect data read from /Test/a/NewFile",
      false);
    goto ErrorExit;
  }

  /* Try renaming to a name that's too long. */

  ErrorCode = 0;
  if (!rename ("/Test/a/NewFile",
  "/Test/a/FileLong901234567890123456789012345678901234567890"
  "12345678901234567890123456789012345678901234567890"
  "12345678901234567890123456789012345678901234567890"
  "12345678901234567890123456789012345678901234567890"
  "12345678901234567890123456789012345678901234567890"
  "123456") || (ErrorCode = errno) != ENAMETOOLONG)
  {
    SetTestResult ("Renaming to a 256 byte long name should fail "
      "with ENAMETOOLONG", ErrorCode);
    goto ErrorExit;
  }

  /* Try renaming it to itself.  May succeed, or fail with EPERM. */

  ErrorCode = 0;
  if (!rename ("/Test/a/NewFile", "/Test/a/NewFile") ||
  (ErrorCode = errno) != EINVAL)
  {
    SetTestResult ("Rename /Test/a/NewFile to itself should fail with EINVAL",
      ErrorCode);
    if (ErrorCode != 0) /* AGMSRAMFileSystem allows rename to itself. */
      goto ErrorExit;
  }

  /* Try renaming a file to ".." which should fail (reserved word). */

  ErrorCode = 0;
  if (!rename ("/Test/a/NewFile", "/Test/a/b/..") ||
  (ErrorCode = errno) != EINVAL)
  {
    SetTestResult ("Renaming a file to \"..\" should fail with EINVAL",
      ErrorCode);
    goto ErrorExit;
  }

  /* Try renaming a file to "..." which should also fail (reserved word). */

  if (!g_FakeMount) /* Regular BFS allows ... as a name so don't test BFS. */
  {
    ErrorCode = 0;
    if (!rename ("/Test/a/NewFile", "/Test/a/b/...") ||
    (ErrorCode = errno) != EINVAL)
    {
      SetTestResult ("Renaming a file to \"...\" should fail with EINVAL",
        ErrorCode);
      goto ErrorExit;
    }
  }

  /* Try renaming . to some other name, may fail. */

  ErrorCode = 0;
  if (!rename ("/Test/a/b/.", "/Test/a/Dot") ||
  (ErrorCode = errno) != EINVAL)
  {
    SetTestResult ("Renaming \".\" should fail with EINVAL", ErrorCode);
    goto ErrorExit;
  }

  /* Try renaming a file over an empty directory. */

  ErrorCode = 0;
  if (!rename ("/Test/a/NewFile", "/Test/a/b") ||
  (ErrorCode = errno) != EISDIR)
  {
    SetTestResult ("Renaming a file to a directory should fail with EISDIR",
      ErrorCode);
    if (ErrorCode != 0) /* BFS allows it.  So does AGMSRAMFileSystem. */
      goto ErrorExit;
  }

  /* Try renaming a file over a non-empty directory. */

  ErrorCode = 0;
  if (!rename ("/Test/a/b", "/Test/a/bb") ||
  (ErrorCode = errno) != ENOTEMPTY)
  {
    SetTestResult ("Renaming a file over a non-empty directory should "
      "fail with EISDIR", ErrorCode);
    if (ErrorCode != ENOTEMPTY) /* BFS allows it, so error is nonempty. */
      goto ErrorExit;
  }

  File.SetTo ("/Test/a/ExistingFile", B_READ_WRITE | B_CREATE_FILE);
  File.Write ("Existing", strlen ("Existing"));
  File.Unset ();

  File.SetTo ("/Test/a/bb/AnotherFile", B_READ_WRITE | B_CREATE_FILE);
  File.Write ("Another", strlen ("Another"));
  File.Unset ();

  if (!VerifyDirectoryTree ("/Test/a",
  ",d,*\n"
  "b,f,36\n"
  "bb,d,*\n"
  "bb/AnotherFile,f,7\n"
  "bb/cc,d,*\n"
  "bbb,d,*\n"
  "ExistingFile,f,8\n"))
    goto ErrorExit;

  /* Rename a file over an existing file. */

  if (rename ("/Test/a/b", "/Test/a/ExistingFile"))
  {
    SetTestResult ("Renaming a file over an existing file failed", errno);
    goto ErrorExit;
  }

  /* Rename a file over an existing file in a different directory. */

  if (rename ("/Test/a/ExistingFile", "/Test/a/bb/AnotherFile"))
  {
    SetTestResult ("Renaming a file over an existing file in another "
      "directory failed", errno);
    goto ErrorExit;
  }

  if (!VerifyDirectoryTree ("/Test/a",
  ",d,*\n"
  "bb,d,*\n"
  "bb/AnotherFile,f,36\n"
  "bb/cc,d,*\n"
  "bbb,d,*\n"))
    goto ErrorExit;

  /* Rename a symbolic link. */

  if (symlink ("/Test/a/bb/AnotherFile", "/Test/a/SymLink"))
  {
    SetTestResult ("Unable to make link /Test/a/SymLink to AnotherFile",
      errno);
    goto ErrorExit;
  }

  if (rename ("/Test/a/SymLink", "/Test/a/bb/SymLinkRenamed"))
  {
    SetTestResult ("Renaming a symbolic link failed", errno);
    goto ErrorExit;
  }

  if (!VerifyDirectoryTree ("/Test/a",
  ",d,*\n"
  "bb,d,*\n"
  "bb/AnotherFile,f,36\n"
  "bb/cc,d,*\n"
  "bb/SymLinkRenamed,s,/Test/a/bb/AnotherFile\n"
  "bbb,d,*\n"))
    goto ErrorExit;

  /* Rename a symbolic link over a directory. */

  ErrorCode = 0;
  if (!rename ("/Test/a/bb/SymLinkRenamed", "/Test/a/bbb") ||
  (ErrorCode = errno) != EISDIR)
  {
    SetTestResult ("Renaming a symbolic link to a directory should fail "
      "with EISDIR", ErrorCode);
    if (ErrorCode != 0) /* BFS allows it.  So does AGMSRAMFileSystem. */
      goto ErrorExit;
  }

  /* Rename a file over a symbolic link. */

  if (rename ("/Test/a/bb/AnotherFile", "/Test/a/bbb"))
  {
    SetTestResult ("Renaming a file over a symbolic link failed", errno);
    goto ErrorExit;
  }

  if (!VerifyDirectoryTree ("/Test/a",
  ",d,*\n"
  "bb,d,*\n"
  "bb/cc,d,*\n"
  "bbb,f,36\n"))
    goto ErrorExit;

  SetTestResult ("Rename test OK.");

ErrorExit:
  File.Unset ();
  RecursivelyDelete ("/Test/a");
  return true;
}



/******************************************************************************
 * Test symbolic link operations.
 */

class TestSymbolicLink : public TestAbstract
{
public:
  TestSymbolicLink (BRect NewFrame)
    : TestAbstract (NewFrame, "TestSymbolicLink")
    {};

  virtual bool DoTest ();
  virtual const char *GetDescription ()
    { return "Test Symbolic Link operations."; };
};



bool TestSymbolicLink::DoTest ()
{
  int         ErrorCode;
  BFile       File;
  int         i;
  const char *NotAValidLinkString = "../NotAValidLink";
  char        TempString [B_PATH_NAME_LENGTH];
  const char *WriteString = "This is a file with a string in it.\n";

  /* Make the directory structure /Test/a/b/c. */

  if ((ErrorCode = create_directory ("/Test/a/b/c", 0777)) != 0)
  {
    SetTestResult ("Failed to make /Test/a/b/c directories", ErrorCode);
    goto ErrorExit;
  }

  /* Make a file /Test/a/TheFile. */

  ErrorCode = File.SetTo ("/Test/a/TheFile", B_READ_WRITE | B_CREATE_FILE);
  if (ErrorCode != 0)
  {
    SetTestResult ("Unable to create /Test/a/TheFile", ErrorCode);
    goto ErrorExit;
  }
  ErrorCode = File.Write (WriteString, strlen (WriteString));
  if (ErrorCode != (int) strlen (WriteString))
  {
    SetTestResult ("Unable to write to /Test/a/TheFile", ErrorCode);
    goto ErrorExit;
  }
  File.Unset ();

  /* Make absolute symbolic link dAbsToTheFile, pointing back to TheFile. */

  if (symlink ("/Test/a/TheFile", "/Test/a/b/c/dAbsToTheFile"))
  {
    SetTestResult ("Unable to make link /Test/a/b/c/dAbsToTheFile to TheFile",
      errno);
    goto ErrorExit;
  }

  /* Make relative symbolic link eRelToB, pointing back to b. */

  if (symlink ("..", "/Test/a/b/c/eRelToB"))
  {
    SetTestResult ("Unable to make link /Test/a/b/c/eRelToB to directory b",
      errno);
    goto ErrorExit;
  }

  /* Make a broken symbolic link fBroken. */

  if (symlink (NotAValidLinkString, "/Test/a/b/c/fBroken"))
  {
    SetTestResult ("Unable to make link /Test/a/b/c/fBroken",
      errno);
    goto ErrorExit;
  }

  if (!VerifyDirectoryTree ("/Test/a",
  ",d,*\n"
  "TheFile,f,36\n"
  "b,d,*\n"
  "b/c,d,*\n"
  "b/c/dAbsToTheFile,s,/Test/a/TheFile\n"
  "b/c/eRelToB,s,..\n"
  "b/c/fBroken,s,../NotAValidLink\n"
  ))
    goto ErrorExit;

  /* Verify that readlink doesn't include the NUL at the end of the string,
  just the string size.  Possibly stupid, but that's the POSIX standard. */

  memset (TempString, '?', sizeof (TempString));
  i = readlink ("/Test/a/b/c/fBroken", TempString, sizeof (TempString));
  if (i < 0)
  {
    SetTestResult ("Readlink failed for /Test/a/b/c/fBroken", errno);
    goto ErrorExit;
  }
  if (i != (int) strlen (NotAValidLinkString))
  {
    sprintf (TempString, "Readlink is returning the wrong size (%d) of the "
      "symlink string (should be %d).", i, (int) strlen (NotAValidLinkString));
    SetTestResult (TempString, false);
    goto ErrorExit;
  }
  if (TempString[strlen (NotAValidLinkString)] != '?')
  {
    SetTestResult ("Readlink is overwriting past the end of the string, "
      "the standard says no NUL at the end of the string!", false);
    goto ErrorExit;
  }
  TempString[strlen (NotAValidLinkString)] = 0;
  if (strcmp (TempString, NotAValidLinkString) != 0)
  {
    SetTestResult ("Readlink's returned string isn't what we expected.",
      false);
    goto ErrorExit;
  }

  /* Try opening a symbolic file, should end up reading the data in the real
  actual file. */

  ErrorCode = File.SetTo ("/Test/a/b/c/dAbsToTheFile", B_READ_WRITE);
  if (ErrorCode != 0)
  {
    SetTestResult ("Unable to open /Test/a/b/c/dAbsToTheFile", ErrorCode);
    goto ErrorExit;
  }
  ErrorCode = File.Read (TempString, sizeof (TempString));
  if (ErrorCode != (int) strlen (WriteString))
  {
    SetTestResult ("Read wrong amount of data from /Test/a/b/c/dAbsToTheFile",
      ErrorCode);
    goto ErrorExit;
  }
  File.Unset ();
  if (memcmp (TempString, WriteString, strlen (WriteString)) != 0)
  {
    SetTestResult ("Incorrect data read from /Test/a/b/c/dAbsToTheFile",
      false);
    goto ErrorExit;
  }

  /* Examine a symbolic directory listing.  Sneakily append a dot to the path
  so that it doesn't think that the name is just a symbolic link. */

  if (!VerifyDirectoryTree ("/Test/a/b/c/eRelToB/.",
  ",d,*\n"
  "c,d,*\n"
  "c/dAbsToTheFile,s,/Test/a/TheFile\n"
  "c/eRelToB,s,..\n"
  "c/fBroken,s,../NotAValidLink\n"
  ))
  {
    SetTestResult ("Directory listing of symbolic linked directory is wrong.",
      false);
    goto ErrorExit;
  }

  /* Try opening a broken link. */

  ErrorCode = File.SetTo ("/Test/a/b/c/fBroken", B_READ_WRITE);
  if (ErrorCode != B_ENTRY_NOT_FOUND)
  {
    SetTestResult ("Expected to get a file not found error when opening "
      "a broken link of /Test/a/b/c/fBroken", ErrorCode);
    goto ErrorExit;
  }

  /* Try calling readlink with a buffer that's too small.  Should return the
  size of the full symbolic string, but only write the size specified to the
  buffer. */

  memset (TempString, '?', sizeof (TempString));
  i = readlink ("/Test/a/b/c/fBroken", TempString, 4);
  if (i < 0)
  {
    SetTestResult ("Readlink failed for /Test/a/b/c/fBroken with a "
      "too small buffer", errno);
    goto ErrorExit;
  }
  if (i != (int) strlen (NotAValidLinkString))
  {
    sprintf (TempString, "Readlink is returning the wrong size (%d) of the "
      "symlink string (should be %d).", i, (int) strlen (NotAValidLinkString));
    SetTestResult (TempString, false);
    goto ErrorExit;
  }
  if (TempString [4 /* past end of our short buffer */] != '?')
  {
    SetTestResult ("Readlink is overwriting past the end of the too "
      "small buffer.", false);
    goto ErrorExit;
  }
  if (memcmp (TempString, NotAValidLinkString, 4) != 0)
  {
    SetTestResult ("Readlink's returned buffer too small string isn't "
      "what we expected.", false);
    goto ErrorExit;
  }

  /* Try calling readlink with a NULL buffer.  Does it return an error? */

  ErrorCode = 0;
  i = readlink ("/Test/a/b/c/fBroken", NULL, sizeof (TempString));
  if (i >= 0 || (ErrorCode = errno) != EFAULT)
  {
    SetTestResult ("Readlink with a NULL buffer should fail with "
      "a EFAULT error.", ErrorCode);
    goto ErrorExit;
  }

  /* Try calling readlink with a zero size buffer. */

  i = readlink ("/Test/a/b/c/fBroken", TempString, 0);
  if (i != (int) strlen (NotAValidLinkString))
  {
    sprintf (TempString, "Readlink should return full string size for a "
      "zero length buffer, but returned %d instead of %d.",
      i, (int) strlen (NotAValidLinkString));
    SetTestResult (TempString, false);
    goto ErrorExit;
  }

  /* Try calling readlink with a NULL and zero size buffer.  Should work. */

  i = readlink ("/Test/a/b/c/fBroken", NULL, 0);
  if (i != (int) strlen (NotAValidLinkString))
  {
    sprintf (TempString, "Readlink should return full size when called with "
      "a NULL buffer and zero size, returned %d instead of %d.",
      i, (int) strlen (NotAValidLinkString));
    SetTestResult (TempString, false);
    goto ErrorExit;
  }

  SetTestResult ("Symbolic Link test OK.");

ErrorExit:
  File.Unset ();
  RecursivelyDelete ("/Test/a");
  return true;
}



/******************************************************************************
 * Test the umask feature.
 */

class TestUMask : public TestAbstract
{
public:
  TestUMask (BRect NewFrame)
    : TestAbstract (NewFrame, "TestUMask")
    {};

  virtual bool DoTest ();
  virtual const char *GetDescription ()
    { return "Test umask permissions."; };
};


bool TestUMask::DoTest ()
{
  int         FileHandle0 = -1;
  int         FileHandle1 = -1;
  mode_t      OldUMask = umask (0 /* specifies rwxrwxrwx bits to disable */);
  struct stat StatBuf;

  FileHandle0 = open ("/Test/File1", O_RDWR | O_CREAT, 0777);
  if (FileHandle0 < 0)
  {
    SetTestResult ("Unable to create /Test/File1", errno);
    goto ErrorExit;
  }
  close (FileHandle0); FileHandle0 = -1;

  memset (&StatBuf, 0, sizeof (StatBuf));
  if (stat ("/Test/File1", &StatBuf) != 0)
  {
    SetTestResult ("Stat of /Test/File1 failed", errno);
    goto ErrorExit;
  }
  if ((StatBuf.st_mode & 0777) != 0777)
  {
    SetTestResult ("File /Test/File1 has wrong permissions.", false);
    goto ErrorExit;
  }

  umask (0070);
  FileHandle0 = open ("/Test/File2", O_RDWR | O_CREAT, 0765);
  if (FileHandle0 < 0)
  {
    SetTestResult ("Unable to create /Test/File2", errno);
    goto ErrorExit;
  }
  close (FileHandle0); FileHandle0 = -1;

  memset (&StatBuf, 0, sizeof (StatBuf));
  if (stat ("/Test/File2", &StatBuf) != 0)
  {
    SetTestResult ("Stat of /Test/File2 failed", errno);
    goto ErrorExit;
  }
  if ((StatBuf.st_mode & 0777) != 0705)
  {
    SetTestResult ("File /Test/File2 has wrong permissions.", false);
    goto ErrorExit;
  }

  umask (0772);
  FileHandle0 = open ("/Test/File3", O_RDWR | O_CREAT, -1);
  if (FileHandle0 < 0)
  {
    SetTestResult ("Unable to create /Test/File3", errno);
    goto ErrorExit;
  }
  close (FileHandle0); FileHandle0 = -1;

  memset (&StatBuf, 0, sizeof (StatBuf));
  if (stat ("/Test/File3", &StatBuf) != 0)
  {
    SetTestResult ("Stat of /Test/File3 failed", errno);
    goto ErrorExit;
  }
  if ((StatBuf.st_mode & 0777) != 0005)
  {
    SetTestResult ("File /Test/File3 has wrong permissions.", false);
    goto ErrorExit;
  }

  SetTestResult ("UMask tests OK.");

ErrorExit:
  umask (OldUMask);
  if (FileHandle0 >= 0)
    close (FileHandle0);
  if (FileHandle1 >= 0)
    close (FileHandle1);
  unlink ("/Test/File1");
  unlink ("/Test/File2");
  unlink ("/Test/File3");
  return true;
}



/******************************************************************************
 * Unmount the test file system.  Handles both fake and real mounts.  Closes
 * global stuff too.
 */

class TestUnmount : public TestAbstract
{
public:
  TestUnmount (BRect NewFrame)
    : TestAbstract (NewFrame, "TestUnmount")
    {m_TestEnabled = true;};

  virtual bool DoTest ();

  virtual const char *GetDescription ()
    { return "Unmount the /Test (fake or not) file system."; };
};


bool TestUnmount::DoTest ()
{
  int  ErrorCode = 0;
  char TempString [PATH_MAX + 80];

  if (g_FakeMount)
  {
    if (unlink ("/Test"))
    {
      ErrorCode = errno;
      SetTestResult ("Unable to remove symbolic link /Test", ErrorCode);
    }

    if (rmdir (g_FakeDirString) && ErrorCode == 0)
    {
      ErrorCode = errno;
      sprintf (TempString, "Unable to delete \"%s\" directory",
        g_FakeDirString);
      SetTestResult (TempString, ErrorCode);
    }
  }
  else
  {
    if (unmount ("/Test"))
    {
      ErrorCode = errno;
      SetTestResult ("Unmount failed", ErrorCode);
    }
    if (rmdir ("/Test") && ErrorCode == 0)
    {
      ErrorCode = errno;
      SetTestResult ("Can't remove /Test directory", ErrorCode);
    }
  }

  if (ErrorCode == 0)
    SetTestResult (g_FakeMount ? "Cleaned up fake /Test stuff." :
    "Unmounted real file system from /Test successfully.");
  return true;
}



/******************************************************************************
 * Test disk volume related things.
 */

class TestVolumeStuff : public TestAbstract
{
public:
  TestVolumeStuff (BRect NewFrame)
    : TestAbstract (NewFrame, "TestVolumeStuff")
    {};

  virtual bool DoTest ();
  virtual const char *GetDescription ()
    { return "Test Volume related things."; };
};


bool TestVolumeStuff::DoTest ()
{
  struct node_ref     DirNodeRef;
  int                 ErrorCode;
  int32               LocalInt32;
  int64               LocalInt64;
  const char         *LocalStringPntr;
  BMessage           *MsgPntr;
  bool                NameChanged = false;
  const char         *NewLongName = "New Volume Name which is longer than 32.";
  const char         *NewShortName = "Short Name";
  NotificationLooper *NotificationLooperPntr = NULL;
  char                OriginalName [PATH_MAX];
  struct node_ref     RootFSNodeRef;
  char                TempString [PATH_MAX + 256];
  BDirectory          TestDir;
  BVolume             Volume;

  NotificationLooperPntr = new NotificationLooper;
  ErrorCode = 0;
  if (NotificationLooperPntr == NULL ||
  (ErrorCode = NotificationLooperPntr->Run ()) < 0)
  {
    SetTestResult ("Failed to start up the looper for monitoring "
      "notification messages", ErrorCode);
    goto ErrorExit;
  }

  TestDir.SetTo ("/");
  ErrorCode = TestDir.GetNodeRef (&RootFSNodeRef);
  if (ErrorCode != 0)
  {
    SetTestResult ("Unable to get node ref for / (the rootfs)", ErrorCode);
    goto ErrorExit;
  }

  TestDir.SetTo ("/Test");
  ErrorCode = TestDir.GetNodeRef (&DirNodeRef);
  if (ErrorCode != 0)
  {
    SetTestResult ("Unable to get node ref for /Test", ErrorCode);
    goto ErrorExit;
  }
  TestDir.Unset ();

  ErrorCode = watch_node (&DirNodeRef, B_WATCH_ALL,
    NotificationLooperPntr, NotificationLooperPntr);
  if (ErrorCode != 0)
  {
    SetTestResult ("watch_node failed for /Test", ErrorCode);
    goto ErrorExit;
  }

  Volume.SetTo (DirNodeRef.device);
  ErrorCode = Volume.InitCheck ();
  if (ErrorCode != 0)
  {
    SetTestResult ("Unable to set up BVolume for /Test", ErrorCode);
    goto ErrorExit;
  }

  ErrorCode = Volume.GetName (OriginalName);
  if (ErrorCode != 0)
  {
    SetTestResult ("Unable to read volume name", ErrorCode);
    goto ErrorExit;
  }
  sprintf (TempString, "Original volume name is \"%s\".", OriginalName);
  SetTestResult (TempString, false);

  /* Try changing the volume name.  Temporarily, in case this is /boot. */

  ErrorCode = Volume.SetName (NewShortName);
  if (ErrorCode != 0)
  {
    SetTestResult ("Unable to change volume name", ErrorCode);
    goto ErrorExit;
  }
  NameChanged = true;
  ErrorCode = Volume.GetName (TempString);
  if (strcmp (TempString, NewShortName) != 0)
  {
    SetTestResult ("Change to short volume name didn't work correctly.",
      false);
    goto ErrorExit;
  }

  /* The system should have generated a B_ENTRY_MOVED notification.  Actually,
  it comes from the rootfs directory entry for the disk volume being changed to
  the new name. */

  if (NotificationLooperPntr->WaitForOpCodes (1) != 1 ||
  NotificationLooperPntr->HasOverflowed () || (MsgPntr =
  NotificationLooperPntr->GetMessageForOpCode (B_ENTRY_MOVED)) == NULL)
  {
    SetTestResult ("Should just get a single B_ENTRY_MOVED notification "
      "message when renaming the volume.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt32 ("device", &LocalInt32) != B_OK ||
  LocalInt32 != RootFSNodeRef.device)
  {
    SetTestResult ("B_ENTRY_MOVED notification message has an incorrect "
      "device field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt64 ("from directory", &LocalInt64) != B_OK ||
  LocalInt64 != RootFSNodeRef.node)
  {
    SetTestResult ("B_ENTRY_MOVED notification message has an incorrect "
      "from directory field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt64 ("to directory", &LocalInt64) != B_OK ||
  LocalInt64 != RootFSNodeRef.node)
  {
    SetTestResult ("B_ENTRY_MOVED notification message has an incorrect "
      "to directory field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindInt64 ("node", &LocalInt64) != B_OK ||
  LocalInt64 != DirNodeRef.node)
  {
    SetTestResult ("B_ENTRY_MOVED notification message has an incorrect "
      "node field.", false);
    goto ErrorExit;
  }

  if (MsgPntr->FindString ("name", &LocalStringPntr) != B_OK ||
  strcmp (LocalStringPntr, NewShortName) != 0)
  {
    SetTestResult ("B_ENTRY_MOVED notification message has an incorrect "
      "name field.", false);
    goto ErrorExit;
  }

  /* Try an overly long volume name (should be able to do file name length). */

  ErrorCode = Volume.SetName (NewLongName);
  if (ErrorCode != 0)
  {
    SetTestResult ("Unable to change volume name to long name", ErrorCode);
    goto ErrorExit;
  }
  NameChanged = true;
  ErrorCode = Volume.GetName (TempString);
  if (strcmp (TempString, NewLongName) != 0)
  {
    SetTestResult ("Change to long volume name didn't work correctly (new "
      "name isn't the same string).", false);
    goto ErrorExit;
  }

  SetTestResult ("Volume stuff tests successful.");

ErrorExit:
  stop_watching (NotificationLooperPntr, NotificationLooperPntr);
  if (NotificationLooperPntr != NULL)
    NotificationLooperPntr->PostMessage (B_QUIT_REQUESTED);
  if (NameChanged)
    Volume.SetName (OriginalName);
  TestDir.Unset ();
  Volume.Unset ();
  return true;
}



/******************************************************************************
 * Test file reading and writing.
 */

class TestWriteFile : public TestAbstract
{
public:
  TestWriteFile (BRect NewFrame);
  virtual ~TestWriteFile ();
  virtual void MessageReceived (BMessage *MessagePntr);
  virtual float AddCustomControls (float X);
  virtual void PreTestPreparations ();
  virtual bool DoTest ();

  virtual const char *GetDescription ()
    { return "Test writing and seeking in files."; };

private:
  void GetCheckAndSetTextBoxes ();

  int           m_BigBufSize;
  BTextControl *m_BigBufSizeTextBoxPntr;
  int           m_BigTimes; /* Number of times to write the big buffer. */
  BTextControl *m_BigTimesTextBoxPntr;
};


TestWriteFile::TestWriteFile (BRect NewFrame)
: TestAbstract (NewFrame, "TestWriteFile"),
  m_BigBufSize (50000000), /* Typically 1/4 of total real RAM is good. */
  m_BigBufSizeTextBoxPntr (NULL),
  m_BigTimes (1),
  m_BigTimesTextBoxPntr (NULL)
{
  /* Make sure there is space for our big buffer size text box. */

  if (m_GUIHeight < g_TextBoxHeight)
    m_GUIHeight = g_TextBoxHeight;
}


TestWriteFile::~TestWriteFile ()
{
  /* Child views (controls) are automatically destroyed and deallocated by
  the parent destructor.  So there isn't much to do here. */

  m_BigBufSizeTextBoxPntr = NULL;
  m_BigTimesTextBoxPntr = NULL;
}


void TestWriteFile::MessageReceived (BMessage *MessagePntr)
{
  switch (MessagePntr->what)
  {
    case MSG_BIG_BUF_SIZE:
    case MSG_BIG_BUF_TIMES:
      GetCheckAndSetTextBoxes ();
      break;

    default:
      TestAbstract::MessageReceived (MessagePntr);
      break;
  }
}


float TestWriteFile::AddCustomControls (float X)
{
  float Margin;
  BRect TempRect;
  char  TempString [80];

  /* Add the text control used to specify the buffer size to use. */

  Margin = ceilf ((m_GUIHeight - g_TextBoxHeight) / 2);
  TempRect = Bounds ();
  TempRect.top += Margin;
  TempRect.bottom = TempRect.top + g_TextBoxHeight;
  TempRect.left = X;
  TempRect.right = X + 100;

  sprintf (TempString, "%d", m_BigBufSize);
  m_BigBufSizeTextBoxPntr = new BTextControl (TempRect,
    "Big Buffer Size Text Control",
    "Size:" /* label */,
    TempString,
    new BMessage (MSG_BIG_BUF_SIZE),
    B_FOLLOW_LEFT | B_FOLLOW_TOP,
    B_WILL_DRAW | B_FRAME_EVENTS | B_FULL_UPDATE_ON_RESIZE | B_NAVIGABLE);
  if (m_BigBufSizeTextBoxPntr == NULL) goto ErrorExit;
  m_BigBufSizeTextBoxPntr->SetDivider (be_plain_font->StringWidth ("Size: "));
  AddChild (m_BigBufSizeTextBoxPntr);
  m_BigBufSizeTextBoxPntr->SetTarget (this);
  X += ceilf (m_BigBufSizeTextBoxPntr->Bounds().Width());

  /* Add the text control for the number of times to write the big buffer. */

  Margin = ceilf ((m_GUIHeight - g_TextBoxHeight) / 2);
  TempRect = Bounds ();
  TempRect.top += Margin;
  TempRect.bottom = TempRect.top + g_TextBoxHeight;
  TempRect.left = X;
  TempRect.right = X + 70;

  sprintf (TempString, "%d", m_BigTimes);
  m_BigTimesTextBoxPntr = new BTextControl (TempRect,
    "Big Buffer Times Text Control",
    "Times:" /* label */,
    TempString,
    new BMessage (MSG_BIG_BUF_TIMES),
    B_FOLLOW_LEFT | B_FOLLOW_TOP,
    B_WILL_DRAW | B_FRAME_EVENTS | B_FULL_UPDATE_ON_RESIZE | B_NAVIGABLE);
  if (m_BigTimesTextBoxPntr == NULL) goto ErrorExit;
  m_BigTimesTextBoxPntr->SetDivider (be_plain_font->StringWidth ("Times: "));
  AddChild (m_BigTimesTextBoxPntr);
  m_BigTimesTextBoxPntr->SetTarget (this);
  X += ceilf (m_BigTimesTextBoxPntr->Bounds().Width());

  return X;

ErrorExit:
  return -1.0;
}


void TestWriteFile::PreTestPreparations ()
{
  GetCheckAndSetTextBoxes ();
  TestAbstract::PreTestPreparations ();
}


bool TestWriteFile::DoTest ()
{
  int       AmountRead;
  int       AmountWritten;
  char     *BigBuffer = NULL;
  bigtime_t Duration;
  int       ErrorCode;
  int       FileHandle0 = -1;
  int       FileHandle1 = -1;
  int       i;
  int       MaxLoops;
  char      ReadBuffer [1024];
  bigtime_t StartTime;
  char      TempString [1024];
  int       TotalRead;
  int       TotalWritten;

  /* Create a pair of files. */

  FileHandle0 = open ("/Test/Write1", O_RDWR | O_CREAT | O_TRUNC, S_IRWXU);
  if (FileHandle0 < 0)
  {
    SetTestResult ("Create /Test/Write1 failed", errno);
    goto ErrorExit;
  }

  FileHandle1 = creat ("/Test/Write2", S_IRWXU);
  if (FileHandle1 < 0)
  {
    SetTestResult ("Create /Test/Write2 failed", errno);
    goto ErrorExit;
  }

  /* Write up to 5 megabytes of data, in small 10 byte chunks.  Cut it off
  after 5 seconds, in case debug messages in ARFS_write are slowing the kernel
  code. */

  StartTime = system_time ();
  for (i = 0; i < 500000; i++)
  {
    sprintf (TempString, "%09d\n", i);
    if (write (FileHandle0, TempString, strlen (TempString)) != 10)
    {
      SetTestResult ("Writing small stuff to /Test/Write1 failed", errno);
      goto ErrorExit;
    }
    if ((Duration = system_time () - StartTime) > 5000000)
      break; /* Ran out of time, stop early. */
    if (g_StopTestRequested)
      goto ErrorExit;
  }
  MaxLoops = i;

  if (Duration <= 0) Duration = 1;
  sprintf (TempString, "Wrote %d 10 byte chunks at %f bytes per second.",
    MaxLoops, MaxLoops * 10.0 * 1000000 / Duration);
  SetTestResult (TempString, false);

  /* Try reading it back in, verify that the 5 megabytes is correct. */

  if (lseek (FileHandle0, 0, SEEK_SET) < 0)
  {
    SetTestResult ("lseek to zero of /Test/Write1 failed", errno);
    goto ErrorExit;
  }

  memset (ReadBuffer, 0, sizeof (ReadBuffer));
  StartTime = system_time ();
  for (i = 0; i < MaxLoops; i++)
  {
    TotalRead = 0;
    while (TotalRead < 10)
    {
      AmountRead = read (FileHandle0, ReadBuffer, 10 - TotalRead);
      if (AmountRead <= 0)
      {
        SetTestResult ("First reading small stuff from /Test/Write1 failed",
          errno);
        goto ErrorExit;
      }
      TotalRead += AmountRead;
    }
    sprintf (TempString, "%09d\n", i);
    if (strcmp (TempString, ReadBuffer) != 0)
    {
      SetTestResult ("Data read from /Test/Write1 is wrong.", false);
      goto ErrorExit;
    }
    if (g_StopTestRequested)
      goto ErrorExit;
  }
  Duration = system_time () - StartTime;
  if (Duration <= 0) Duration = 1;
  sprintf (TempString, "Read %d 10 byte chunks at %f bytes per second.",
    MaxLoops, MaxLoops * 10.0 * 1000000 / Duration);
  SetTestResult (TempString, false);

  /* Test seeking - read it backwards. */

  memset (ReadBuffer, 0, sizeof (ReadBuffer));
  StartTime = system_time ();
  for (i = MaxLoops - 1; i >= 0; i--)
  {
    if (lseek (FileHandle0, i * 10, SEEK_SET) < 0)
    {
      ErrorCode = errno;
      sprintf (TempString, "lseek to %d in /Test/Write1 failed", i * 10);
      SetTestResult (TempString, ErrorCode);
      goto ErrorExit;
    }
    TotalRead = 0;
    while (TotalRead < 10)
    {
      AmountRead = read (FileHandle0, ReadBuffer, 10 - TotalRead);
      if (AmountRead <= 0)
      {
        SetTestResult ("Second reading small stuff from /Test/Write1 failed",
          errno);
        goto ErrorExit;
      }
      TotalRead += AmountRead;
    }
    sprintf (TempString, "%09d\n", i);
    if (strcmp (TempString, ReadBuffer) != 0)
    {
      SetTestResult ("Data read from /Test/Write1 is wrong.", false);
      goto ErrorExit;
    }
    if (g_StopTestRequested)
      goto ErrorExit;
  }
  Duration = system_time () - StartTime;
  if (Duration <= 0) Duration = 1;
  sprintf (TempString,
    "Read backwards %d 10 byte chunks at %f bytes per second.",
    MaxLoops, MaxLoops * 10.0 * 1000000 / Duration);
  SetTestResult (TempString, false);

  /* Flush the VNode layer by listing a lot of files, this will flush out our
  vnode and cause the write_vnode routines to be called.  For the RAM file
  system, this packs the data into one big chunk. */

  close (FileHandle0);

  system ("ls -R /boot/home/config >/dev/null");

  FileHandle0 = open ("/Test/Write1", O_RDONLY, S_IRWXU);
  if (FileHandle0 < 0)
  {
    SetTestResult ("Read-only open of /Test/Write1 failed", errno);
    goto ErrorExit;
  }

  /* Try reading it back in again after the flush, verify that the 5 megabytes
  is correct. */

  memset (ReadBuffer, 0, sizeof (ReadBuffer));
  StartTime = system_time ();
  for (i = 0; i < MaxLoops; i++)
  {
    TotalRead = 0;
    while (TotalRead < 10)
    {
      AmountRead = read (FileHandle0, ReadBuffer, 10 - TotalRead);
      if (AmountRead <= 0)
      {
        SetTestResult ("Third reading small stuff from /Test/Write1 failed",
          errno);
        goto ErrorExit;
      }
      TotalRead += AmountRead;
    }
    sprintf (TempString, "%09d\n", i);
    if (strcmp (TempString, ReadBuffer) != 0)
    {
      SetTestResult ("Data read from /Test/Write1 is wrong.", false);
      goto ErrorExit;
    }
    if (g_StopTestRequested)
      goto ErrorExit;
  }
  Duration = system_time () - StartTime;
  if (Duration <= 0) Duration = 1;
  sprintf (TempString,
    "After flush, read %d 10 byte chunks at %f bytes per second.",
    MaxLoops, MaxLoops * 10.0 * 1000000 / Duration);
  SetTestResult (TempString, false);
  snooze (3000000);

  /* Try writing a big file.  Usually a few dozen megabytes.  Write several
  times so that it doesn't use up twice the total size (which can make it swap
  a lot). */

  BigBuffer = (char *) malloc (m_BigBufSize);
  if (BigBuffer == NULL)
  {
    sprintf (TempString, "Unable to allocate %d bytes for the big buffer.",
      m_BigBufSize);
    SetTestResult (TempString, false);
    goto ErrorExit;
  }

  TotalWritten = 0;
  StartTime = system_time ();
  for (i = 0; i < m_BigTimes; i++)
  {
    memset (BigBuffer, i, m_BigBufSize);
    if (g_StopTestRequested) /* The memset can take a long time if swapping. */
      goto ErrorExit;
    AmountWritten = write (FileHandle1, BigBuffer, m_BigBufSize);
    if (AmountWritten <= 0)
    {
      SetTestResult ("Failed to write whole big buffer to /Test/Write2", errno);
      goto ErrorExit;
    }
    TotalWritten += AmountWritten;
  }
  Duration = system_time () - StartTime;
  if (Duration <= 0) Duration = 1;
  if (TotalWritten != m_BigBufSize * m_BigTimes)
  {
    sprintf (TempString, "Only wrote %d bytes of %d to /Test/Write2",
      TotalWritten, m_BigBufSize * m_BigTimes);
    SetTestResult (TempString, false);
    goto ErrorExit;
  }
  sprintf (TempString,
    "Wrote %d byte buffer %d times, averaging %f bytes per second.",
    m_BigBufSize, m_BigTimes, TotalWritten * 1000000.0 / Duration);
  SetTestResult (TempString, false);

  close (FileHandle1);

  if (g_StopTestRequested)
    goto ErrorExit;

  system ("ls -R /boot/home/config >/dev/null");

  FileHandle1 = open ("/Test/Write2", O_RDONLY, S_IRWXU);
  if (FileHandle1 < 0)
  {
    SetTestResult ("Read-only open of /Test/Write2 failed", errno);
    goto ErrorExit;
  }

  if (g_StopTestRequested)
    goto ErrorExit;

  TotalRead = 0;
  StartTime = system_time ();
  while (TotalRead < TotalWritten)
  {
    AmountRead = read (FileHandle1, BigBuffer, m_BigBufSize);
    if (AmountRead <= 0)
    {
      SetTestResult ("Failed to read big buffer from /Test/Write2", errno);
      goto ErrorExit;
    }
    TotalRead += AmountRead;
  }
  Duration = system_time () - StartTime;
  if (Duration <= 0) Duration = 1;
  if (TotalRead != TotalWritten)
  {
    sprintf (TempString, "Only read %d bytes of %d from /Test/Write2",
      TotalRead, TotalWritten);
    SetTestResult (TempString, false);
    goto ErrorExit;
  }
  sprintf (TempString,
    "Read %d bytes with a %d byte buffer at %f bytes per second.",
    TotalRead, m_BigBufSize, TotalRead * 1000000.0 / Duration);
  SetTestResult (TempString, false);

  close (FileHandle1);
  FileHandle1 = -1;

  /* Check the logs here to make sure it doesn't pack already-packed data. */

  system ("ls -R /boot/home/config >/dev/null");

  snooze (3000000);
  SetTestResult ("File creation tests finished successfully.");

ErrorExit:
  if (BigBuffer != NULL)
    free (BigBuffer);
  if (FileHandle0 >= 0)
    close (FileHandle0);
  if (FileHandle1 >= 0)
    close (FileHandle1);
  unlink ("/Test/Write1");
  unlink ("/Test/Write2");

  return true;
}


void TestWriteFile::GetCheckAndSetTextBoxes ()
{
  char TempString [80];

  if (m_BigBufSizeTextBoxPntr != NULL)
  {
    m_BigBufSize = atoi (m_BigBufSizeTextBoxPntr->Text ());
    if (m_BigBufSize < 1)
      m_BigBufSize = 1;
    sprintf (TempString, "%d", m_BigBufSize);
    m_BigBufSizeTextBoxPntr->SetText (TempString);
  }

  if (m_BigTimesTextBoxPntr != NULL)
  {
    m_BigTimes = atoi (m_BigTimesTextBoxPntr->Text ());
    if (m_BigTimes < 1)
      m_BigTimes = 1;
    sprintf (TempString, "%d", m_BigTimes);
    m_BigTimesTextBoxPntr->SetText (TempString);
  }
}



/******************************************************************************
 * Test the wstat (write file status info) internal function.
 */

class TestWStat : public TestAbstract
{
public:
  TestWStat (BRect NewFrame);
  virtual ~TestWStat ();
  virtual void MessageReceived (BMessage *MessagePntr);
  virtual float AddCustomControls (float X);
  virtual bool DoTest ();

  virtual const char *GetDescription ()
    { return "Test wstat system call."; };

  bool       m_GigaSizeTest;
  BCheckBox *m_GigaSizeTestCheckBoxPntr;
};


TestWStat::TestWStat (BRect NewFrame)
: TestAbstract (NewFrame, "TestWStat"),
  m_GigaSizeTest (false),
  m_GigaSizeTestCheckBoxPntr (NULL)
{
  /* Make sure there is space for our Giga Size Test checkbox. */

  if (m_GUIHeight < g_CheckBoxHeight)
    m_GUIHeight = g_CheckBoxHeight;
}


TestWStat::~TestWStat ()
{
  /* Child views (controls) are automatically destroyed and deallocated by
  the parent destructor.  So there isn't much to do here. */

  m_GigaSizeTestCheckBoxPntr = NULL;
}


void TestWStat::MessageReceived (BMessage *MessagePntr)
{
  switch (MessagePntr->what)
  {
    case MSG_GIGA_SIZE_TEST:
      m_GigaSizeTest = m_GigaSizeTestCheckBoxPntr->Value ();
      break;

    default:
      TestAbstract::MessageReceived (MessagePntr);
      break;
  }
}


float TestWStat::AddCustomControls (float X)
{
  float Height;
  float Margin;
  BRect TempRect;
  float Width;

  /* Add the checkbox used to select the giga size test. */

  Margin = ceilf ((m_GUIHeight - g_CheckBoxHeight) / 2);
  TempRect = Bounds ();
  TempRect.top += Margin;
  TempRect.bottom = TempRect.top + g_CheckBoxHeight;
  TempRect.left = X;

  m_GigaSizeTestCheckBoxPntr = new BCheckBox (TempRect,
    "GigaSizeTestBox",
    "Slow GigaTest",
    new BMessage (MSG_GIGA_SIZE_TEST));
  if (m_GigaSizeTestCheckBoxPntr == NULL) goto ErrorExit;
  AddChild (m_GigaSizeTestCheckBoxPntr);
  m_GigaSizeTestCheckBoxPntr->SetTarget (this);
  m_GigaSizeTestCheckBoxPntr->SetValue (m_GigaSizeTest);
  m_GigaSizeTestCheckBoxPntr->ResizeToPreferred ();
  m_GigaSizeTestCheckBoxPntr->GetPreferredSize (&Width, &Height);
  X += ceilf (Width);

  return X;

ErrorExit:
  return -1.0;
}


bool TestWStat::DoTest ()
{
  char   EmptyByte;
  int    ErrorCode;
  gid_t  Group;
  int    i;
  uid_t  OldUserID = geteuid ();
  uid_t  Owner;
  mode_t Permissions;
  off_t  Size;
  char   TempString [1024];
  BFile  TestFile;
  time_t Time;

  ErrorCode = TestFile.SetTo ("/Test/File1", B_READ_WRITE | B_CREATE_FILE);
  if (ErrorCode != 0)
  {
    SetTestResult ("Unable to create /Test/File1", ErrorCode);
    goto ErrorExit;
  }

  ErrorCode = TestFile.SetPermissions (0666);
  if (ErrorCode != 0)
  {
    SetTestResult ("SetPermissions failed", ErrorCode);
    goto ErrorExit;
  }
  ErrorCode = TestFile.GetPermissions (&Permissions);
  if (ErrorCode != 0)
  {
    SetTestResult ("GetPermissions failed", ErrorCode);
    goto ErrorExit;
  }
  if ((Permissions & S_IUMSK) != 0666)
  {
    sprintf (TempString, "Setting permission bits didn't work, got 0%o "
      "instead of 0666.", Permissions);
    SetTestResult (TempString, false);
    goto ErrorExit;
  }

  ErrorCode = TestFile.SetPermissions (-1);
  ErrorCode = TestFile.GetPermissions (&Permissions);
  if ((Permissions & S_IUMSK) != 07777)
  {
    sprintf (TempString, "Setting all permission bits didn't work, got 0%o "
      "instead of 07777.", Permissions);
    SetTestResult (TempString, false);
    goto ErrorExit;
  }

  ErrorCode = TestFile.SetOwner (5);
  ErrorCode = TestFile.GetOwner (&Owner);
  if (Owner != 5)
  {
    SetTestResult ("Setting owner didn't work.", false);
    goto ErrorExit;
  }

  ErrorCode = TestFile.SetGroup (1234);
  ErrorCode = TestFile.GetGroup (&Group);
  if (Group != 1234)
  {
    SetTestResult ("Setting group didn't work.", false);
    goto ErrorExit;
  }

  ErrorCode = TestFile.SetCreationTime (123456789);
  ErrorCode = TestFile.GetCreationTime (&Time);
  if (Time != 123456789)
  {
    SetTestResult ("Setting creation time didn't work.", false);
    goto ErrorExit;
  }

  ErrorCode = TestFile.SetModificationTime (123455555);
  ErrorCode = TestFile.GetModificationTime (&Time);
  if (Time != 123455555)
  {
    SetTestResult ("Setting modification time didn't work.", false);
    goto ErrorExit;
  }

  ErrorCode = TestFile.SetAccessTime (122222222);
  ErrorCode = TestFile.GetAccessTime (&Time);
  if (Time != 122222222)
  {
    SetTestResult ("Setting access time didn't work (BFS doesn't support it).",
      false);
    if (!g_FakeMount) /* BFS doesn't do access time. */
      goto ErrorExit;
  }

  if (!TestFile.IsFile ())
  {
    SetTestResult ("This should be identified as file.", false);
    goto ErrorExit;
  }

  if (TestFile.IsDirectory ())
  {
    SetTestResult ("This should not be identified as a directory.", false);
    goto ErrorExit;
  }

  if (TestFile.IsSymLink ())
  {
    SetTestResult ("This should not be identified as a symbolic link.", false);
    goto ErrorExit;
  }

  ErrorCode = TestFile.Write ("Testing", 7);
  if (ErrorCode != 7)
  {
    SetTestResult ("Failed to write to /Test/File1", ErrorCode);
    goto ErrorExit;
  }

  ErrorCode = TestFile.GetSize (&Size);
  if (Size != 7)
  {
    SetTestResult ("GetSize got wrong size, should be 7.", false);
    goto ErrorExit;
  }

  /* Make sure that you have write access before changing a file. */

  ErrorCode = TestFile.SetPermissions (0444); /* r--r--r-- read only perms. */
  setuid (42);
  ErrorCode = TestFile.SetSize (50000000);
  if (ErrorCode != EACCES)
  {
    SetTestResult ("SetSize on a read only file should get EACCES",
      ErrorCode);
    goto ErrorExit;
  }

  /* Back to being root user, should be able to write file even if permissions
  say it is read only. */

  setuid (OldUserID);
  ErrorCode = TestFile.SetSize (50000000);
  if (ErrorCode != 0)
  {
    SetTestResult ("SetSize (50000000) failed", ErrorCode);
    goto ErrorExit;
  }

  ErrorCode = TestFile.GetSize (&Size);
  if (Size != 50000000)
  {
    SetTestResult ("GetSize got wrong size, should be 50000000.", false);
    goto ErrorExit;
  }

  /* See if sparse data in the middle is blanked out, or is there data left
  over from other files in there? */

  ErrorCode = TestFile.WriteAt (1024 * 1024 * 10, "Middle", 6);
  if (ErrorCode != 6)
  {
    SetTestResult ("Unable to write in the middle of a sparse file",
      ErrorCode);
    goto ErrorExit;
  }

  memset (TempString, 0, sizeof (TempString));
  ErrorCode =
    TestFile.ReadAt (1024 * 1024 * 10 - 512, TempString, sizeof (TempString));
  if (ErrorCode != sizeof (TempString))
  {
    SetTestResult ("Unable to read from the middle of a sparse file",
      ErrorCode);
    goto ErrorExit;
  }

  /* Verify that empty sparse space around the written data is zero (the POSIX
  standard says it should be zero). */

  EmptyByte = 0 /* TempString [0] */;
  for (i = 0; i < 512; i++)
    if (TempString[i] != EmptyByte)
      break;
  if (i != 512)
  {
    SetTestResult ("Sparse data before the write isn't blanked out, "
      "security hazard!", false);
    goto ErrorExit;
  }
  if (memcmp (TempString + 512, "Middle", 6) != 0)
  {
    SetTestResult ("Data read in the middle of a sparse area is wrong.",
      false);
    goto ErrorExit;
  }
  for (i = 512 + 6; i < (int) sizeof (TempString); i++)
    if (TempString[i] != EmptyByte)
      break;
  if (i != sizeof (TempString))
  {
    SetTestResult ("Sparse data after the write isn't blanked out, "
      "security hazard!", false);
    goto ErrorExit;
  }

  ErrorCode = TestFile.SetSize (4);
  if (ErrorCode != 0)
  {
    SetTestResult ("SetSize (4) didn't work", ErrorCode);
    goto ErrorExit;
  }

  ErrorCode = TestFile.GetSize (&Size);
  if (Size != 4)
  {
    SetTestResult ("GetSize got wrong size, should be 4.", false);
    goto ErrorExit;
  }

  memset (TempString, 0, sizeof (TempString));
  ErrorCode = TestFile.ReadAt (0, TempString, sizeof (TempString));
  if (ErrorCode != 4)
  {
    SetTestResult ("Unable to read from the middle of a sparse file",
      ErrorCode);
    goto ErrorExit;
  }
  if (strcmp (TempString, "Test") != 0)
  {
    SetTestResult ("Contents of shrunk file are wrong.", false);
    goto ErrorExit;
  }

  /* Try a size which is too big for RAM (about 1GB available under BeOS to
  the kernel drivers, even if you have more memory). */

  if (m_GigaSizeTest)
  {
    SetTestResult ("SetSize(1500000000) test in progress, can be slow due "
      "to swapping, can't stop early (reset your computer to abort).");
    ErrorCode = TestFile.SetSize (1500000000);
    if (ErrorCode == 0)
    {
      SetTestResult ("SetSize (1500000000) succeeded, "
        "guess it's not a RAM based file system.", false);
      if (!g_FakeMount) /* BFS can handle it, even with not much free space. */
        goto ErrorExit;
    }
  }
  if (g_StopTestRequested)
    goto ErrorExit;

  /* Try a size which is too big for BFS (33GB internal structure limit), and
  overflows 32 bits too. */

  if (m_GigaSizeTest)
  {
    SetTestResult ("SetSize(1000000000000) test in progress, can be slow due "
      "to swapping, can't stop early (reset your computer to abort).");
    ErrorCode = TestFile.SetSize (1000000000000LL);
    if (ErrorCode == 0)
    {
      SetTestResult ("SetSize (1000000000000) succeeded, does this file system "
        "support sparse files?", false);
      if (!g_FakeMount) /* BFS can handle it, usually. */
        goto ErrorExit;
    }
  }

  SetTestResult ("wstat tests finished successfully.");

ErrorExit:
  setuid (OldUserID);
  TestFile.Unset ();
  unlink ("/Test/File1");

  return true;
}



/******************************************************************************
 * This view just draws a solid colour background (since we can't change
 * the window's background).  Child views can be added to this.
 */

class BackgroundColourView : public BView
{
public:
  int red, green, blue;
  int dred, dgreen, dblue;

  BackgroundColourView (BRect NewBounds)
  : BView (NewBounds, "Background View", B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP,
    B_WILL_DRAW | B_PULSE_NEEDED),
    red (255), green (255), blue (255),
    dred (1), dgreen (2), dblue (3)
  {
  };

  virtual void Pulse ();
};


void BackgroundColourView::Pulse ()
{
  BView *ChildViewPntr;
  int    i;

  red += dred; green += dgreen; blue += dblue;
  // Note range 159-255 has 97 possible values, a prime number, so it won't
  // repeat the colours for a long time.  Also, 159-255 is bright enough to
  // keep black text readable.
  if (red < 159 || red > 255) {dred = -dred; red += dred;};
  if (green < 159 || green > 255) {dgreen = -dgreen; green += dgreen;};
  if (blue < 159 || blue > 255) {dblue = -dblue; blue += dblue;};

  SetViewColor (red, green, blue);
  Invalidate ();

  for (i = 0; (ChildViewPntr = ChildAt (i)) != NULL; i++)
  {
    ChildViewPntr->SetViewColor (red, green, blue);
    ChildViewPntr->SetLowColor (red, green, blue);
    ChildViewPntr->Invalidate ();
  }
}



/******************************************************************************
 * This view contains a bunch of test subviews, in order of the test sequence.
 * It is usually a subview of a scroll view so that you can have lots of tests.
 * You can also tell this view to set up particular test sequences.  It is
 * sized to fit in the visible scrolling area, even though the subviews (tests)
 * may be bigger or smaller.
 */

class ArrayOfTestsView : public BView
{
public:
  ArrayOfTestsView (BRect NewBounds);

  bool AddStandardTests ();
  virtual void AttachedToWindow ();
  virtual void FrameResized (float Width, float Height);
  virtual void MessageReceived (BMessage *MessagePntr);
  void ResizeScrollBars ();
  virtual void TargetedByScrollView (BScrollView *ScrollViewPntr);

private:
  bool AddATest (TestAbstract *TestPntr);
  float m_WholeHeight;
  float m_WholeWidth;
  BScrollView *m_ScrollViewPntr;
};


ArrayOfTestsView::ArrayOfTestsView (BRect NewBounds)
: BView (NewBounds, "Array of Tests View",
  B_FOLLOW_ALL_SIDES, B_FRAME_EVENTS),
  m_WholeHeight (20),
  m_WholeWidth (500),
  m_ScrollViewPntr (NULL)
{
};


bool ArrayOfTestsView::AddATest (TestAbstract *TestPntr)
{
  float         Height;
  float         Width;

  if (TestPntr == NULL)
  {
    DisplayErrorMessage ("AddATest: NULL pointer, test not allocated?");
    return false;
  }

  AddChild (TestPntr);

  TestPntr->GetPreferredSize (&Width, &Height);

  if (Width > m_WholeWidth)
    m_WholeWidth = Width;

  m_WholeHeight += Height + 2; /* Off by 1 bug plus 1 for gap between rows. */
  return true;
}


bool ArrayOfTestsView::AddStandardTests ()
{
  BRect         TempRect;

  TempRect = Bounds ();
  m_WholeHeight = 0;
  m_WholeWidth = 0;

  if (!AddATest (new TestMount (TempRect))) goto ErrorExit;
  TempRect.top = m_WholeHeight;

  if (!AddATest (new TestFakeSymbolicMount (TempRect))) goto ErrorExit;
  TempRect.top = m_WholeHeight;

  if (!AddATest (new TestMakeDirectory (TempRect))) goto ErrorExit;
  TempRect.top = m_WholeHeight;

  if (!AddATest (new TestCreateFile (TempRect))) goto ErrorExit;
  TempRect.top = m_WholeHeight;

  if (!AddATest (new TestWriteFile (TempRect))) goto ErrorExit;
  TempRect.top = m_WholeHeight;

  if (!AddATest (new TestPermissions (TempRect))) goto ErrorExit;
  TempRect.top = m_WholeHeight;

  if (!AddATest (new TestRename (TempRect))) goto ErrorExit;
  TempRect.top = m_WholeHeight;

  if (!AddATest (new TestUMask (TempRect))) goto ErrorExit;
  TempRect.top = m_WholeHeight;

  if (!AddATest (new TestWStat (TempRect))) goto ErrorExit;
  TempRect.top = m_WholeHeight;

  if (!AddATest (new TestMaxTotalSize (TempRect))) goto ErrorExit;
  TempRect.top = m_WholeHeight;

  if (!AddATest (new TestSymbolicLink (TempRect))) goto ErrorExit;
  TempRect.top = m_WholeHeight;

  if (!AddATest (new TestAGMSLink (TempRect))) goto ErrorExit;
  TempRect.top = m_WholeHeight;

  if (!AddATest (new TestNotification (TempRect))) goto ErrorExit;
  TempRect.top = m_WholeHeight;

  if (!AddATest (new TestVolumeStuff (TempRect))) goto ErrorExit;
  TempRect.top = m_WholeHeight;

  if (!AddATest (new TestAttributes (TempRect))) goto ErrorExit;
  TempRect.top = m_WholeHeight;

  if (!AddATest (new TestIndices (TempRect))) goto ErrorExit;
  TempRect.top = m_WholeHeight;

  if (!AddATest (new TestQuery (TempRect))) goto ErrorExit;
  TempRect.top = m_WholeHeight;

  if (!AddATest (new TestUnmount (TempRect))) goto ErrorExit;
  TempRect.top = m_WholeHeight;

  /* Compensate for off by 1 and also don't include black separator line. */
  m_WholeHeight -= 2;
  return true;

ErrorExit:
  DisplayErrorMessage ("AddStandardTests: Unable to create all the tests.");
  return false;
}


void ArrayOfTestsView::AttachedToWindow ()
{
  ResizeScrollBars ();
  SetViewColor (0, 0, 0); /* Black for black lines between test rows. */
}


void ArrayOfTestsView::FrameResized (float Width, float Height)
{
  ResizeScrollBars ();
}


void ArrayOfTestsView::MessageReceived (BMessage *MessagePntr)
{
  float       BigStep;
  float       DeltaX; /* Usually -1.0, 0.0 or +1.0. */
  float       DeltaY;
  BScrollBar *ScrollBarPntr;
  float       SmallStep;

  switch (MessagePntr->what)
  {
    case B_MOUSE_WHEEL_CHANGED:
      if (m_ScrollViewPntr == NULL) break; /* Not initialised yet. */

      /* Do sideways scrolling, in case someone ever invents a mouse wheel
      which goes sideways.  Don't do scrolling if the scroll bar isn't active
      (whole area width fits in window width).  Also use actual delta, in case
      fractional movements are ever implemented. */

      if (MessagePntr->FindFloat ("be:wheel_delta_x", &DeltaX) != 0) break;
      if (DeltaX != 0.0 && m_WholeWidth > Bounds().Width())
      {
        ScrollBarPntr = m_ScrollViewPntr->ScrollBar (B_HORIZONTAL);
        if (ScrollBarPntr == NULL) break;
        ScrollBarPntr->GetSteps (&SmallStep, &BigStep);
        ScrollBarPntr->SetValue (ScrollBarPntr->Value() + DeltaX * SmallStep);
      }

      /* Vertical mouse wheel scrolling. */

      if (MessagePntr->FindFloat ("be:wheel_delta_y", &DeltaY) != 0) break;
      if (DeltaY != 0.0 && m_WholeHeight > Bounds().Height())
      {
        ScrollBarPntr = m_ScrollViewPntr->ScrollBar (B_VERTICAL);
        if (ScrollBarPntr == NULL) break;
        ScrollBarPntr->GetSteps (&SmallStep, &BigStep);
        ScrollBarPntr->SetValue (ScrollBarPntr->Value() + DeltaY * SmallStep);
      }
      break;

    default:
      BView::MessageReceived (MessagePntr);
      break;
  }
}


void ArrayOfTestsView::ResizeScrollBars ()
{
  float       Height;
  BScrollBar *ScrollBarPntr;
  float       Width;

  Height = Bounds().Height();
  Width = Bounds().Width();

  if (m_ScrollViewPntr == NULL)
    return;

  ScrollBarPntr = m_ScrollViewPntr->ScrollBar (B_HORIZONTAL);
  if (ScrollBarPntr != NULL)
  {
    if (m_WholeWidth <= Width)
      ScrollBarPntr->SetRange (0.0, 0.0); // Fully fits, turn off scroll bar.
    else
    {
      ScrollBarPntr->SetRange (0.0, m_WholeWidth - Width);
      ScrollBarPntr->SetProportion (Width / m_WholeWidth);
      ScrollBarPntr->SetSteps (10.0, Width / 2);
    }
  }

  ScrollBarPntr = m_ScrollViewPntr->ScrollBar (B_VERTICAL);
  if (ScrollBarPntr != NULL)
  {
    if (m_WholeHeight <= Height)
      ScrollBarPntr->SetRange (0.0, 0.0); // Fully fits, turn off scroll bar.
    else
    {
      ScrollBarPntr->SetRange (0.0, m_WholeHeight - Height);
      ScrollBarPntr->SetProportion (Height / m_WholeHeight);
      ScrollBarPntr->SetSteps (g_TextBoxHeight + 1.0, Height / 2);
    }
  }
}


void ArrayOfTestsView::TargetedByScrollView (BScrollView *ScrollViewPntr)
{
  m_ScrollViewPntr = ScrollViewPntr;
}



/******************************************************************************
 * This window displays a list of tests (using a scrolling subview) and has
 * a GO button to run the tests and an ABOUT button.
 */

class TestListWindow : public BWindow
{
public:
  TestListWindow ();

  bool InitialiseContents ();
  virtual void MessageReceived (BMessage *MessagePntr);
  virtual bool QuitRequested ();
  void RunTests ();

private:
  /* Disable the default copy and assignment operators by privatizing. */
  TestListWindow (const TestListWindow&);
  TestListWindow& operator= (const TestListWindow&);

  ArrayOfTestsView *m_ArrayOfTestsViewPntr;
  BButton *m_RunButtonPntr;
  bool m_StopAtFirstFailure;
  BCheckBox *m_StopFirstCheckboxPntr;
  bool m_TestRunning;
};


TestListWindow::TestListWindow ()
: BWindow (
    BRect (30, 30, 1000, 350),
    "Alexander G. M. Smith's Random Access Memory File System Mounter",
    B_DOCUMENT_WINDOW, 0),
  m_ArrayOfTestsViewPntr (NULL),
  m_RunButtonPntr (NULL),
  m_StopAtFirstFailure (false),
  m_StopFirstCheckboxPntr (NULL),
  m_TestRunning (false)
{
}


/* The window is now open but empty.  Add a few buttons and subviews.
Returns TRUE if successful, FALSE if failed.  Called from the application
thread, with the window already existing, so it needs to lock where
appropriate. */

bool TestListWindow::InitialiseContents ()
{
  BackgroundColourView *BackgroundViewPntr;
  float                 BackgroundHeight;
  BScrollView          *ScrollViewPntr;
  BRect                 TempRect;
  float                 X;

  /* Create and add the background colour view, which also holds the buttons
  and other controls outside the main scrolling subview of tests.  It fills
  the top part of the window. */

  BackgroundHeight = g_ButtonHeight;
  if (g_CheckBoxHeight > BackgroundHeight)
    BackgroundHeight = g_CheckBoxHeight;
  BackgroundHeight += 5;
  TempRect = Bounds ();
  TempRect.bottom = TempRect.top + BackgroundHeight;
  X = TempRect.left;

  BackgroundViewPntr = new BackgroundColourView (TempRect);
  if (BackgroundViewPntr == NULL) goto ErrorExit;
  AddChild (BackgroundViewPntr);

  /* Add some controls to the background view.  Start with Run Test button,
  which will indirectly command the application thread to run the tests. */

  TempRect = Bounds ();
  TempRect.left = X;
  TempRect.top += ceilf ((BackgroundHeight - g_ButtonHeight) / 2);
  TempRect.bottom = TempRect.top + g_ButtonHeight;
  m_RunButtonPntr = new BButton (TempRect, "Run Button", "Run Tests",
    new BMessage (MSG_RUN_BUTTON));
  if (m_RunButtonPntr == NULL) goto ErrorExit;
  BackgroundViewPntr->AddChild (m_RunButtonPntr);
  Lock ();
  m_RunButtonPntr->ResizeToPreferred ();
  X = m_RunButtonPntr->Bounds().right + g_MarginBetweenControls;
  Unlock ();

  /* Add the Stop At First Failure checkbox. */

  TempRect = Bounds ();
  TempRect.left = X;
  TempRect.top += ceilf ((BackgroundHeight - g_CheckBoxHeight) / 2);
  TempRect.bottom = TempRect.top + g_CheckBoxHeight;
  m_StopFirstCheckboxPntr = new BCheckBox (TempRect, "Stop Check",
    "Stop at First Failure", new BMessage (MSG_STOP_AT_FIRST_FAILURE));
  if (m_StopFirstCheckboxPntr == NULL) goto ErrorExit;
  BackgroundViewPntr->AddChild (m_StopFirstCheckboxPntr);
  Lock ();
  m_StopFirstCheckboxPntr->SetValue (m_StopAtFirstFailure);
  m_StopFirstCheckboxPntr->ResizeToPreferred ();
  X = m_StopFirstCheckboxPntr->Bounds().right + g_MarginBetweenControls;
  Unlock ();

  /* Add the scrolling list of tests. */

  TempRect = Bounds ();
  TempRect.top = TempRect.top + BackgroundHeight + 2;
  TempRect.right -= B_V_SCROLL_BAR_WIDTH;
  TempRect.bottom -= B_H_SCROLL_BAR_HEIGHT;
  m_ArrayOfTestsViewPntr = new ArrayOfTestsView (TempRect);
  if (m_ArrayOfTestsViewPntr == NULL) goto ErrorExit;
  if (!m_ArrayOfTestsViewPntr->AddStandardTests ()) goto ErrorExit;

  ScrollViewPntr = new BScrollView (
    "BScrollView containing an ArrayOfTestsView",
    m_ArrayOfTestsViewPntr,
    B_FOLLOW_ALL_SIDES,
    B_WILL_DRAW /* for drawing the border */,
    true /* horizontal scroll bar */,
    true /* vertical scroll bar */,
    B_PLAIN_BORDER);
  if (ScrollViewPntr == NULL) goto ErrorExit;

  AddChild (ScrollViewPntr);
  return true;

ErrorExit:
  DisplayErrorMessage ("Unable to initialise the window contents.");
  return false;
}


void TestListWindow::MessageReceived (BMessage *MessagePntr)
{
  switch (MessagePntr->what)
  {
    case MSG_RUN_BUTTON:
      if (m_TestRunning)
      {
        g_StopTestRequested = true;
        m_RunButtonPntr->SetLabel ("Stopping...");
      }
      else
        be_app->PostMessage (MSG_RUN_TESTS);
      break;

    case MSG_STOP_AT_FIRST_FAILURE:
      m_StopAtFirstFailure = !m_StopAtFirstFailure;
      m_StopFirstCheckboxPntr->SetValue (m_StopAtFirstFailure);
      break;

    case B_MOUSE_WHEEL_CHANGED:
      if (m_ArrayOfTestsViewPntr != NULL)
        m_ArrayOfTestsViewPntr->MessageReceived (MessagePntr);
      break;

    default:
      BWindow::MessageReceived (MessagePntr);
      break;
  }
}


/* Override the virtual function (usually called when the close box is clicked)
to make the whole application quit when the window is closed, except when a
test is running.  Doesn't actually close the window if a test is running (so
that the window doesn't disappear while the application is running a test),
but sends a message to the application to close, which it will execute once it
has finished the test. */

bool TestListWindow::QuitRequested ()
{
  if (m_TestRunning)
  {
    g_StopTestRequested = true;
    m_RunButtonPntr->SetLabel ("Closing...");
    be_app->PostMessage (B_QUIT_REQUESTED);
    return false;
  }

  /* Send a possibly redundant quit message, in case this was invoked by the
  window closing (otherwise the application would exist with no windows).
  In the case where this is invoked by the application quitting, the extra
  message won't do anything. */

  be_app->PostMessage (B_QUIT_REQUESTED);
  return true;
}


/* This is called by the application thread to do the tests, not the window
thread as you'd normally expect.  Simultaneously, the window thread can show
test results, and let you click on the stop button (which sets
g_StopTestRequested to true). */

void TestListWindow::RunTests ()
{
  int           TestIndex;
  TestAbstract *TestPntr;
  bool          TotalSuccess;

  if (m_TestRunning || m_ArrayOfTestsViewPntr == NULL ||
  m_RunButtonPntr == NULL)
    return; /* Don't allow unexpected calls. */

  if (!Lock ()) return; /* Window closed or something wrong. */
  m_TestRunning = true;
  g_StopTestRequested = false;
  m_RunButtonPntr->SetLabel ("Stop Test");

  /* Mark all the tests as not being done yet, clearing out the results of
  previous tests, in case we stop early.  Also updates text boxes to reflect
  whatever typing the user has done. */

  for (TestIndex = 0; (TestPntr = dynamic_cast<TestAbstract*>
  (m_ArrayOfTestsViewPntr->ChildAt (TestIndex))) != NULL; TestIndex++)
  {
    TestPntr->PreTestPreparations ();
  }
  Unlock ();

  TestIndex = -1;
  TotalSuccess = true;
  while ((!m_StopAtFirstFailure || TotalSuccess)
  && !g_StopTestRequested)
  {
    /* Get the next test object, from the array of tests view. */

    TestIndex++;
    if (!Lock ())
      break; /* Window closed or something wrong with it. */
    TestPntr = dynamic_cast<TestAbstract*>
      (m_ArrayOfTestsViewPntr->ChildAt (TestIndex));
    Unlock ();
    if (TestPntr == NULL)
      break; /* No more tests to do. */

    /* Run the test.  Returns TRUE if OK to continue to next test. */

    if (!TestPntr->TestAndFrills ())
    {
      TotalSuccess = false;
      break; /* A test has failed badly and wants to stop. */
    }

    if (!TestPntr->GetTestWasSuccessful ())
      TotalSuccess = false; /* A test has failed ordinarily. */
  }

  m_TestRunning = false;
  if (Lock ())
  {
    m_RunButtonPntr->SetLabel ("Run Tests");
    Unlock ();
  }

  if (!TotalSuccess)
    DisplayErrorMessage ("Testing has failed, not all tests were successful.",
    0, "Test Results");
}



/******************************************************************************
 * The main application class.  It opens a window and responds to commands
 * from top level buttons to do the tests.  The separate window thread keeps
 * the UI live and handles simpler (non-time-consuming) things like the
 * settings for each test.
 */

class MounterApp : public BApplication
{
public:
  MounterApp ();
  virtual void MessageReceived (BMessage *MessagePntr);
  virtual void ReadyToRun ();

  TestListWindow *m_TestListWindowPntr;

private:
  /* Disable the default copy and assignment operators by privatizing. */
  MounterApp (const MounterApp&);
  MounterApp& operator= (const MounterApp&);
};


MounterApp::MounterApp ()
: BApplication ("application/x-vnd.agmsmith.AGMSRAMFSMounter"),
  m_TestListWindowPntr (NULL)
{
}


void MounterApp::MessageReceived (BMessage *MessagePntr)
{
  switch (MessagePntr->what)
  {
    case MSG_RUN_TESTS:
      m_TestListWindowPntr->RunTests ();
      break;

    default:
      BApplication::MessageReceived (MessagePntr);
      break;
  }
}


void MounterApp::ReadyToRun ()
{
  float         JunkFloat;
  BButton      *TempButtonPntr;
  BCheckBox    *TempCheckBoxPntr;
  font_height   TempFontHeight;
  BMenuBar     *TempMenuBarPntr;
  BMenuItem    *TempMenuItemPntr;
  BPopUpMenu   *TempPopUpMenuPntr;
  BRadioButton *TempRadioButtonPntr;
  BRect         TempRect;
  const char   *TempString = "Testing My Things";
  BStringView  *TempStringViewPntr;
  BTextControl *TempTextPntr;

  /* Create the new window showing a set of tests, initially empty. */

  m_TestListWindowPntr = new TestListWindow;
  if (m_TestListWindowPntr == NULL)
  {
    DisplayErrorMessage ("Unable to create window.");
    Quit (); /* Tell the BApplication object to quit running. */
    return;
  }

  m_TestListWindowPntr->Show ();

  /* Set the spacing between buttons and other controls to the width of
  the letter "M" in the user's desired font. */

 g_MarginBetweenControls = (int) be_plain_font->StringWidth ("M");

  /* Also find out how much space a line of text uses. */

  be_plain_font->GetHeight (&TempFontHeight);
  g_LineOfTextHeight = ceilf (
    TempFontHeight.ascent + TempFontHeight.descent + TempFontHeight.leading);

  /* Start finding out the height of various user interface gadgets, which can
  vary based on the current font size.  Make a temporary gadget, which is
  attached to our window, then resize it to its prefered size so that it
  accomodates the font size and other frills it needs. */

  TempRect = m_TestListWindowPntr->Bounds ();

  /* Find the height of a single line of text in a BStringView. */

  TempStringViewPntr = new BStringView (TempRect, TempString, TempString);
  if (TempStringViewPntr != NULL)
  {
    m_TestListWindowPntr->Lock ();
    m_TestListWindowPntr->AddChild (TempStringViewPntr);
    TempStringViewPntr->GetPreferredSize (&JunkFloat, &g_StringViewHeight);
    m_TestListWindowPntr->RemoveChild (TempStringViewPntr);
    m_TestListWindowPntr->Unlock ();
    delete TempStringViewPntr;
  }

  /* Find the height of a button, which seems to be larger than a text
  control and can make life difficult.  Make a temporary button, which
  is attached to our window so that it resizes to accomodate the font size. */

  TempButtonPntr = new BButton (TempRect, TempString, TempString, NULL);
  if (TempButtonPntr != NULL)
  {
    m_TestListWindowPntr->Lock ();
    m_TestListWindowPntr->AddChild (TempButtonPntr);
    TempButtonPntr->GetPreferredSize (&JunkFloat, &g_ButtonHeight);
    m_TestListWindowPntr->RemoveChild (TempButtonPntr);
    m_TestListWindowPntr->Unlock ();
    delete TempButtonPntr;
  }

  /* Find the height of a text box. */

  TempTextPntr = new BTextControl (TempRect, TempString, NULL /* label */,
    TempString, NULL);
  if (TempTextPntr != NULL)
  {
    m_TestListWindowPntr->Lock ();
    m_TestListWindowPntr->AddChild (TempTextPntr);
    TempTextPntr->GetPreferredSize (&JunkFloat, &g_TextBoxHeight);
    m_TestListWindowPntr->RemoveChild (TempTextPntr);
    m_TestListWindowPntr->Unlock ();
    delete TempTextPntr;
  }

  /* Find the height of a checkbox control. */

  TempCheckBoxPntr = new BCheckBox (TempRect, TempString, TempString, NULL);
  if (TempCheckBoxPntr != NULL)
  {
    m_TestListWindowPntr->Lock ();
    m_TestListWindowPntr->AddChild (TempCheckBoxPntr);
    TempCheckBoxPntr->GetPreferredSize (&JunkFloat, &g_CheckBoxHeight);
    m_TestListWindowPntr->RemoveChild (TempCheckBoxPntr);
    m_TestListWindowPntr->Unlock ();
    delete TempCheckBoxPntr;
  }

  /* Find the height of a radio button control. */

  TempRadioButtonPntr =
    new BRadioButton (TempRect, TempString, TempString, NULL);
  if (TempRadioButtonPntr != NULL)
  {
    m_TestListWindowPntr->Lock ();
    m_TestListWindowPntr->AddChild (TempRadioButtonPntr);
    TempRadioButtonPntr->GetPreferredSize (&JunkFloat, &g_RadioButtonHeight);
    m_TestListWindowPntr->RemoveChild (TempRadioButtonPntr);
    m_TestListWindowPntr->Unlock ();
    delete TempRadioButtonPntr;
  }

  /* Find the height of a pop-up menu. */

  TempMenuBarPntr = new BMenuBar (TempRect, TempString,
    B_FOLLOW_LEFT | B_FOLLOW_TOP, B_ITEMS_IN_COLUMN,
    true /* resize to fit items */);
  TempPopUpMenuPntr = new BPopUpMenu (TempString);
  TempMenuItemPntr = new BMenuItem (TempString, new BMessage (12345), 'g');

  if (TempMenuBarPntr != NULL && TempPopUpMenuPntr != NULL &&
  TempMenuItemPntr != NULL)
  {
    TempPopUpMenuPntr->AddItem (TempMenuItemPntr);
    TempMenuBarPntr->AddItem (TempPopUpMenuPntr);

    m_TestListWindowPntr->Lock ();
    m_TestListWindowPntr->AddChild (TempMenuBarPntr);
    TempMenuBarPntr->GetPreferredSize (&JunkFloat, &g_PopUpMenuHeight);
    m_TestListWindowPntr->RemoveChild (TempMenuBarPntr);
    m_TestListWindowPntr->Unlock ();
    delete TempMenuBarPntr; // It will delete contents too.
  }

  if (!m_TestListWindowPntr->InitialiseContents ())
  {
    DisplayErrorMessage ("Something went wrong while initialising the "
      "contents of the Test list window.");
    Quit (); /* Tell this BApplication to shut down. */
  }
}



/******************************************************************************
 * Finally, the main program which drives it all.
 */

int main (int argc, char** argv)
{
  MounterApp MyApp;

  MyApp.Run ();

  return 0;
}
